/* NSC_LIC_BSD
 *
 * National Semiconductor Corporation Open Source License for
 *
 *    Video library for Geode SCx2x0
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c) 2000-2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use by Licensee of this Software in source and
 * binary forms, with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   * Redistributions in binary form must either:
 *     (a) reproduce the above copyright notice, this list of conditions
 *         and the following disclaimer in the documentation and/or
 *         other materials provided with the distribution, or
 *     (b) if Licensee elects to redistribute any Software not in
 *         accordance with (a), then Licensee shall indemnify National
 *         Semiconductor Corporation and all copyright holders and
 *         contributors for any claims related to such Software (and pay
 *         all costs and damages, including all attorney fees)
 *
 *   * Neither the name of the National Semiconductor Corporation nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF
 * YOUR JURISDICTION. It is Licensee's responsibility to comply with
 * any export regulations applicable in Licensee's jurisdiction. Under
 * CURRENT (July 2001) U.S. export regulations this software
 * is eligible for export from the U.S. and can be downloaded by or
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran,
 * Syria, Sudan, Afghanistan and any other country to which the U.S.
 * has embargoed goods and services.
 *
 * END_NSC_LIC_BSD */
 
 /*--------------------------------------------------------------------------
 * Copyright (c) 2000-2001 by National Semiconductor Corporation
 * All rights reserved.
 *--------------------------------------------------------------------------
 * File Contents:
 *           Functions for configuring and controlling the SCx2x0 video
 *           overlay on graphics and video alpha blending with graphics.
 * Project : Video library for Geode SCx2x0
 * Created : November, 2000
 * Author  : Galit Heller & Eyal Cohen
 *--------------------------------------------------------------------------
 * $Date: 8/10/01 6:58a $
 * $RCSfile: vl_alpha.c $
 * $ProjectRevision: 3.1 $
 * $Revision: 6 $
 *------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/*-------------------------   Include files   ----------------------------*/
/*------------------------------------------------------------------------*/
#include "gfx_rtns.h"
#include "vl_rtns.h"
#include "vl_defs.h"

/*------------------------------------------------------------------------*/
/*----------------   Constants and macros definitions  -------------------*/
/*------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/*---------------------   Data Types definitions   -----------------------*/
/*------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/*----------------------   Variable definitions   ------------------------*/
/*------------------------------------------------------------------------*/
unsigned long color_key;
unsigned long color_mask;
boolean gfx_overlay;

/*-------------*/
/*-- Externs --*/
/*-------------*/

/**************************************************************************/
/**************************   Alpha Blending   ****************************/
/**************************************************************************/


// Three alpha windows are available for use by the following functions.
// The numbers of the windows are 0 - 2.


//---------------------------------------------------------------------
// Description:
//   Set the specified alpha window position and size relative to the
//   graphics resolution.
//
// Parameters:
//   number          - Alpha window number (0..2).
//   x , y           - Window position.
//   width , height  - Window size.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//
// Side Effects:
//   If graphics aspect ratio is kept, the actual horizontal start position is
//   the nearest 8/12 (NTSC/PAL) multiple nearest to x. If the video x
//   start and the alpha x start are not rounded by the same amount, then
//   in borderline cases the alpha window might continue past the video window.
//---------------------------------------------------------------------
VL_STATUS VL_VideoSetAlphaWindow(unsigned char number, short x, short y,
   unsigned short width, unsigned short height)
{
	if (gfx_select_alpha_region(number) != GFX_STATUS_OK)
		return VL_STATUS_BAD_PARAMETER;
		
	// IF ASPECT RATIO SCALING IS PERFORMED ON THE VIDEO, ADJUSTMENTS MUST BE MADE:
	// HORIZONTAL START POSITION AND SIZE ARE MULTIPLED BY THE ASPECT RATIO SCALING, AND
	// HORIZONTAL START POSITION IS ROUNDED TO 8 or 12 PIXEL MULTIPLES
	// (TO MINIMIZE SCALING ARTIFACTS)
	if (VidDisplay.horzFactor.num != VidDisplay.horzFactor.denom) {
		x = ASPECT_RATIO_VIDEO_PRESCALE(x);
		width = ASPECT_RATIO_VIDEO_PRESCALE(width);
		if (VidAttr.TVStandard == VL_TV_STANDARD_NTSC)
			x = (x + 4) & (~0x7);
		else if (VidAttr.TVStandard == VL_TV_STANDARD_PAL)
			x = ((x + 6) / 12) * 12;
	}		

	gfx_set_alpha_window(x, y + VidDisplay.extra_y_offset, width, height);
	return VL_STATUS_OK;
} 

//---------------------------------------------------------------------
// Description:
//   Set the relevant values for the alpha window operation desired.
//
// Parameters:
//   number       - Alpha window number (0..2).
//   alpha        - Alpha value.
//   color        - Alpha color.
//   color_enable - Alpha color enable.
//   priority     - Alpha window priority (0..2).
//   fade         - Alpha increment/decrement value per frame.
//   enable       - Alpha window enable/disable.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS VL_VideoSetAlphaOperation(unsigned char number, unsigned char alpha,
   unsigned long color, boolean color_enable, int priority, char fade, boolean enable)
{
	if (gfx_select_alpha_region(number) != GFX_STATUS_OK)
		return VL_STATUS_BAD_PARAMETER;
		
	gfx_set_alpha_priority(priority);
	gfx_set_alpha_color(color);
	gfx_set_alpha_color_enable(color_enable);
	
	// Enable alpha window before setting fade in order to see it
	gfx_set_alpha_enable(enable);
	gfx_set_alpha_value(alpha, fade);
	return VL_STATUS_OK;
}   


/**************************************************************************/
/*****************************   Color Key   ******************************/
/**************************************************************************/


//---------------------------------------------------------------------
// Description:
//   Set color key and mask values.
//   The color key value is programmed immediately only if graphics
//   and video are both displayed at the time of the call, otherwise
//   it is saved internally and used when needed.
//
// Parameters:
//   key   - Color/Chroma key value.
//   mask  - Color/Chroma key mask.
//
// Return Value:
//   VL_STATUS_OK.
//---------------------------------------------------------------------
VL_STATUS VL_VideoSetColorKey(unsigned long key, unsigned long mask)
{
	// Save the color key and mask
	color_key = key;
	color_mask = mask;

	if (gfx_overlay)
		gfx_set_video_color_key(key, mask, gfx_get_video_color_key_src());

	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Configures how video overlay on graphics will appear.
//
// Parameters:
//   video_overlay_mode:
//     VL_VIDEO_OVERLAY_COLOR_KEY  : 
//       Video overlay will appear wherever graphics pixels match the
//       color key comparison.
//     VL_VIDEO_OVERLAY_CHROMA_KEY :
//       Video overlay will appear wherever video pixels do NOT match
//       the color key comparison.
//     VL_VIDEO_OVERLAY_COLOR_KEY_INSIDE_ALPHA_AND_VIDEO_OUTSIDE : 
//       Inside alpha windows, video will appear if graphics pixels
//       match the color key comparison; outside the (enabled) alpha
//       windows, video will always appear.
//     VL_VIDEO_OVERLAY_CHROMA_KEY_INSIDE_ALPHA_AND_GRAPHICS_OUTSIDE :
//       Inside alpha windows, video will appear if video pixels do NOT match
//       the color key comparison; outside the (enabled) alpha windows, video
//       will not appear.
//     VL_VIDEO_OVERLAY_ALWAYS_ON:
//       Disables color key comparison, meaning only video is displayed
//       in the video window area.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS VL_VideoSetOverlay(VL_VideoOverlayModeType video_overlay_mode)
{
	// Save the new value
	if (video_overlay_mode == VL_VIDEO_OVERLAY_ALWAYS_ON)
		gfx_overlay = FALSE;
	else		
		gfx_overlay = TRUE;

	switch (video_overlay_mode) {
		case VL_VIDEO_OVERLAY_ALWAYS_ON:
			// Disable color key comparison - set color mask to zero
			// and color key source to graphics
			gfx_set_video_color_key(0, 0, 1);
			break;			
		case VL_VIDEO_OVERLAY_COLOR_KEY:
			// Enable color key comparison (use the saved color key and mask)
			gfx_set_video_color_key(color_key, color_mask, 1);
			gfx_set_no_ck_outside_alpha(0);
			break;			
		case VL_VIDEO_OVERLAY_CHROMA_KEY:
			// Enable color key comparison (use the saved color key and mask)
			gfx_set_video_color_key(color_key, color_mask, 0);
			gfx_set_no_ck_outside_alpha(0);
			break;
		case VL_VIDEO_OVERLAY_COLOR_KEY_INSIDE_ALPHA_AND_VIDEO_OUTSIDE:
			// Enable color key comparison (use the saved color key and mask)
			gfx_set_video_color_key(color_key, color_mask, 1);
			gfx_set_no_ck_outside_alpha(1);
			break;			
		case VL_VIDEO_OVERLAY_CHROMA_KEY_INSIDE_ALPHA_AND_GRAPHICS_OUTSIDE:
			// Enable color key comparison (use the saved color key and mask)
			gfx_set_video_color_key(color_key, color_mask, 0);
			gfx_set_no_ck_outside_alpha(1);
			break;
		default:					
			return VL_STATUS_BAD_PARAMETER;
	}
    return VL_STATUS_OK;
}
