/* NSC_LIC_BSD
 *
 * National Semiconductor Corporation Open Source License for
 *
 *    Video library for Geode SCx2x0
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c) 2000-2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use by Licensee of this Software in source and
 * binary forms, with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   * Redistributions in binary form must either:
 *     (a) reproduce the above copyright notice, this list of conditions
 *         and the following disclaimer in the documentation and/or
 *         other materials provided with the distribution, or
 *     (b) if Licensee elects to redistribute any Software not in
 *         accordance with (a), then Licensee shall indemnify National
 *         Semiconductor Corporation and all copyright holders and
 *         contributors for any claims related to such Software (and pay
 *         all costs and damages, including all attorney fees)
 *
 *   * Neither the name of the National Semiconductor Corporation nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF
 * YOUR JURISDICTION. It is Licensee's responsibility to comply with
 * any export regulations applicable in Licensee's jurisdiction. Under
 * CURRENT (July 2001) U.S. export regulations this software
 * is eligible for export from the U.S. and can be downloaded by or
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran,
 * Syria, Sudan, Afghanistan and any other country to which the U.S.
 * has embargoed goods and services.
 *
 * END_NSC_LIC_BSD */
 
/*--------------------------------------------------------------------------
 * Copyright (c) 2000-2001 by National Semiconductor Corporation
 * All rights reserved.
 *--------------------------------------------------------------------------
 * File Contents:
 *           Functions for handling VBI capture and CC output for the Geode
 *           SCx2x0
 * Project : Video library for Geode SCx2x0
 * Created : November, 2000
 * Author  : Galit Heller & Eyal Cohen
 *--------------------------------------------------------------------------
 * $Date: 8/30/01 7:48a $
 * $RCSfile: vl_vbi.c $
 * $ProjectRevision: 3.1 $
 * $Revision: 8 $
 *------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/*-------------------------   Include files   ----------------------------*/
/*------------------------------------------------------------------------*/
#include "gfx_rtns.h"
#include "gfx_defs.h"
#include "gfx_regs.h"
#include "vl_rtns.h"
#include "vl_os.h"

/*------------------------------------------------------------------------*/
/*----------------   Constants and macros definitions  -------------------*/
/*------------------------------------------------------------------------*/
#undef VL_VBI_DEBUG
#define CC_BUF_SIZE 20

#if !defined(NULL)
#define NULL 0
#endif

/*------------------------------------------------------------------------*/
/*------------   Data Types definitions and declarations   ---------------*/
/*------------------------------------------------------------------------*/

// Structure for keeping information of the VBI settings
typedef struct VbiStateStruct {
	VL_RING_ID ccRingId;
	unsigned long virt_odd_base;
	unsigned long phys_odd_base;
	unsigned long virt_even_base;
	unsigned long phys_even_base;
	int cc_line_num;
	int vip_last_field;
	int tvenc_last_field;
	unsigned char cc_chars[2];
	boolean cc_exist;
	boolean vbi_capture_enabled;
	boolean cc_passthru_enabled;
} VbiStateStruct;

/*------------------------------------------------------------------------*/
/*-------------   Variable definitions and declarations   ----------------*/
/*------------------------------------------------------------------------*/
VL_VbiStruct VbiAttr;
VbiStateStruct VbiState;
int ic;

/*-------------*/
/*-- Externs --*/
/*-------------*/

/**************************************************************************/
/****************************   VBI module  *******************************/
/**************************************************************************/

/*
	Use the following sequence to start VBI capture :
	1. Configure the video decoder to output the requested VBI lines and format
	   by calling the function gfx_set_decoder_vbi_format(start, end, format)
	   as many times as needed.
	2. Call VL_VbiCapture() to start capture.
	   The function will call VL_VbiGetEmptyBuf() twice to get the first
	   buffers for the odd and even fields and configure VIP's VBI bases.
	   Each buffer should be big enough to contain one field of VBI data
	   (both raw and sliced).
	   Buffers will be sent to the application using the two supplied
	   functions as follows :
	   - VIP interrupt handler is called.
	   - The function VL_VbiNotifyFullBuf() is called to notify the
		 application that a buffer filled with VBI data is ready. The address
		 passed to the application is the virtual address of the buffer.
	   - The function VL_VbiGetEmptyBuf() is called to get another empty
		 buffer from the application. The application should return the physical
		 address of a 16-byte aligned buffer and the matching virtual address.
		 The new address is programmed in the relevant VIP's VBI base address.
	3. Call VL_VbiCCPassThruEnable() to enable/disable CC passing thru to the TV.
*/

/*
	VIP field interrupt - triggered at the end of every field.
    The interrupt handler is responsible for the following things :
    - Save the new vbi data to a buffer retrieved from the application
      and notify the application that it's ready for use.
    - Extract CC (closed caption) information from the vbi buffer and
      insert it to a ring buffer.
*/

#ifdef VL_VBI_DEBUG

typedef struct {
	unsigned long time;
	unsigned long data;
} vbi_debug_struct;

#define VBI_BUF_SIZE 100
vbi_debug_struct vbi_buf[VBI_BUF_SIZE];
int vbi_buf_counter = 0;
int vbi_cur_tick, vbi_prev_tick = 0;

#endif

static void vbiIntHandler()
{
	unsigned long value, status;
	unsigned long vbi_base;
	int vip_current_field;
	int i;
	int numChars;

	value = READ_VIP32(SC1200_VIP_CONTROL);
	// Return if video and VBI capture is not enabled
	if (((value & SC1200_CAPTURE_RUN_MODE_MASK) != SC1200_CAPTURE_RUN_MODE_START) ||
	    ((value & (SC1200_VIP_DATA_CAPTURE_EN | SC1200_VIP_VBI_CAPTURE_EN)) == 0))
	{
		// Clear pending field interrupt
		WRITE_VIP32(SC1200_VIP_STATUS, SC1200_VIP_CLEAR_FIELD_INT);
#ifdef VL_VBI_DEBUG
		OS_IoWriteByte(0x80, 0xEC);
#endif
		return;
	}

	ic++;

#ifdef VL_VBI_DEBUG
	OS_IoWriteByte(0x80, ic);
#endif

#ifdef VL_VBI_DEBUG
	if (vbi_buf_counter < VBI_BUF_SIZE) {
		vbi_cur_tick = tickGet();
		vbi_buf[vbi_buf_counter].time = vbi_cur_tick - vbi_prev_tick;
		vbi_buf[vbi_buf_counter].data = READ_VIP32(SC1200_VIP_STATUS);
		vbi_prev_tick = vbi_cur_tick;
		vbi_buf_counter++;
	}
#endif

	VbiState.cc_exist = 0;

	i = 0;
	while (1) {
		status = READ_VIP32(SC1200_VIP_STATUS);
		if (status & (SC1200_VIDEO_DATA_CAPTURE_ACTIVE | SC1200_VBI_DATA_CAPTURE_ACTIVE))
			break;
		if (i++ >= 3000) {
			// Clear pending field interrupt
			WRITE_VIP32(SC1200_VIP_STATUS, SC1200_VIP_CLEAR_FIELD_INT);
#ifdef VL_VBI_DEBUG
			OS_IoWriteByte(0x80, 0xCE);
#endif
			return;
		}
	}

	// Clear pending field interrupt
//	WRITE_VIP32(SC1200_VIP_STATUS, SC1200_VIP_CLEAR_FIELD_INT);

	vip_current_field = status & SC1200_VIP_CURRENT_FIELD_ODD;

#ifdef VL_VBI_DEBUG
	if (vip_current_field == VbiState.vip_last_field)
		logMsg("ERROR - Field not changed !!!\n", 0, 0, 0, 0, 0, 0);
#endif

	// Save the current VIP field
	VbiState.vip_last_field = vip_current_field;

	// Extract CC (closed caption) information from the vbi buffer and
	// insert it to a ring buffer.

	if (VbiState.cc_passthru_enabled && VbiState.ccRingId)
	{
		// Even field is being processed - odd field is ready
		if (vip_current_field == EVEN_FIELD)
		{
			vbi_base = VbiState.virt_odd_base;
			vbi_base += (VbiState.cc_line_num * gfx_get_vbi_pitch());
			value = (*(unsigned long *)vbi_base);
			// Dxx2xx - line 21, x5xxxx - US CC, xxCxxx - FID, xxxx4B - Dword count.
			if ((value & 0xFFFFFF00) == 0xD5C24B00)
			{
				value = (*(unsigned long *)(vbi_base + 4));
				value &= 0x00007F7F;
				if (value != 0)
				{
					VbiState.cc_chars[0] = (unsigned char) value;
					VbiState.cc_chars[1] = (unsigned char) (value >> 8);
					VbiState.cc_exist = 1;
					numChars = OS_RingBufPut(VbiState.ccRingId, (char *)VbiState.cc_chars, 2);
#ifdef VL_VBI_DEBUG
					if (numChars != 2)
						logMsg("ERROR - Ring buffer is full [Tick = %ld] !!!\n", tickGet(), 0, 0, 0, 0, 0);
#endif
				}
			}
		}
	}

	// Notify the upper layer that a buffer filled with VBI data is ready
	// and, get an empty buffer to capture the next VBI field
	if (vip_current_field == EVEN_FIELD) {		// Even field is being processed ?
		// Notify that odd field is full
		VbiAttr.VL_VbiNotifyFullBuf(VbiState.virt_odd_base, ODD_FIELD);
		// Get buf for next odd field
		VbiAttr.VL_VbiGetEmptyBuf(&VbiState.phys_odd_base, &VbiState.virt_odd_base, ODD_FIELD);
		// Set new odd base in VIP
		gfx_set_vbi_base(0, VbiState.phys_odd_base);
	}
	else {
		// Notify that even field is full
		VbiAttr.VL_VbiNotifyFullBuf(VbiState.virt_even_base, EVEN_FIELD);
		// Get buf for next even field
		VbiAttr.VL_VbiGetEmptyBuf(&VbiState.phys_even_base, &VbiState.virt_even_base, EVEN_FIELD);
		// Set new even base in VIP
		gfx_set_vbi_base(VbiState.phys_even_base, 0);
	}
	// Clear pending field interrupt
	/* <sojan> Moved this from the top </sojan> */
	WRITE_VIP32(SC1200_VIP_STATUS, SC1200_VIP_CLEAR_FIELD_INT);
}

//---------------------------------------------------------------------
// Description:
//   Extract CC data from the internal ring buffer and output it to
//   the TV (by programming the TV encoder).
//
// Parameters:
//   None.
//
// Return Value:
//   None.
//---------------------------------------------------------------------
void VidSendCCDataToTVenc()
{
	unsigned char cc_data[2];
	unsigned long odd_field;

	if (!VbiState.cc_passthru_enabled)
		return;

	odd_field = gfx_test_tvenc_odd_field();

	if (VbiState.tvenc_last_field == odd_field)
		return;

	// Save the current TV encoder field
	VbiState.tvenc_last_field = odd_field;

	// Continue only if even field in TV encoder
	if (VbiState.tvenc_last_field == ODD_FIELD)
		return;

	if (VbiState.ccRingId != NULL) {
		if (OS_RingBufGet(VbiState.ccRingId, (char *)cc_data, 2) == 2)
			gfx_set_tv_cc_data(cc_data[0], cc_data[1]);
	}
}

//---------------------------------------------------------------------
// Description:
//   Clear CC data from TV.
//
// Parameters:
//   None.
//
// Return Value:
//   None.
//---------------------------------------------------------------------
void clear_cc1_data_from_tv()
{
    char cc_data[2];

    if (VbiState.ccRingId)
    {
        // Empty the CC ring buffer
        while (OS_RingBufGet(VbiState.ccRingId, (char *)cc_data, 2) == 2);

        // Erase Non-Displayed Memory
        cc_data[0] = 0x14;
        cc_data[1] = 0x2e;
        OS_RingBufPut(VbiState.ccRingId, cc_data, 2);
        OS_RingBufPut(VbiState.ccRingId, cc_data, 2);

		// Erase Displayed Memory
        cc_data[0] = 0x14;
        cc_data[1] = 0x2c;
        OS_RingBufPut(VbiState.ccRingId, cc_data, 2);
        OS_RingBufPut(VbiState.ccRingId, cc_data, 2);

		// End of Caption
        cc_data[0] = 0x14;
        cc_data[1] = 0x2f;
        OS_RingBufPut(VbiState.ccRingId, cc_data, 2);
        OS_RingBufPut(VbiState.ccRingId, cc_data, 2);
    }
}

//---------------------------------------------------------------------
// Description:
//   Configure VIP interrupt.
//
// Parameters:
//   None.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS config_vip_interrupt()
{
	unsigned long pmr;
	unsigned char reg_val;
	unsigned char int_val;
	unsigned char int_mask;
	unsigned short int_reg;

	if((VbiAttr.VipIRQ ==  0) ||
       (VbiAttr.VipIRQ ==  2) ||
       (VbiAttr.VipIRQ ==  8) ||
       (VbiAttr.VipIRQ == 13) ||
       (VbiAttr.VipIRQ >  15))
		return VL_STATUS_BAD_PARAMETER;

	// Set the requested IRQ to LEVEL
	if (VbiAttr.VipIRQ <= 7)
	{
		int_reg = 0x04d0;
		int_val = VbiAttr.VipIRQ;
	}
	else
	{
		int_reg = 0x04d1;
		int_val = VbiAttr.VipIRQ - 8;
	}

	reg_val = OS_IoReadByte(int_reg);
	reg_val |= (1 << int_val);
	OS_IoWriteByte(int_reg, reg_val);

	// Route the requested PCI interrupt to the requested IRQ

	switch (VbiAttr.VipINT)        // INTA = 1, INTB = 2, INTC = 3, INTD = 4
	{
		case 1:
			int_reg = 0x5c;
			int_mask = 0xf0;
			int_val = VbiAttr.VipIRQ;
			break;
		case 2:
			int_reg = 0x5c;
			int_mask = 0x0f;
			int_val = VbiAttr.VipIRQ << 4;
			break;
		case 3:
			int_reg = 0x5d;
			int_mask = 0xf0;
			int_val = VbiAttr.VipIRQ;
			break;
		case 4:
			int_reg = 0x5d;
			int_mask = 0x0f;
			int_val = VbiAttr.VipIRQ << 4;
			break;
		default:
			return VL_STATUS_BAD_PARAMETER;
	}

	OS_PciReadByte(0, 0x12, 0, int_reg, &reg_val);
	reg_val &= int_mask;
	reg_val |= int_val;
	OS_PciWriteByte(0, 0x12, 0, int_reg, reg_val);

	// Config VIP Interrupt Pin Register to use INTC
	OS_PciWriteByte(0, 0x12, 4, 0x3d, VbiAttr.VipINT);

	// If INTC is used, configure it to be internal only
	if (VbiAttr.VipINT == 3) {
		pmr = OS_IoReadDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR);
		pmr &= ~0x10;
		OS_IoWriteDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR, pmr);
	}

	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//	 Initializes the VBI path by configuring VIP module,
//	 according to the values specified.
//
//   This function must be called prior to using any of the other
//   VBI functions.
//
//   For the usage of VBI capture, the application needs to supply
//   two functions to manipulate VBI buffers (for more details refer to
//   the Video Library Reference Manual section "Application layer").
//
//   The function initializes the following :
//   * Configure VIP module :
//     - Set Task A
//     - Disable VBI interrupt
//     - Disable capture VBI data
//     - Configure VBI pitch
//   * Configure and connect VIP interrupt.
//   * Initialize a ring buffer for CC handling.
//   * Disable CC passing thru to the TV.
//   * Save the two function pointers for use during VBI capture.
//
// Parameters:
//     vbi_struct fields :
//     VipIRQ              - The IRQ level for the VBI interrupt.
//     VipINT              - The PCI interrupt for the VBI interrupt.
//     VL_VbiGetEmptyBuf   - Function to get the next VBI buffer.
//     VL_VbiNotifyFullBuf - Function to notify on a full VBI buffer.
//     VbiLineSize         - VBI line size (in bytes).
//
//   When VL_VideoInit() returns, the VBI is not enabled.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS VL_VbiInit(VL_VbiStruct * vbi_struct)
{
	VbiAttr = *vbi_struct;

    // Configure VIP for VBI capture
	VbiState.virt_odd_base  = 0;
	VbiState.phys_odd_base  = 0;
	VbiState.virt_even_base = 0;
	VbiState.phys_even_base = 0;
	gfx_set_vbi_base(VbiState.phys_even_base, VbiState.phys_odd_base);
	gfx_set_vbi_pitch(VbiAttr.VbiLineSize);
	gfx_set_vbi_mode(VBI_TASK_A);

	// Disable VBI Field interrupt
	gfx_set_vbi_interrupt(0);

	// Disable capture VBI data
	gfx_set_vbi_enable(0);

	// Initialize the video processor to get VBI data from memory
	gfx_set_vbi_source(VBI_SOURCE_MEMORY);

	// Configure VIP interrupt
	if (config_vip_interrupt() != VL_STATUS_OK)
		return VL_STATUS_ERROR;

	// Connect the interrupt service routine
	if (OS_InterruptConnect(VbiAttr.VipIRQ, (void (*)(int))vbiIntHandler, "VL_Vbi", 0) != VL_STATUS_OK)
		return VL_STATUS_ERROR;

	VbiState.ccRingId = OS_RingCreate(CC_BUF_SIZE);
	if (VbiState.ccRingId == NULL)
		return VL_STATUS_ERROR;

	VbiState.cc_line_num = 0;
	VbiState.cc_exist = 0;
	VbiState.vbi_capture_enabled = 0;
	VbiState.cc_passthru_enabled = 0;
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Shuts down the VBI:
//   * Disables VBI capture.
//   * Disables VBI interrupt.
//	 * Releases all internally-allocated resources.
//
// Parameters:
//   None.
//
// Return Value:
//   VL_STATUS_OK.
//
// Side Effects:
//   None of the VBI functions may be called after calling this function,
//   except for VL_VbiInit().
//---------------------------------------------------------------------
VL_STATUS VL_VbiEnd()
{
	// Disable VBI capture
	VL_VbiCapture(0);

	// Free the CC ring buffer
	if (VbiState.ccRingId != NULL)
		OS_RingDelete(VbiState.ccRingId);
	VbiState.ccRingId = NULL;

	// Disconnect the VBI interrupt handler
	OS_InterruptDisconnect(VbiAttr.VipIRQ);
    return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Enable/Disable VBI capture to memory :
//   * Get the first VBI buffers to capture odd/even VBI data.
//   * Enable VIP interrupt.
//   * Enable capture VBI data.
//
// Parameters:
//   enable - TRUE to enable VBI capture, FALSE to disable VBI capture.
//
// Return Value:
//   VL_STATUS_OK.
//---------------------------------------------------------------------
VL_STATUS VL_VbiCapture(boolean enable)
{
	ic = 0;
	VbiState.vip_last_field = VbiState.tvenc_last_field = 0;

	if (VbiState.vbi_capture_enabled == enable)
		return VL_STATUS_OK;

	if (enable)
	{
		// Get the first 2 VBI buffers to capture odd/even VBI data
		VbiAttr.VL_VbiGetEmptyBuf(&VbiState.phys_odd_base, &VbiState.virt_odd_base, 1);
		VbiAttr.VL_VbiGetEmptyBuf(&VbiState.phys_even_base, &VbiState.virt_even_base, 0);
		gfx_set_vbi_base(VbiState.phys_even_base, VbiState.phys_odd_base);
		// Enable capture VBI data
		gfx_set_vbi_enable(1);
		// Enable VBI Field interrupt
		gfx_set_vbi_interrupt(1);
	}
	else
	{
		// Disable VBI Field interrupt
		gfx_set_vbi_interrupt(0);
		// Disable capture VBI data
		gfx_set_vbi_enable(0);
	}
	VbiState.vbi_capture_enabled = enable;
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Checks whether VBI is captured.
//
// Parameters:
//   None.
//
// Return Value:
//    TRUE if VBI is currently captured, FALSE otherwise.
//---------------------------------------------------------------------
boolean VL_VbiIsCaptured()
{
	return VbiState.vbi_capture_enabled;
}

//---------------------------------------------------------------------
// Description:
//   Enables or disables CC passing thru to the TV.
//   CC passing thru to the TV is available only if VBI capture is enabled.
//
// Parameters:
//   OnOff - On to enable CC passing thru to the TV, Off to disable.
//   cc_line_num - The line number where CC information is located in memory,
//                 relative to the beginning of the field.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS VL_VbiCCPassThruEnable(boolean OnOff, int cc_line_num)
{
	// CC PassThru is available only if VBI capture is enabled
	if (OnOff)
		if (VbiState.vbi_capture_enabled == 0)
			return VL_STATUS_ERROR;

	// Save CC line number
	VbiState.cc_line_num = cc_line_num;

	if (OnOff) {
		gfx_set_tv_cc_enable(1);
		VbiState.cc_passthru_enabled = 1;
	}
	else {
		VbiState.cc_passthru_enabled = 0;
		clear_cc1_data_from_tv();
		gfx_set_tv_cc_enable(0);
	}
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Checks whether CC passing thru to the TV is enabled.
//
// Parameters:
//   None.
//
// Return Value:
//    TRUE if CC passing thru to the TV is enabled, FALSE otherwise.
//---------------------------------------------------------------------
boolean VL_VbiCCPassThruIsEnabled()
{
	return VbiState.cc_passthru_enabled;
}

//---------------------------------------------------------------------
// Description:
//   Get CC data of the last field.
//
// Parameters:
//   ch1, ch2 - pointers to two chars to be filled with CC data.
//
// Return Value:
//    TRUE if CC data exists, FALSE otherwise.
//---------------------------------------------------------------------
boolean VL_VbiGetCCData(unsigned char * ch1, unsigned char * ch2)
{
	if (VbiState.cc_exist) {
		*ch1 = VbiState.cc_chars[0];
		*ch2 = VbiState.cc_chars[1];
	}
	return VbiState.cc_exist;
}
