/*
 *
 * National Semiconductor Corporation Open Source License for 
 *
 *    Geode Video4linux2 driver
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c)  2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 *   * Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer. 
 *
 *   * Redistributions in binary form must reproduce the above 
 *     copyright notice, this list of conditions and the following 
 *     disclaimer in the documentation and/or other materials provided 
 *     with the distribution. 
 *
 *   * Neither the name of the National Semiconductor Corporation nor 
 *     the names of its contributors may be used to endorse or promote 
 *     products derived from this software without specific prior 
 *     written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER 
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE,
 * INTELLECTUAL PROPERTY INFRINGEMENT, OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF 
 * YOUR JURISDICTION. It is licensee's responsibility to comply with 
 * any export regulations applicable in licensee's jurisdiction. Under 
 * CURRENT (2001) U.S. export regulations this software 
 * is eligible for export from the U.S. and can be downloaded by or 
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed 
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran, 
 * Syria, Sudan, Afghanistan and any other country to which the U.S. 
 * has embargoed goods and services. 
 *
 */
 
 /* File Name            : geodedrv.c
 *
 * File Contents:
 *     Implementation for the /dev/video and /dev/vout devices for the
 *     GeodeV4l2 Linux driver.
 * [Notes]
 *	Ensure that Video4Linux2 has been installed on the system before
 *	proceeding to compile this code. Red-Hat Linux versions upto 7.0
 *	come bundled with Video4Linux 1 which will not work for this code.
 *
 * Project              : Geode Video4Linux2 driver
 * Author               : Sojan James
 * Date:  		: 
 *---------------------------------------------------------------------*/

#include "geodedrv.h"
#include "vbi.h"
#define GEODE_V4L2_VERSION "1.1.4d"

#define FRAME_BUFFER_START        0x40800000
#define FRAME_BUFFER_LINE_SIZE    2048
#define TOTAL_GRAPHICS_MEMORY     0x400000 /* 4MB memory */
/* The Video buffer starts at 2MB offset  (default)*/
#define LINES_IN_NTSC             1024
#define LINES_IN_PAL              1024
#define LINES_IN_OTHERS		  1024  
/* assume a max resolution of 800x600 */
#define FRAME_BUFFER_SIZE(tvStandard)   (FRAME_BUFFER_LINE_SIZE * LINES_IN_##tvStandard)
/* Set the video buffer to start 4 lines after Frame buffer end */
#define VIDEO_BUFFER_START(tvStandard)  (FRAME_BUFFER_START + FRAME_BUFFER_SIZE(tvStandard) )

/* IRQ to be used by timer interrupt */
static int IRQ = 8;
/* IRQ to be used by the VBI interrupt */
int VBI_IRQ = 11 ;

/* The Standard of the video input 1->NTSC, 2->PAL*/
int VIN_STD = VSOURCE_STD_NTSC;
/* The Standard of video out 1->NTSC, 2->PAL */
int VOUT_STD = VSOURCE_STD_NTSC;

/* Show display on CRT (RGB mixing, no DVIP) */
static int USE_CRT = TRUE;

/* Do go into DVIP */
static int NO_DVIP = FALSE;

/* Default de-interlacing mode */
/* 1 = NONE, 2 = INTERLEAVE, 3 = BOB */
static int DEINTERLACE 	= 3;

/* TV output */
/* 1-> Composite, 2-> S-Video, 3-> YUV, 4 -> SCART */
static int TV_OUT       = 1;

/* TV aspect ratio */
/* 1-> graphics only
   2-> graphics and video
   3-> video only (default)
 */
static int TV_ASPECT_RATIO = 3;


/* the minor numbers of the devices*/
static int 	VOUT_MINOR  = 16;
static int 	VIDEO_MINOR = 0;
int		VBI_MINOR   = 224;

int		MAX_VBI_STREAM_BUFFERS	= 12;
int		MAX_VBI_OPEN_BUFFERS	= 10;

/* The NTSC and PAL addresses depend on the current standard the framebuffer 
   is currently in */ 
static int NTSC_START_ADDRESS = VIDEO_BUFFER_START(NTSC);
static int PAL_START_ADDRESS = VIDEO_BUFFER_START(PAL);
static int OTHERS_START_ADDRESS = VIDEO_BUFFER_START(OTHERS);

#ifdef NO_DECODER
#define DECODERSTR  "No Decoder\n"
#define TUNERSTR "No Tuner\n"
#else
#define DECODERSTR  "SAA7114 decoder\n"
#ifndef EXCLUDE_TUNER
#ifdef CYGNUS
#ifdef CYGNUS_PAL_TUNER
#define TUNERSTR "Cygnus PAL Tuner\n"
#else  
#define TUNERSTR "Cygnus NTSC Tuner\n"
#endif
#else
#define TUNERSTR "Unknown Tuner\n"
#endif
#else
#define TUNERSTR "No Tuner\n"
#endif
#endif
#ifdef USE_SYSTEM_I2C
#define I2CSTR  "Linux I2C Subsystem required\n"
#else
#define I2CSTR  "Using built-in I2C module\n"
#endif
#define DESC_STR  "Geode V4l2 driver, Ver "GEODE_V4L2_VERSION", National Semiconductors\n" TUNERSTR DECODERSTR I2CSTR

/* The module parameters, settable at load time */
MODULE_PARM(IRQ,"i");
MODULE_PARM(VBI_IRQ,"i");
MODULE_PARM(VIN_STD,"i");
MODULE_PARM(VOUT_STD,"i");
MODULE_PARM(TV_OUT,"i");
MODULE_PARM(TV_ASPECT_RATIO,"i");
MODULE_PARM(USE_CRT,"i");
MODULE_PARM(NO_DVIP,"i");
MODULE_PARM(DEINTERLACE,"i");
MODULE_PARM(VOUT_MINOR,"i");
MODULE_PARM(VIDEO_MINOR,"i");
MODULE_PARM(VBI_MINOR,"i");
MODULE_PARM(MAX_VBI_STREAM_BUFFERS,"i");
MODULE_PARM(MAX_VBI_OPEN_BUFFERS,"i");
MODULE_AUTHOR("Sojan James");
MODULE_DESCRIPTION(DESC_STR);

/* This is the timer that is going to query the decoder for a signal lock every
   second */
static struct   timer_list geode_timer;
extern void     signalISR(); /* This is the VideoLib signal monitor */
static int      init_videolib();
static void     vout_get_hw_parms(struct vout_device * dev);


/* forward declarations */
void       OS_Interface_init (void);
void       OS_Interface_close (void);
int 	   vbi_decode_cleanup(void);
static int get_genlock(int input, int vout_std);

/* The global instances. vbi_dev and capture_dev is not static because they are
 * accessed from vbi.c 
 */
static struct vout_device	vout_dev;
struct capture_device		capture_dev;
extern struct vbi_device	vbi_dev;

static VL_VideoStruct   g_vs;
static VL_MemBuf        g_video_buf;
int                     g_GFX_pitch;
int                     g_vip_disabled_by_timer = FALSE;
static int              g_VideoLib_initialized = FALSE;

static unsigned long    flags;
static int              g_overlay_mode;
static int              g_overlay_set_in_s_win = TRUE;
int                     g_DVIP_if_possible = FALSE;
int                     g_inDVIP = FALSE;		    
static int              g_timer_activated = 0;
				/* is the timer already activated*/
static int 		g_input_window_index;


/* For DVIP, we need to FLIP the FID in the decoder,
   until we find out a better way to do things, we use
   this */

#ifdef FLIP_DEC_FID_FOR_DVIP
static int              g_decoder_fid_flip_hack = FALSE;
#endif


static struct v4l2_queryctrl capture_control[] =
{ 
 {V4L2_CID_BRIGHTNESS, "Brightness", 0, 0xFF, 1,0x40<<1,V4L2_CTRL_TYPE_INTEGER},
 {V4L2_CID_CONTRAST,   "Contrast",   0, 0xFF, 1,0x44<<1,V4L2_CTRL_TYPE_INTEGER},
 {V4L2_CID_SATURATION, "Saturation", 0, 0xFF, 1,0x40<<1,V4L2_CTRL_TYPE_INTEGER},
 {V4L2_CID_HUE,        "Hue",        0, 0xFF, 1,0x00<<1,V4L2_CTRL_TYPE_INTEGER},
 {V4L2_CID_WHITENESS,  "Sharpness",  0, 0x0F, 1, 0x05, V4L2_CTRL_TYPE_INTEGER},
};
/* Note: (above) Whiteness is actually "Sharpness"  */

static struct v4l2_fmtdesc capture_formats[] = 
{
	{index:0,description:"Geode Capture UYVY", pixelformat:V4L2_PIX_FMT_UYVY,
	 flags:V4L2_FMT_CS_601YUV , depth: 16 }
};
static struct v4l2_fmtdesc vout_formats[] = 
{
	{index:0,description:"Geode UYVY", pixelformat:V4L2_PIX_FMT_UYVY,
	 flags:V4L2_FMT_CS_601YUV , depth: 16 },
	{index:1,description:"Geode RGB", pixelformat:V4L2_PIX_FMT_RGB565,
	 flags:0 , depth: 16 }
};

static struct v4l2_fxdesc vout_effects[] = 
{
	{index:0, name: "Video Overlay Mode",flags:0,inputs:1,controls:1}
};

static struct v4l2_queryctrl vout_effects_control[] = 
{
	{GEODE_CID_OVERLAY_MODE, "Overlay Mode", 1,5,1,5, V4L2_CTRL_TYPE_INTEGER}
};


/*----------------------------------------------------------------------
 * Description:
 *      Find the index into the capture_control array for the specified V4L2_CID_x
 * Return:
 * 	    (integer) Index into capture_control array
 * side-effects:
 * Notes:
 *	This has been lifted from the sample driver distributed along with 
 *	the V4L2 package.
 *----------------------------------------------------------------------*/
static int find_vctrl( int id )
{
	int i;
	if( ( id == V4L2_CID_PRIVATE_BASE ) 
		|| ( id <  V4L2_CID_BASE )
		|| ( id >  V4L2_CID_LASTP1 ) ) {
		return -EDOM;
	}
	for( i = MAXCONTROLS - 1; i >= 0; i-- ) {
		if( capture_control[i].id == id ) {
			break;
		}
	}
	if( i < 0 ) {
		i = -EINVAL;
	}
	return i;
}

/*---------------------------------------------------------------------
 * Description:
 *   Function to check if the decoder is stable.
 * return:
 *     TRUE for signal lock,
 *     FALSE for no signal lock.
 * side-effects:
 * Notes:
 *	This is currently not used in the driver. This function is much 
 * 	faster than decoder_detect_signal as it does not perform an i2c 
 *	read. It just returns the state of the decoder during the previous
 *	read by the timer handler. 
 *----------------------------------------------------------------------*/
int signal_ok()  
{
	return capture_dev.videc.decoder_is_stable;
}

/*---------------------------------------------------------------------
 * Description:
 *	Check if the decoder detects a signal.
 * return:
 *	TRUE for signal lock
 *	FALSE for no signal lock
 * side-effects:
 * Notes:
 *	This function performs an I2C read from the decoder and locks the 
 *	system up for the entire duration of the call. This function should
 *	not be used in any situation where it is called very often. Doing so
 *	will slow the system down.
 *----------------------------------------------------------------------*/
int decoder_detect_signal()
{
#ifndef NO_DECODER
	unsigned char status;
	save_flags(flags);
	cli();
	status = tv_i2c_saa7114_read_byte_data(0x1F); /* status byte */
	restore_flags(flags);
	if( status &  0x01) /* check for color signal */
		return TRUE;
	else
		return FALSE;	
#else
	return FALSE;
#endif		
}

/*---------------------------------------------------------------------
 * Description:
 * 	Fill the video buffer with blue 
 * return:
 *  void	
 * side-effects:
 * Notes:
 *	This function is used to give a nice looking blue screen when the
 *	decoder loses the signal.
 *----------------------------------------------------------------------*/
void blue_screen()
{
	unsigned long val = 0x1C641CFF; /* RGB: 0,0,255 in YUV 4:2:2 */
	int i,j;
	int pitch = gfx_get_vip_pitch();
	int width,height;
	unsigned long ptr = gfx_get_vip_base(1) ;
	unsigned long p = ptr;		 
    
	if(ptr < (unsigned long)gfx_phys_fbptr) return ;
	ptr = ptr - (unsigned long)gfx_phys_fbptr;
	width = g_video_buf.width / 2;
	height = g_video_buf.height;
	for(i = 0; i < height ; i++) {
		p = ptr;
		for(j = 0 ; j < width; j++) {
			WRITE_FB32(p,val);
			p+=4;
		}
		ptr+=pitch;
	} 
	/* sometimes, we get 0 from gfx_get_vip_base() */
	/* (The first time actually) */
	ptr = gfx_get_vip_base(0) ;
	if(ptr < (unsigned long)gfx_phys_fbptr) return;
	ptr  = ptr - (unsigned long)gfx_phys_fbptr;
	for(i = 0; i < height ; i++) {
		p  = ptr;
		for(j = 0 ; j < width; j++) {
			WRITE_FB32(ptr,val);
			p+=4;
		}
		ptr+=pitch;
	}
}
 
#ifdef FLIP_DEC_FID_FOR_DVIP
static void flip_decoder_fid(int flip)
{
        unsigned char byte;
        byte = tv_i2c_saa7114_read_byte_data(0xC2);
        if(flip) {
                byte = byte | 0x40;
                g_decoder_fid_flip_hack = TRUE;
        } else {
                byte = byte & (~(0x40));
                g_decoder_fid_flip_hack = FALSE;
        }
        tv_i2c_saa7114_write_byte_data(0xC2,byte);
}
#endif


/*---------------------------------------------------------------------
 * Description:
 *	Timer loop to check for a valid input signal and switch to and
 *	from DVIP if the user requires so, and conditions permit.
 * return:
 *  void
 * side-effects:
 * Notes:
 *	The VidSetDirectVIP call changes the timings using durango functions
 *	This disables the alpha windows. We should restore them.
 *----------------------------------------------------------------------*/
#ifndef NO_DECODER
 void timer_handler(unsigned long val)
 {
	/* DEBUG printk("VID 400:%x\n",READ_VID32(0x400)); */
        // check if we are in DVIP, we need to flip
        // the decoder FID if we are in DVIP.
#ifdef FLIP_DEC_FID_FOR_DVIP
        unsigned long dvip_status = READ_VID32(0x400);

        if(dvip_status & 0x02) {
                // we are in DVIP, VideoLib might have drifted to DVIP.
                if(!g_decoder_fid_flip_hack) {
                        // not flipped, do it.
                        flip_decoder_fid(dvip_status);
                }
        }
#endif
	/* check if we have a signal lock */		 
	if(decoder_detect_signal() || (capture_dev.input == VSOURCE_MPEG_DEC)) {
		capture_dev.videc.decoder_is_stable	= TRUE;
		if(g_vip_disabled_by_timer) {
	    	        g_vip_disabled_by_timer = FALSE;	
		        if(VL_VideoIsPaused()) VL_VideoResume();
                }        
		if(!g_inDVIP && g_DVIP_if_possible && (!NO_DVIP) && (!USE_CRT)) 
		{
			VidSetDirectVIP(1,0);
			g_inDVIP = TRUE;	
		}
	}	
	else { /* No signal */
		/* Lets confirm it once more */
		if(!decoder_detect_signal()) { /* we've definitely lost it */
			VL_VideoPause();
			g_vip_disabled_by_timer = TRUE;
			blue_screen();
			capture_dev.videc.decoder_is_stable	= FALSE;
			signalISR(); /*Tell VideoLib that we lost the signal*/
			if(g_inDVIP && g_DVIP_if_possible)
				g_inDVIP = FALSE;
			debug_msg("timer_handler: lost signal\n");
		}	
	}
	if(g_timer_activated) {
		geode_timer.expires = jiffies + HZ; /* one second*/
		add_timer(&geode_timer);
	}
 }
#endif
/* VOUT Device
 * The output device
 */  


/*---------------------------------------------------------------------
 * Description:
 *	The open function for the vout device
 * return:
 *	Zero on success, error code on failure
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int v4l2_vout_open(struct v4l2_device *v, int o_flags, void **idptr)
{
	struct vout_device *dev = (struct vout_device *)v;
	int     i, n;
        int     cap;
	unsigned long cpu_flags;
	
        for (i = 0, n = -1, cap = 0; i < MAX_OPENS; ++i) {
                if (!dev->open_data[i].isopen)
                        n = i;/* available open_data structure */
                else if (!dev->open_data[i].noncapturing)
                        cap = 1;/* another open is already capturing */
        }
        if (n == -1) {/* No available open_data structures */
                err_msg("No more opens on this device\n");
                return -EBUSY;
        }
        if (o_flags & O_NOIO)/*  Non-capturing open */
                dev->open_data[n].noncapturing = TRUE;
        else if (cap) {
                err_msg("No more capturing opens on this device\n");
	        return -EBUSY;
        }
        else {
                dev->open_data[n].noncapturing = FALSE;
                /*  Keep track of whether there is a capturing open  */
                ++dev->capturing_opens;
        }
        MOD_INC_USE_COUNT;
        ++dev->open_count;
        dev->open_data[n].isopen = TRUE;
        dev->open_data[n].dev = dev;
        *idptr = &dev->open_data[n];
        if (dev->open_count == 1)
	{
#ifndef NO_DECODER                
		save_flags(cpu_flags);
		cli();
		if(!g_timer_activated)
			add_timer(&geode_timer);
		g_timer_activated++;	
		restore_flags(cpu_flags);
#endif                
		debug_msg("First open of vout device\n");
	}
	return 0;
}

/*---------------------------------------------------------------------
 * Description:
 *	The close function for the vout device
 * return:
 *	void	
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static void v4l2_vout_close( void *id )
{
	struct vout_device_open *o = (struct vout_device_open *)id;
	struct vout_device *dev = o->dev;
	if (!o->noncapturing) {
                --dev->capturing_opens;
                debug_msg("Video output device closed\n");
        }
        o->isopen = FALSE;
        --dev->open_count;
        if (dev->open_count == 0) {
#ifdef RESET_ON_CLOSE		
		save_flags(flags);
		cli();
		g_timer_activated--;
		if(!g_timer_activated)
			del_timer(&geode_timer);
		restore_flags(flags);	
		VL_VideoEnable(0);
#endif		
		debug_msg("Last close on vout device\n");
        }
        MOD_DEC_USE_COUNT;
}


/*---------------------------------------------------------------------
 * Description:
 *	The write function for the vout device
 * return:
 *	Error 	
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static long v4l2_vout_write( void *id, const char *buf, unsigned long count, int noblock )
{
	return -EINVAL;
}

/*---------------------------------------------------------------------
 * Description:
 *	Setup the overlay hardware.
 * return:
 *  void	
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static void set_overlay_info( overlay_info_t* overlay_info )
{ 
	extern unsigned char gfx_overlay;
	if(overlay_info->Enable)
        {
                tv_os_dbg_msg("Enabling overlay\n");
                VL_VideoSetColorKey(overlay_info->ColorKey,
                    overlay_info->ColorMask);
		gfx_set_video_color_key(overlay_info->ColorKey,
			overlay_info->ColorMask, overlay_info->Graphics);
		gfx_overlay = overlay_info->Enable;
        }
        else
        {  /* Disable */
                VL_VideoSetColorKey(0,0);
		gfx_set_video_color_key(overlay_info->ColorKey,0,1);
        }

}

/*---------------------------------------------------------------------
 * Description:
 *	Setup the alpha blend parameters.
 * return:
 *  void
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static void set_alpha_blend_info( alpha_blend_info_t* alpha_blend_info )
{
	int enable = FALSE;
	if(alpha_blend_info->Color) enable = TRUE;
	VL_VideoSetAlphaWindow(
    	(unsigned char)alpha_blend_info->WindowNumber,
	    alpha_blend_info->StartX,
    	alpha_blend_info->StartY,
    	alpha_blend_info->Width,
    	alpha_blend_info->Height);

	VL_VideoSetAlphaOperation(
	    (unsigned char) alpha_blend_info->WindowNumber,
    	alpha_blend_info->Alpha,
	    alpha_blend_info->Color, enable,
    	alpha_blend_info->Priority,
    	alpha_blend_info->Delta,
    	alpha_blend_info->Enable);
	debug_msg("ALPHA_BLEND_INFO\n");
}


/*---------------------------------------------------------------------
 * Description:
 *	Sets up the Video window for the output device.
 * return:
 *	Zero
 * side-effects:
 *	May corrupt the video if the decoder is not setup for the same size
 * 	as the window
 * Notes:
 *	Use the VIDIOC_S_FMT ioctl on /dev/video to set the decoder to the
 *	same size window.
 *----------------------------------------------------------------------*/
static int vout_set_window(struct v4l2_window * window)
			
{
//	if(window->x < 0) window->x = 0;
//	if(window->y < 0) window->y = 0;
        // if Video is paused, Video window settings will not have an
        // effect.
        int was_paused;
        was_paused = VL_VideoIsPaused();
        if(was_paused)
                VL_VideoResume();
	VL_VideoSetWindow(&g_video_buf,window->x,window->y,
			window->width,window->height);
        if(was_paused)
        {
                blue_screen();
                VL_VideoPause();
        }

	debug_msg("vout_set_window: %d, %d, %d, %d\n", window->x,
					  window->y,
					  window->width,
					  window->height);
	window->width =  window->width & ~0x7;
        // If the overlay mode is color key, we also set the color key here
#ifndef WEBMEDIA_COMPAT
        if(g_overlay_set_in_s_win) {
                if( (g_overlay_mode == VL_VIDEO_OVERLAY_COLOR_KEY) ||
                (g_overlay_mode == VL_VIDEO_OVERLAY_CHROMA_KEY)    )
		         return VL_VideoSetColorKey(window->chromakey,0x00FFFFFF);
	}
#endif

	return 0;
}

/*---------------------------------------------------------------------
 * Description:
 *	Sets up the de-interlacing mode. This will come into effect only
 *	after the next time the VIDIOC_S_WIN ioctl is called.
 * return:
 *	Zero on success, error code on failure
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int vout_set_deinterlacing_method(int window,int mode)
{
	if( (mode != GEODE_DEINTERLACE_NONE) &&
	    (mode != GEODE_DEINTERLACE_INTERLEAVE) &&
	    (mode != GEODE_DEINTERLACE_BOB)
	   ) return -EINVAL;
	return VL_VideoSetDeinterlacingMethod(window,mode);
}

/*---------------------------------------------------------------------
 * Description:
 *	Query the hw and set the vout_device values.
 * return:
 *	void
 * side-effects:
 * Notes:
 *  This is used after calling VideoInit, to get the maximum size of the
 *  output video
 *----------------------------------------------------------------------*/
static void vout_get_hw_parms(struct vout_device * dev)
{
	int width,height,depth, refresh;
	gfx_get_display_mode(&width,&height,&depth,&refresh);
	dev->clientfmt.width = width;
	dev->clientfmt.height = height;
	dev->clientfmt.depth = depth;
	if(depth == 16)
		dev->clientfmt.pixelformat = V4L2_PIX_FMT_UYVY;
	else
		dev->clientfmt.pixelformat = V4L2_PIX_FMT_GREY;

	/* Durango currently does not return the refresh rates for PAL
	 *  modes (50Hz). We assume a PAL mode if height = 576
	 */ 
	/* workaround */
	if(height == 576) refresh = 50; /* This must be a PAL mode */
	dev->clientfmt.priv = refresh;/*lets save the refresh rate over here */
	debug_msg("Output at %d x %d, %d bpp, %d Hz\n", 
				width, height,depth,refresh);
}

/*---------------------------------------------------------------------
 * Description:
 *	Set the output standard
 * return:
 *	Zero on success, error code on failure(-1).
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
 static int vout_set_standard( struct vout_device * dev, unsigned long std)
 {
	 vbi_decode_cleanup();
	 if(g_VideoLib_initialized)
	 {
		 /* de-init VideoLib */
		VL_VideoEnd();
		OS_Interface_close();
	 }
	 /* now it's safe to change things */
        g_vs.CRTEnable           = USE_CRT;
	switch(std) {
		case V4L2_STD_NTSC:
        		g_vs.TVStandard     = VL_TV_STANDARD_NTSC;
		        g_video_buf.address = NTSC_START_ADDRESS;
			VOUT_STD = VSOURCE_STD_NTSC; 
			break;
		case V4L2_STD_PAL:
			VOUT_STD = VSOURCE_STD_PAL; 		
                        g_vs.TVStandard = VL_TV_STANDARD_PAL;
                   	g_video_buf.address = PAL_START_ADDRESS;
                default: 
                        g_vs.TVStandard = VSOURCE_STD_NONE;
                   	g_video_buf.address = OTHERS_START_ADDRESS;
			err_msg("standard not supported\n");
            	break;
	}
	OS_Interface_init();
	if (g_VideoLib_initialized)
	{
		g_VideoLib_initialized = FALSE;
		if(init_videolib())
                return -1;

	}
	if(!vbi_decode_init(&vbi_dev))
            return -1;
	return 0;	 
 }

 
/*---------------------------------------------------------------------
 * Description:
 * 	Moves into or out of DVIP mode.
 * return:
 *	Zero on success, error code on failure.
 * side-effects:
 *  	A very noticable flicker is observed on the screen as the video 
 *	timings are changed.
 * Notes:
 *----------------------------------------------------------------------*/
 int vout_enable_dvip(int enable)
 {
	int video_was_enabled = gfx_get_video_enable();
	int alpha_enable = 0;
	int alpha_bits = 0;
	int i = 0;
	if(USE_CRT || NO_DVIP) return -EPERM;
	/* Since VidSetDirectVIP does a mode change, video
	 * and alpha will get disabled, we save the current
	 * values
	 */
	 for(i =0; i < 3; i++)
	 {	 
		 gfx_select_alpha_region(i);
		 gfx_get_alpha_enable(&alpha_enable);
		 if(alpha_enable)
		 { /* this region is enabled*/
			alpha_bits |= 0x01;
		 }
		 alpha_bits  = alpha_bits << 1;
	 }	 
	 alpha_bits  = alpha_bits >> 1;
	 /* The status of the threee alpha regions will be stored
	  * in alpha_bits
	  */
	if(enable)
	{
		if(VL_VideoSetFullScreenSource(VL_VIDEO_SOURCE_DVIP,
			get_genlock(capture_dev.input,VOUT_STD)))
			return -EINVAL;	
		VidSetDirectVIP(1,0);
#ifdef FLIP_DEC_FID_FOR_DVIP
                flip_decoder_fid(enable);
#endif
		g_DVIP_if_possible = TRUE;
	}
	else /* disable */
	{
		VL_VideoSetFullScreenSource(VL_VIDEO_SOURCE_MEMORY,
			get_genlock(capture_dev.input,VOUT_STD));	
		signalISR();
		VidSetDirectVIP(0,0);
		gfx_set_video_enable(1);
#ifdef FLIP_DEC_FID_FOR_DVIP
                flip_decoder_fid(enable);
#endif
		g_DVIP_if_possible = FALSE;
		g_inDVIP = FALSE;
		
	}
	/* restore the alpha regions and the video status */
	 for(i = 2; i >= 0; i--)
	 {	 
		 gfx_select_alpha_region(i);
		 alpha_enable = alpha_bits & 0x01;
		 gfx_set_alpha_enable(alpha_enable);
		 alpha_bits  = alpha_bits >> 1;
	 }	 
	/* enable video */
	if ( video_was_enabled) gfx_set_video_enable(1);
	return 0;
}

/*---------------------------------------------------------------------
 * Description:
 *	Sets up the output format.
 *	
 * return:
 *	Zero on success, error code on failure
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int vout_set_format(struct v4l2_format *fmt)
{
	int ret;
        g_video_buf.width = fmt->fmt.pix.width;
        g_video_buf.height = fmt->fmt.pix.height;
        ret = VL_VideoCfgMemoryWindow(&g_video_buf,0,0,fmt->fmt.pix.width,
                                                           fmt->fmt.pix.height);
	/* set the pixel format*/
	/* This is not supported by VideoLib, so we use durango */
	switch(fmt->fmt.pix.pixelformat)
	{
		case V4L2_PIX_FMT_UYVY:
			gfx_set_video_format(VIDEO_FORMAT_UYVY);	
			break;
		case V4L2_PIX_FMT_RGB565:
			gfx_set_video_format(VIDEO_FORMAT_RGB);	
			break;
		default:
			err_msg("Ignoring unknown format:\n");
			break;	
	}
	return ret;
}

/*----------------------------------------------------------------------
 * Description:
 *      Find the index into the  array for the specified V4L2_CID_x
 * Return:
 * 	    (integer) Index into capture_control array
 * side-effects:
 * Notes:
 *	This has been lifted from the sample driver distributed along with 
 *	the V4L2 package.
 *----------------------------------------------------------------------*/
static int find_voutctrl( int id )
{
	int i;
	if( ( id == V4L2_CID_PRIVATE_BASE ) 
		|| ( id <  V4L2_CID_PRIVATE_BASE )
		|| ( id >  GEODE_CID_LAST ) ) {
		return -EDOM;
	}
	for( i = MAX_VOUT_CONTROLS-1; i >= 0; i-- ) {
		if( vout_effects_control[i].id == id ) {
			break;
		}
	}
	if( i < 0 ) {
		i = -EINVAL;
	}
	return i;
}

static int set_vout_control(struct v4l2_control * vc)
{
	int			i;
	struct v4l2_queryctrl * control; 
	i = find_voutctrl(vc->id);
	if( i < 0 )
	{
		return -EINVAL;
	}
	control = &vout_effects_control[i];
	debug_msg("set_vout_control: i = %d\n", i);
	if( ((vc->value) < (control->minimum)) ||
	    ((vc->value) > (control->maximum)) ) {
			/* out of range */
			return -EINVAL;	
	}
	/* go ahead and set the control. As of now, the only control
	   we have is GEODE_CID_OVERLAY_MODE, so have it here for now.
	   This will have to be moved elsewhere when(if) we need to add
	   more controls 
	 */  
        if(VL_VideoSetOverlay((VL_VideoOverlayModeType) vc->value) == VL_STATUS_OK)	
        {
                g_overlay_mode = (VL_VideoOverlayModeType)vc->value ;
                return 0;
        }
	return -EINVAL;	
}

/*----------------------------------------------------------------------
 * Description:
 *      Enable/Disable preview
 * Return:
 * side-effects:
 *----------------------------------------------------------------------*/
static void vout_set_preview(int enable)
{
	if(enable)
	{
		VL_VideoEnable(1);
		gfx_set_video_enable(1);
	}
	else
		VL_VideoEnable(0);
}


/*---------------------------------------------------------------------
 * Description:
 *	The ioctl function for /dev/vout.
 * return:
 *	Zero on success, error code on failure.
 * Side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int v4l2_vout_ioctl( void *id, unsigned int cmd, void *arg )
{
	struct vout_device_open *o = (struct vout_device_open *)id;
	struct vout_device *dev = o->dev;

	switch(cmd)
	{
		case GEODE_IOCTL_SET_OVERLAY_INFO:
		{
			overlay_info_t* overlay_info = (overlay_info_t*)arg;
                        // If this ioctl is called, the colorkey will not be 
                        // set in the VIDIOC_S_WIN ioctl. This is the ensure
                        // that applications that use this interface and the
                        // applications that set the chromakey in the S_WIN
                        // ioctl work.
                        g_overlay_set_in_s_win = FALSE;
			set_overlay_info( overlay_info );
			debug_msg("(vout)GEODE_IOCTL_SET_OVERLAY_INFO: %x, %x\n",
				overlay_info->ColorKey, overlay_info->ColorMask);

			return 0;
		}
		case GEODE_IOCTL_SET_ALPHA_BLEND_INFO:
		{
			alpha_blend_info_t* alpha_blend_info 
					= (alpha_blend_info_t*)arg;
			set_alpha_blend_info( alpha_blend_info );
			debug_msg("(vout)GEODE_IOCTL_SET_ALPHA_BLEND_INFO: \n");
			return 0;
		}
		case VIDIOC_S_WIN:
		{
			struct v4l2_window *window = arg;
			debug_msg("(vout) VIDIOC_S_WIN\n");
			return vout_set_window(window);
		}
		case VIDIOC_PREVIEW:
		{
			int * enable = arg;
			debug_msg("(vout) VIDIOC_PREVIEW: %d\n",*enable);
		        vout_set_preview(*enable);
                        return 0;
		}
		case VIDIOC_S_STD:
		{
			struct v4l2_standard	*std = arg;
			int			id;
			id = v4l2_video_std_confirm(std);
			debug_msg("(vout) VIDIOC_S_STD\n");
			// Check to see if we support this standard.
			if( !( ( 1 << id ) & dev->standards ) )
			{
			err_msg("VIDIOC_S_STD: bad standard: %u\n", (unsigned)id);
				return -EINVAL;
			}
			return vout_set_standard(dev, id);
		}
		case VIDIOC_G_STD:
		{
			struct v4l2_standard *std = arg;
			v4l2_video_std_construct(std, dev->standard, 0);
			debug_msg("(vout) VIDIOC_G_STD\n");
			return 0;
		}
		case GEODE_IOCTL_S_DVIP:
		{
			int * enable = arg;
			debug_msg("vout) VIDIOC_S_DVIP : %d\n", *enable);
			return vout_enable_dvip(*enable); 
		}
		case VIDIOC_G_FMT:
		{
			struct v4l2_format *fmt = arg;
			fmt->fmt.pix = dev->clientfmt;
			debug_msg("(vout) VIDIOC_G_FMT\n");
			return 0;
		}
		case VIDIOC_S_FMT:
		{
			struct v4l2_format *fmt = arg;
			debug_msg("(vout) VIDIOC_S_FMT\n");
			return vout_set_format(fmt);
		}
		case VIDIOC_ENUMFX:
		{
			struct v4l2_fxdesc * desc = arg;	
			debug_msg("(vout) VIDIOC_ENUMFX : %d\n", desc->index);
			if(desc->index == 0) /* we have only one effect */
			{
				*desc= vout_effects[0];
			}
			return 0;
		}
		case VIDIOC_S_EFFECT:
		{
			int * effect = arg;
			debug_msg("(vout) VIDIOC_S_EFFECT :%d\n", *effect);
			/* we have only one effect for the vout device */
			/* dont need to do anything here */
			if(*effect != 0) /* only one effect */
				return -EINVAL;
			return 0;
		}
		case VIDIOC_S_CTRL:
		{
			struct v4l2_control	*vc = arg;
			debug_msg("(vout) VIDIOC_S_CTRL:\n");
			return set_vout_control(vc);

		}
		case VIDIOC_ENUM_FBUFFMT:
		{
			struct v4l2_fmtdesc *fmtdesc = arg;
			debug_msg("(vout) VIDIOC_ENUM_FBUFFMT\n");
			if (fmtdesc->index < 0 || fmtdesc->index > 1)
				return -EINVAL;
			*fmtdesc = vout_formats[fmtdesc->index];	
			return 0;

		}
		/* Get the base address of the video buffer in the graphics
		   memory (offset)
		 */  
		case VIDIOC_G_FBUF:
		{
			struct v4l2_framebuffer *framebuffer = arg;
			int output_pix_fmt;
			debug_msg("(vout) VIDIOC_G_FBUF\n");
			framebuffer->capability = V4L2_FBUF_CAP_CHROMAKEY
                                                 | V4L2_FBUF_CAP_EXTERNOVERLAY;
			framebuffer->flags = V4L2_FBUF_FLAG_OVERLAY
                                            | V4L2_FBUF_FLAG_CHROMAKEY;
			framebuffer->base[0] = 
				(void*) (g_video_buf.address - FRAME_BUFFER_START);
			/* Just so that no invalid address go out */
			framebuffer->base[1] = framebuffer->base[2] = 
				framebuffer->base[0]  ;
				
			framebuffer->fmt.width = g_video_buf.width;
			framebuffer->fmt.height = g_video_buf.height;
			framebuffer->fmt.depth = 16;
			framebuffer->fmt.flags = 0;	
			output_pix_fmt = gfx_get_video_format();
			switch(output_pix_fmt)
			{
				case VIDEO_FORMAT_RGB:
				framebuffer->fmt.pixelformat = V4L2_PIX_FMT_RGB565; 
				break;
				case VIDEO_FORMAT_UYVY:
				framebuffer->fmt.pixelformat = V4L2_PIX_FMT_UYVY; 
				framebuffer->fmt.flags = V4L2_FMT_CS_601YUV;	
				break;
				default:
				err_msg("Format not RGB or UYVY\n");
			}
			return 0;
		}
		/* Set the base address of the video buffer in the graphics
		   memory (offset)
		*/   
		case VIDIOC_S_FBUF:
		{
			struct v4l2_framebuffer *framebuffer = arg;
			debug_msg("(vout) VIDIOC_S_FBUF\n");
			if(  (framebuffer->base[0] < (void*)0) ||
			     (framebuffer->base[0] > (void*)
			     	TOTAL_GRAPHICS_MEMORY))
			{
				err_msg("Out of bounds\n");
				return -EINVAL;
			}
			g_video_buf.address = FRAME_BUFFER_START + 
						(unsigned long) framebuffer->base[0] ;
			g_video_buf.width = framebuffer->fmt.width;
			g_video_buf.height = framebuffer->fmt.height;
			return 0;

		}
		/* IOCTLS which directly map to VideoLib calls */
		case GEODE_IOCTL_S_COLOR_KEY:
		{
			geode_general_param_t * par = arg;
			debug_msg("(vout) GEODE_IOCTL_S_COLOR_KEY\n");
			return VL_VideoSetColorKey(par->p_1,par->p_2);
						
		}
		case GEODE_IOCTL_S_ALPHA_WINDOW:
		{ 
			geode_general_param_t * par = arg;
			debug_msg("(vout) GEODE_IOCTL_S_ALPHA_WINDOW\n");
			return VL_VideoSetAlphaWindow((unsigned char)par->p_1,
						(short)par->p_2,
						(short)par->p_3,
						(unsigned short)par->p_4,
						(unsigned short)par->p_5);
		}		
		case GEODE_IOCTL_S_ALPHA_OPER:
		{
			geode_general_param_t * par = arg;
			debug_msg("(vout) GEODE_IOCTL_S_ALPHA_OPER\n");
			return VL_VideoSetAlphaOperation(
						(unsigned char)par->p_1,
						(unsigned char)par->p_2,
						par->p_3,
						(int)par->p_4,
						(int)par->p_5,
						(char)par->p_6,
						(int)par->p_7);
		}					
		case GEODE_IOCTL_DEINTERLACE_FULL:
		{
			int * mode = arg;
			debug_msg("(vout) GEODE_IOCTL_DEINTERLACE_FULL\n");
			return vout_set_deinterlacing_method(
			VL_VIDEO_WINDOW_FULL,*mode);
		}
		case GEODE_IOCTL_DEINTERLACE_SCALED:
		{
			int * mode = arg;
			debug_msg("(vout) GEODE_IOCTL_DEINTERLACE_SCALED\n");
			return vout_set_deinterlacing_method(
			VL_VIDEO_WINDOW_SCALED,*mode);
		}
	/* Deprecated ioctls, retained here for backward compatibility */
		case GEODE_IOCTL_S_GRAPHICS_ON_VIDEO:
		{
			extern unsigned long color_key;
			extern unsigned long color_mask;
		        geode_general_param_t * par = arg;
			debug_msg("(vout) GEODE_IOCTL_S_GRAPHICS_ON_VIDEO\n");
			/* if we call this ioctl, we must be using the
                           older deprecated overlay interface too, so we 
                           need to disable setting the color key in S_WIN
			*/
        		g_overlay_set_in_s_win = FALSE;
			err_msg("Deprecated: Use the effects interface for overlay modes\n");
			if( ((int)par->p_1) )
			        gfx_set_video_color_key(color_key,color_mask,
				 			((int)par->p_2) );
			else
			        gfx_set_video_color_key(color_key,0,1);
			
			return 0;	 
		}
		case GEODE_IOCTL_VIDEO_OUTSIDE_ALPHA:
		{
		    geode_general_param_t * par = arg;
		    debug_msg("(vout) GEODE_IOCTL_VIDEO_OUTSIDE_ALPHA\n");
		    err_msg("Deprecated: Use the effects interface for overlay modes\n");
		    gfx_set_no_ck_outside_alpha( (int)par->p_1);
        	    g_overlay_set_in_s_win = FALSE;
                    return 0;

		}
		
		case VIDIOC_STREAMON:
		case VIDIOC_STREAMOFF:

		default:
			return -ENOIOCTLCMD;
	}
	return 0;
}

/*---------------------------------------------------------------------
 * Description:
 *	memory map function for vout device.
 * return:
 *	error code ( Not supported)
 * Side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int v4l2_vout_mmap( void *id, struct vm_area_struct *vma )
{
	return -EINVAL;
}

/*---------------------------------------------------------------------
 * Description:
 *	poll function for vout device.
 * return:
 *	error code ( Not supported)
 * Side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int v4l2_vout_poll( void *id, struct file *file, poll_table *table )
{
	return -EINVAL;
}

/*---------------------------------------------------------------------
 * Description:
 *	read function for vout device.
 * return:
 *	error code
 * Side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static long v4l2_vout_read( void *id, char *buf, unsigned long count, int noblock )
{
	return -EINVAL;
}
/*---------------------------------------------------------------------
 * Description:
 *	Further configuration of the vout device.
 * return:
 *	Zero
 * Side-effects:
 * Notes:
 *	This function is required by the V4L2 interface.
 *----------------------------------------------------------------------*/
static int v4l2_vout_init_done( struct v4l2_device *v )
{
	struct vout_device * dev = (struct vout_device *) v;
	dev->standards	= ( 1 << V4L2_STD_NTSC )   
			| ( 1 << V4L2_STD_PAL );
	switch(VOUT_STD)
	{
		case VSOURCE_STD_NTSC:
			dev->standard = V4L2_STD_NTSC; break;
		case VSOURCE_STD_PAL:
			dev->standard = V4L2_STD_PAL; break;
		/* TODO: Add more standards here */			
                default:
                        err_msg("Unknown standard\n");    
        	        dev->standard = VOUT_STD;

	}
	vout_get_hw_parms(dev);
	return 0;
}


/*---------------------------------------------------------------------
 * Description:
 *	Configure the vout device
 * return:
 *	TRUE
 * Side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int __init config_vout_device( struct vout_device *dev )
{
	sprintf(dev->shortname, "vout0");

	sprintf(dev->v.name, "NSC Geode V4L2 Overlay driver" );
	dev->v.type = V4L2_TYPE_OUTPUT;
	dev->v.minor = VOUT_MINOR;
	dev->v.open = v4l2_vout_open;
	dev->v.close = v4l2_vout_close;
	dev->v.read = v4l2_vout_read;
	dev->v.write = v4l2_vout_write;
	dev->v.ioctl = v4l2_vout_ioctl;
	dev->v.mmap = v4l2_vout_mmap;
	dev->v.poll = v4l2_vout_poll;
	dev->v.initialize = v4l2_vout_init_done;
	dev->v.priv = NULL;
	return 1;
}
/*-------------------------------------------------------------------
 *   VIDEO Device 
 *   The capture device 
 *-------------------------------------------------------------------*/

#ifndef NO_DECODER
/*---------------------------------------------------------------------
 * Description:
 *	Brightness setting for the analog inputs.
 * return:
 *  void	
 * side-effects:
 * Notes:
 *	Values: 0 to 255
 *----------------------------------------------------------------------*/
static void device_brightness(struct capture_device *dev, int x)
{
	tv_saa7114_set_decoder_brightness( (BYTE)x );
}


/*---------------------------------------------------------------------
 * Description:
 *	Contrast setting for the analog inputs.
 * return:
 *  void
 * side-effects:
 * Notes:
 *	Values: 0 to 255
 *----------------------------------------------------------------------*/
static void device_contrast(struct capture_device *dev, int x)
{
	tv_saa7114_set_decoder_contrast( (BYTE)x );
}

/*---------------------------------------------------------------------
 * Description:
 *	Color saturation setting for the analog inputs.
 * return:
 *  void
 * side-effects:
 * Notes:
 *	Values: 0 to 255
 *----------------------------------------------------------------------*/
static void device_saturation(struct capture_device *dev, int x)
{
	tv_saa7114_set_decoder_saturation( (BYTE)x );
}

/*---------------------------------------------------------------------
 * Description:
 *	Hue setting for the analog inputs.
 * return:
 *	None
 * side-effects:
 * Notes:
 *	Values: 0 to 255
 *----------------------------------------------------------------------*/
static void device_hue(struct capture_device *dev, int x)
{
	tv_saa7114_set_decoder_hue( (BYTE)x );
}

/*---------------------------------------------------------------------
 * Description:
 *	Sharpness setting for the analog inputs (except S-Video)
 * return:
 *	None
 * side-effects:
 * Notes:
 *	The name is misleading. This function is named so because it 
 *	changes the luminance register of the 7114, a part of which is 
 *	actually the filter settings. This function will have no effect 
 *	for S-Video input because the filter is bypassed.
 *----------------------------------------------------------------------*/
static void device_luminance(struct capture_device *dev, int x)
{
	tv_saa7114_set_luminance( (BYTE)x );
}

/*---------------------------------------------------------------------
 * Description:
 *	Set all the video controls for the current input.
 * return:
 *	Void
 * side-effects:
 * Notes:
 *	Each video control value for each input is stored independantly.
 *	This makes it possible to have different settings for each input.
 *	The values which have been set before will be restored when the
 *	input has changed.
 *----------------------------------------------------------------------*/
static void device_tone_controls(struct capture_device *dev)
{
	/* dont do anything if input is MPEG */
	int	*ctrl;
	if(dev->input == VSOURCE_MPEG_DEC) return;
	ctrl = dev->source[dev->input].control;
	save_flags(flags);
	cli();
	device_brightness(dev, ctrl[VCTRL_BRIGHTNESS]);
	device_contrast(dev, ctrl[VCTRL_CONTRAST]);
	device_saturation(dev, ctrl[VCTRL_SATURATION]);
 	device_hue(dev, ctrl[VCTRL_HUE]); 
 	device_luminance(dev, ctrl[VCTRL_WHITENESS]); 
	restore_flags(flags);
}
#endif /* NO_DECODER*/


/*---------------------------------------------------------------------
 * Description:
 *	Initialize the capture device.
 * return:
 *	TRUE
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int geode_capture_initialize(struct capture_device *dev)
{
	/*  Video decoder information fields	*/
	dev->videc.standards	= ( 1 << V4L2_STD_NTSC ) 
				| ( 1 << V4L2_STD_PAL );
			/*	
				| ( 1 << V4L2_STD_PAL_M )
				| ( 1 << V4L2_STD_PAL_N )
				| ( 1 << V4L2_STD_PAL_60 )
				| ( 1 << V4L2_STD_NTSC_N ); */
#ifndef NO_DECODER
	dev->videc.hskip[VSOURCE_STD_NTSC]  = 10;
	dev->videc.vskip[VSOURCE_STD_NTSC]  = 22; /* lines 10-21 are VBI */
	dev->videc.width[VSOURCE_STD_NTSC]  = 704;
	dev->videc.height[VSOURCE_STD_NTSC] = 240;

	dev->videc.hskip[VSOURCE_STD_PAL]  = 10; 
	dev->videc.vskip[VSOURCE_STD_PAL]  = 23; /* lines 6-22 are VBI */
	dev->videc.width[VSOURCE_STD_PAL]  = 720;
	dev->videc.height[VSOURCE_STD_PAL] = 288;

/* VBI values */
	dev->videc.vbi_hskip[VSOURCE_STD_NTSC] = 6;
	dev->videc.vbi_vskip[VSOURCE_STD_NTSC] = 10; /* lines 10-21 are VBI */
	dev->videc.vbi_width[VSOURCE_STD_NTSC] = 720;
	dev->videc.vbi_height[VSOURCE_STD_NTSC]= 12;

	dev->videc.vbi_hskip[VSOURCE_STD_PAL] = 3; 
	dev->videc.vbi_vskip[VSOURCE_STD_PAL] = 6; /* lines 6-22 are VBI */
	dev->videc.vbi_width[VSOURCE_STD_PAL] = 720;
	dev->videc.vbi_height[VSOURCE_STD_PAL]= 17;
#endif	
	/* no offset  for the MPEG decoder input */
	dev->videc.hskip[VSOURCE_STD_SIGMA_NTSC] = 0;
	dev->videc.vskip[VSOURCE_STD_SIGMA_NTSC] = 0;
	dev->videc.width[VSOURCE_STD_SIGMA_NTSC] = 720;
	dev->videc.height[VSOURCE_STD_SIGMA_NTSC]= 240;

	dev->videc.hskip[VSOURCE_STD_SIGMA_PAL] = 0;
	dev->videc.vskip[VSOURCE_STD_SIGMA_PAL] = 0;
	dev->videc.width[VSOURCE_STD_SIGMA_PAL] = 720;
	dev->videc.height[VSOURCE_STD_SIGMA_PAL]= 288;
	
	/* we have 4 inputs, tuner, composite, s-video and mpeg-decoder */
	/* the inputs are 0,1,2 and 3. */
#ifndef NO_DECODER	
	dev->videc.num_inputs		= 3; 
	dev->videc.decoder_is_stable	= TRUE;
#else
	dev->videc.num_inputs		= 1; 
	dev->videc.decoder_is_stable	= FALSE;
	
#endif  /*NO_DECODER*/		
	return TRUE;
}

/*---------------------------------------------------------------------
 * Description:
 *	Get the correct genlock value for a particular input and standard.
 * return:
 *	genlock value for given input and output standard.
 * side-effects:
 * Notes:
 *	The genlock values are as suggested by NSTA. Which are as follows.
 *	New genlock values have to be found for the new decoder settings.
 *
 *  Video Path              | NTSC  |  PAL
 *  ------------------------+-------+--------
 *  EM8400 directly to VIP   0x3900  0x3800
 *  EM8400 to 7114 XPORT     0x3900   0x3800
 *  7114 CVBS                0x300   0 ( Note: 0x300 is used for PAL)
 *----------------------------------------------------------------------*/
static int get_genlock(int input, int vout_std)
{
/* Not tested for SIGMA_DIRECT */
	switch(vout_std)
	{
		case VSOURCE_STD_NTSC:
#ifdef SIGMA_DIRECT
		return 0x3900;
#else		
		if(input == VSOURCE_MPEG_DEC)
			return 0x3900;
		else return 0x3900; 
#endif		
		case VSOURCE_STD_PAL:
#ifdef SIGMA_DIRECT
		return 0x3800;
#else				
		if(input == VSOURCE_MPEG_DEC)
			return 0x3800;
		else return 0x2900;		
#endif		
	}
	return 0;
}


/*---------------------------------------------------------------------
 * Description:
 *	Sets the input of the decoder. The inputs are...
 *	0 -> Tuner
 *	1 -> Composite input
 *	2 -> S-Video
 *	3 -> X-Port Digital video
 * return:
 *	TRUE
 * side-effects:
 *	This function will put the hardware into VIP if in DVIP.
 * Notes:
 *	This function is quite involved, especially when switching to and
 *	from the digital input. Also, the field-inversion problem when the
 *	SIGMA decoder is connected is also taken care of.
 *----------------------------------------------------------------------*/
static int geode_capture_set_input(struct capture_device *dev, int input)
{	  
#ifndef SIGMA_DIRECT	
#ifndef NO_DECODER	
	/* first, switch to VIP if in DVIP, we do this by fooling VideoLib
	 *  into believing that we've lost the video signal */
	if(dev->input != input) 
		signalISR(); /* this is a VideoLib function */   
    debug_msg("Setting input to %d, old input %d\n", input, dev->input);
	save_flags(flags);
	cli();
	if(input != VSOURCE_MPEG_DEC) {
	        VL_FieldPolarityInvert(0);
		g_input_window_index=VIN_STD;
		tv_saa7114_set_input(dev->input,input,
			dev->videc.hskip[g_input_window_index],	
			dev->videc.vskip[g_input_window_index],
                        dev->videc.width[g_input_window_index],
                        dev->videc.height[g_input_window_index],
                        g_video_buf.width,
                        g_video_buf.height, VIN_STD-1
		);
	    if(VIN_STD == VSOURCE_STD_PAL) {
                tv_saa7114_set_tv_standard(TV_SAA7114_STANDARD_PAL);
		    /* maybe a different setting for PAL */
	    }
	}
	if( (input != VSOURCE_MPEG_DEC) && (dev->input == VSOURCE_MPEG_DEC)) {
		/* MPEG to non-MPEG */
		VL_FieldPolarityInvert(0);
	}
	else if(input == VSOURCE_MPEG_DEC) { /* TO MPEG */
		g_input_window_index = VSOURCE_STD_SIGMA_NTSC;
		switch(VOUT_STD)
		{
			case VSOURCE_STD_PAL:
			g_input_window_index = VSOURCE_STD_SIGMA_PAL;
			break;
			/* TODO: add more standards here */
		}
		tv_saa7114_set_input(dev->input,input,
			dev->videc.hskip[g_input_window_index],	
			dev->videc.vskip[g_input_window_index],
            		dev->videc.width[g_input_window_index],
            		dev->videc.height[g_input_window_index],
           	 	g_video_buf.width,
            		g_video_buf.height, VIN_STD-1
		);
		VL_FieldPolarityInvert(1);
	}
	restore_flags(flags);
    dev->input = input;

	blue_screen();
#endif /* NO_DECODER */
	/* Turn off DVIP */
	VL_VideoSetFullScreenSource(VL_VIDEO_SOURCE_MEMORY,
		get_genlock(input,VOUT_STD));
	g_DVIP_if_possible = FALSE;
	g_inDVIP = FALSE;
#endif /* !SIGMA_DIRECT */	
        return TRUE;
}


/*---------------------------------------------------------------------
 * Description:
 *	Set the capture standard. (Decoder setting)
 * return:
 *      1 on success & 0 on faliure
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int geode_capture_set_standard(struct capture_device *dev, int std)
{
#ifndef NO_DECODER	
	int x;
	dev->videc.standard = std;
	save_flags(flags);
	cli();
	switch( std )
	{
		case V4L2_STD_NTSC:
			x = VSOURCE_STD_NTSC;
                        dev->videc.frame_period = 333667;
                        tv_saa7114_set_tv_standard(TV_SAA7114_STANDARD_NTSC);
//                        tv_saa7114_set_decoder_input_offset(
  //                                              dev->videc.hskip[x],
    //                                            dev->videc.vskip[x]
      //                                          );
        //                tv_saa7114_set_decoder_input_size(
          //                                      dev->videc.width[x],
            //                                    dev->videc.height[x]*2);
              //          tv_saa7114_set_decoder_output_size(
              //                                  g_video_buf.width,
              //                                  g_video_buf.height);
                        tv_saa7114_set_decoder_scale(
					dev->videc.hskip[x],
					dev->videc.vskip[x],
                                        dev->videc.width[x],
                                        dev->videc.height[x],
                                        g_video_buf.width,
                                        g_video_buf.height, VD_NTSC ); 

                        tv_saa7114_set_vbi_input_offset(
                                                dev->videc.vbi_hskip[x],
                                                dev->videc.vbi_vskip[x]
                                                );
                        tv_saa7114_set_vbi_input_size(
                                                dev->videc.vbi_width[x],
                                                dev->videc.vbi_height[x]);
                        tv_saa7114_set_vbi_output_size(
                                                dev->videc.vbi_width[x],
                                                dev->videc.vbi_height[x]); 

                        VIN_STD = VSOURCE_STD_NTSC;  /* NTSC */
                      //  g_video_buf.address = NTSC_START_ADDRESS;
		break;
                case V4L2_STD_PAL:
			x = VSOURCE_STD_PAL;
                        dev->videc.frame_period = 400000;
                        tv_saa7114_set_tv_standard(TV_SAA7114_STANDARD_PAL);
                        tv_saa7114_set_decoder_scale(
					dev->videc.hskip[x],
					dev->videc.vskip[x],
                                        dev->videc.width[x],
                                        dev->videc.height[x],
                                        g_video_buf.width,
                                        g_video_buf.height, VD_PAL ); 
                        tv_saa7114_set_vbi_input_offset(
                                                dev->videc.vbi_hskip[x],
                                                dev->videc.vbi_vskip[x]
                                                );
                        tv_saa7114_set_vbi_input_size(
                                                dev->videc.vbi_width[x],
                                                dev->videc.vbi_height[x]);
                        tv_saa7114_set_vbi_output_size(
                                                dev->videc.vbi_width[x],
                                                dev->videc.vbi_height[x]);
                        VIN_STD = VSOURCE_STD_PAL;
                       // g_video_buf.address = PAL_START_ADDRESS;
                        break;
                case V4L2_STD_SECAM:
                        dev->videc.frame_period = 400000;
                        return 0; /* error as SECAM not supported */     

                default:        
                       // g_video_buf.address = OTHERS_START_ADDRESS;
                        
	}
	restore_flags(flags);
#endif	
	return 1;
}

/*---------------------------------------------------------------------
 * Description:
 *  sets the vcr mode on/off for capture device 
 * return:
 *  1 on success 
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int geode_capture_set_vcrmode(struct capture_device *dev, int x)
{
	dev->source[dev->input].vcrmode = x;
	/*  TODO: Switch decoder to VCR sync timing mode  */
	return 1;
}

/*---------------------------------------------------------------------
 * Description:
 *  returns if the capture device is stable
 * return:
 *  int
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int geode_capture_is_stable(struct capture_device *dev)
{
	return signal_ok();
}


/*---------------------------------------------------------------------
 * Description:
 *	Set the video input. This is the function which is actually called
 *	The limit checking for the input is done here.
 * return:
 *  void 
 * side-effects:
 *	Changes the input.
 * Notes:
 *----------------------------------------------------------------------*/
static void set_video_input(struct capture_device *dev,
		                    int input)
{
#ifndef NO_DECODER	
	if( input < 0 || input > dev->videc.num_inputs )
	{
		return;
	}
	dev->videc.set_input(dev, input);
	dev->videc.set_vcrmode(dev, dev->source[input].vcrmode);
	device_tone_controls(dev);
#endif	
}

/*---------------------------------------------------------------------
 * Description:
 *	open function for video device	
 * return:
 *	Zero on success, error code on failure.
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int v4l2_open(struct v4l2_device *v, int o_flags, void **idptr)
{
	struct capture_device *dev = (struct capture_device *)v;
	int     i, n;
    int     cap;
	unsigned long cpu_flags;

        for (i = 0, n = -1, cap = 0; i < MAX_OPENS; ++i)
        {
                if (!dev->open_data[i].isopen)
                        n = i;/* available open_data structure */
                else if (!dev->open_data[i].noncapturing)
                        cap = 1;/* another open is already capturing */
        }
        if (n == -1)/* No available open_data structures */
        {
                err_msg("No more opens on this device\n");
                return -EBUSY;
        }

        if (o_flags & O_NOIO)/*  Non-capturing open */
                dev->open_data[n].noncapturing = 1;
        else if (cap)
        {
                err_msg("No more capturing opens on this device\n");
                return -EBUSY;
        }
        else
        {
                dev->open_data[n].noncapturing = 0;
                /*  Keep track of whether there is a capturing open  */
                ++dev->capturing_opens;
        }

        ++dev->open_count;
        dev->open_data[n].isopen = 1;
        dev->open_data[n].dev = dev;
        *idptr = &dev->open_data[n];

        if (dev->open_count == 1)
        {
#ifndef NO_DECODER                
		save_flags(cpu_flags);
		cli();
		if(!g_timer_activated)
			add_timer(&geode_timer);
		g_timer_activated++;	
		restore_flags(cpu_flags);
#endif                
		debug_msg("First open on capture device\n");

        }
        debug_msg("Open succeeded\n");		
	MOD_INC_USE_COUNT;
	return 0;
}


/*---------------------------------------------------------------------
 * Description:
 *	close function for video device	
 * return:
 *	Zero on success, error code on failure.
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static void v4l2_close(void *id)
{
	struct capture_device_open *o = (struct capture_device_open *)id;
	struct capture_device *dev = o->dev;
    
	if (!o->noncapturing)
        {
                --dev->capturing_opens;
                debug_msg("Close\n");
        }
        o->isopen = 0;
        --dev->open_count;
        if (dev->open_count == 0)
        {	
#ifdef RESET_ON_CLOSE		
		save_flags(flags);
		cli();
		g_timer_activated--;
		if(!g_timer_activated)
			del_timer(&geode_timer);
		restore_flags(flags);	
#endif		
		debug_msg("Last close on capture device\n");
        }
	MOD_DEC_USE_COUNT;

}

/*---------------------------------------------------------------------
 * Description:
 *	write function for video device	
 * return:
 *	long
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static long v4l2_write(void* id, const char* buf, unsigned long count, int noblock )
{
	return -EINVAL;
}

/*---------------------------------------------------------------------
 * Description:
 *	function implementing ioctl calls for video device	
 * return:
 *	zero on success, error code on faliure	
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int v4l2_ioctl(void *id, /* private pointer */
                	   unsigned int	cmd, /* command */
	                   void		*arg /* argument to the ioctl */)
{
	struct capture_device_open *o = (struct capture_device_open *)id;
	struct capture_device *dev = o->dev;

	switch(cmd)
	{
	case VIDIOC_QUERYCAP:
	{
		struct v4l2_capability *capb = arg;
		debug_msg("(capture) VIDIOC_QUERYCAP\n");
		strcpy(capb->name, dev->v.name);
		capb->type = V4L2_TYPE_CAPTURE;
		capb->flags = V4L2_FLAG_READ
#ifndef EXCLUDE_TUNER                                
			      | V4L2_FLAG_TUNER
#endif                                
                              | V4L2_FLAG_DATA_SERVICE
			      | V4L2_FLAG_PREVIEW ;

		capb->inputs = dev->videc.num_inputs+1; 
		capb->outputs = 1;
		capb->audios = 0;
		capb->maxwidth = dev->videc.width[VIN_STD];
		capb->maxheight = dev->videc.height[VIN_STD]*2;
		switch(VIN_STD)
		{
			case VSOURCE_STD_NTSC:
				capb->maxframerate = 30;
				break;
			case VSOURCE_STD_PAL:
				capb->maxframerate = 25;
				break;
		}
		capb->minwidth = MIN_WIDTH;
		capb->minheight = MIN_HEIGHT;
		return 0;
	}
	case VIDIOC_G_FMT:
	{
		struct v4l2_format *fmt = arg;
		debug_msg("(capture) VIDIOC_G_FMT\n");
		fmt->fmt.pix = dev->clientfmt;
		return 0;
	}

	case VIDIOC_S_FMT:
	{
		struct v4l2_format *fmt = arg;
		int std = VD_NTSC;
		debug_msg("(capture) VIDIOC_S_FMT\n");
		/* Limit checking */
                if( (fmt->fmt.pix.width < 0) || (fmt->fmt.pix.height < 0))
                        return  -EINVAL;
                if( (fmt->fmt.pix.width > vout_dev.clientfmt.width) ||
                  ( (fmt->fmt.pix.height) > vout_dev.clientfmt.height)
                 )
                   return -EINVAL;
#ifndef NO_DECODER
                switch(VIN_STD)
                {
                        case 1: /* NTSC */
			std = VD_NTSC; break;
                        case 2: /* PAL */
			std = VD_PAL; break;
                }
#endif		
                dev->clientfmt = fmt->fmt.pix;
                dev->clientfmt.height = fmt->fmt.pix.height & ( ~0x01) ; 
                debug_msg("S_FMT: Height %d, width %d\n", dev->clientfmt.height,
                                                         dev->clientfmt.width );
		/* compute the size of the image if the user wants to read it */
		dev->clientfmt.sizeimage = (dev->clientfmt.width * dev->clientfmt.height) * 2;
		/* WARNING!! We assume only 16bit UYVY capture */
                /* Adjust to 8 pixel multiples */
                //dev->clientfmt.width = dev->clientfmt.width & ~0x7;
#ifndef NO_DECODER		
		save_flags(flags);
		cli();
                tv_saa7114_set_decoder_output_size(
                                dev->clientfmt.width ,
                                dev->clientfmt.height);
#endif		
                g_video_buf.width = dev->clientfmt.width;
                g_video_buf.height = dev->clientfmt.height ;
                VL_VideoCfgCaptureBuf(&g_video_buf);
#ifndef NO_DECODER		
		tv_saa7114_set_decoder_scale(
				dev->videc.hskip[g_input_window_index],
				dev->videc.vskip[g_input_window_index],
				dev->videc.width[g_input_window_index],
				dev->videc.height[g_input_window_index],
				g_video_buf.width,
				g_video_buf.height, std ); 
		restore_flags(flags);
#endif		
		//blue_screen();
		return 0;
	}
	case VIDIOC_S_WIN:	
	{
		struct v4l2_window *window = arg;
		debug_msg("(capture) VIDIOC_S_WIN %d %d %d %d\n",
 			window->x,window->y,window->width,window->height);
#ifndef NO_DECODER		
		save_flags(flags);
		cli();
		  tv_saa7114_set_decoder_output_size(
                                    window->width,
                                    window->height);
#endif		  
                    g_video_buf.width = window->width;
                    g_video_buf.height = window->height ;
                    VL_VideoCfgCaptureBuf(&g_video_buf);
#ifndef NO_DECODER		    
		tv_saa7114_set_decoder_scale(
				dev->videc.hskip[g_input_window_index],
				dev->videc.vskip[g_input_window_index],
				dev->videc.width[g_input_window_index],
				dev->videc.height[g_input_window_index],
				g_video_buf.width,
				g_video_buf.height, VIN_STD -1  ); 
		restore_flags(flags);    
#endif		    
                vout_set_window(window); 
		return 0;
	}


	case VIDIOC_G_INPUT:
	{
		debug_msg("(capture) VIDIOC_G_INPUT\n");
		memcpy(arg, &dev->input, sizeof(dev->input));
		return 0;
	}

	case VIDIOC_S_INPUT:
	{
		int	input = *(int *)arg;
		debug_msg("(capture) VIDIOC_S_INPUT\n");
		if (input < 0 || input > dev->videc.num_inputs)
		{
			err_msg("VIDIOC_S_INPUT: out of range %d\n",input);
			return -EINVAL;
		}
		set_video_input(dev, input);
		return 0;
	}


	case VIDIOC_G_STD:
	{
		struct v4l2_standard *std = arg;
		debug_msg("(capture) VIDIOC_G_STD\n");
		v4l2_video_std_construct(std, dev->videc.standard, 0);
		return 0;
	}
	/* TODO: This ioctl should be allowed _only_ if capture is
             * inactive and no other process has opened this device
         */
	case VIDIOC_S_STD:
	{
		struct v4l2_standard	*std = arg;
		int			id;
		debug_msg("(capture) VIDIOC_S_STD\n");
		id = v4l2_video_std_confirm(std);

		// Check to see if we support this standard.
		if( !( ( 1 << id ) & dev->videc.standards ) )
		{
			err_msg("VIDIOC_S_STD: bad standard: %u\n", (unsigned)id);
			return -EINVAL;
		}
		dev->videc.set_standard(dev, id);
		return 0;
	}

	case VIDIOC_ENUMSTD:
	{
		struct v4l2_enumstd *estd = arg;
		unsigned long	b, i;
		debug_msg("(capture) VIDIOC_ENUMSTD\n");
		if (estd->index < 0 || estd->index > 30)
			return -EINVAL;
		for (b = 1, i = 0; b < 32; ++b)
		{
			if (((1 << b) & dev->videc.standards) == 0)
				continue;
			if (i == estd->index)
			{
				v4l2_video_std_construct(&estd->std, b, 0);
				estd->inputs = (unsigned long)-1; /* all inputs */
				estd->outputs = 0;
				return 0;
			}
			++i;
		}
		return -EINVAL;
	}

	case VIDIOC_ENUMINPUT:
	{
		struct v4l2_input *vi = arg;
		debug_msg("(capture) VIDIOC_ENUMINPUT\n");
		if (vi->index < 0 || vi->index > dev->videc.num_inputs)
			return -EINVAL;
		*vi = dev->source[vi->index].input;
		return 0;
	}

	case VIDIOC_QUERYCTRL:
	{
		struct v4l2_queryctrl	*qc = arg;
		int			i;
		debug_msg("(capture) VIDIOC_QUERYCTRL\n");
		i = find_vctrl(qc->id);
		if (i == -EINVAL)
		{
			qc->flags = V4L2_CTRL_FLAG_DISABLED;
			return 0;
		}
		if (i < 0)
		{
			return -EINVAL;
		}
		/*  V4L2 filled in category and group, preserve them */
		capture_control[i].category = qc->category;
		memcpy(capture_control[i].group, qc->group,sizeof(qc->group));
		*qc = capture_control[i];
		return 0;
	}


	case VIDIOC_G_CTRL:
	{
		struct v4l2_control	*vc = arg;
		int			i;
		i = find_vctrl(vc->id);
		if (i < 0)
			return -EINVAL;
		vc->value = dev->source[dev->input].control[i];
		return 0;
	}

	case VIDIOC_S_CTRL:
	{
		struct v4l2_control	*vc = arg;
		int			i;
		debug_msg("(capture) VIDIOC_S_CTRL\n");
		i = find_vctrl(vc->id);
		if( i < 0 )
		{
			return -EINVAL;
		}
		dev->source[dev->input].control[i] = vc->value;
#ifndef NO_DECODER		
		device_tone_controls(dev);
#endif		
		return 0;
	}

	case VIDIOC_G_TUNER:	
	{
#ifndef EXCLUDE_TUNER
		debug_msg("(capture) VIDIOC_G_TUNER\n");
		memcpy(arg, &dev->tuner, sizeof(dev->tuner));
#endif
		return 0;				  
	}

	case VIDIOC_S_TUNER:	
	{
#ifndef EXCLUDE_TUNER
		debug_msg("(capture) VIDIOC_S_TUNER\n");
		memcpy(&dev->tuner, arg, sizeof(dev->tuner));
#endif
		return 0;
	}

	case VIDIOC_G_FREQ:	
	{
#ifndef EXCLUDE_TUNER
		unsigned long *freq = (unsigned long*)arg;
		debug_msg("(capture) VIDIOC_G_FREQ\n");
		*freq = tv_tuner_get_freq();
#endif
		return 0;
	}

	case VIDIOC_S_FREQ:	
	{
#ifndef EXCLUDE_TUNER
		int freq = *(unsigned long*)arg;
		int tuner_i2c_add;
		int tuner_type;
		debug_msg("(capture) VIDIOC_S_FREQ %d\n", freq);
#ifdef CYGNUS
#ifdef CYGNUS_PAL_TUNER
		tuner_i2c_add = 0xC2;
		tuner_type = TUNER_TYPE_ALPS_TMDZ2;
#else
		tuner_i2c_add = 0xC6;
		tuner_type = TUNER_TYPE_ALPS_TSBH1;
#endif
#else
		tuner_i2c_add = 0xC6;
		tuner_type = TUNER_TYPE_PHILIPS_NTSC;
#endif
#ifdef FREQ_IS_CHANNEL
		tv_tuner_change_channel(tuner_i2c_add,tuner_type,0,freq );
#else
		tv_tuner_change_frequency(tuner_i2c_add,tuner_type,freq );
#endif



#endif




		return 0;
	}
	case VIDIOC_STREAMON:
	{
		debug_msg("(capture) VIDIOC_STREAMON\n");
		VL_VideoResume();
		return 0;	
	}
	case VIDIOC_STREAMOFF:
	{
		debug_msg("(capture) VIDIOC_STREAMOFF\n");
		VL_VideoPause();
		return 0;
	}
	/* We support only one capture format: UYVY */
	case VIDIOC_ENUM_PIXFMT:
	{
		struct v4l2_fmtdesc *fmtdesc = arg;
		debug_msg("(capture) VIDIOC_ENUM_PIXFMT\n");
		if (fmtdesc->index < 0 || fmtdesc->index > 0)
			return -EINVAL;
		*fmtdesc = capture_formats[fmtdesc->index];	
		return 0;
	}
	case VIDIOC_PREVIEW:
	{
		int * enable = arg;
		debug_msg("(capture) VIDIOC_PREVIEW: %d\n",*enable);
	        vout_set_preview(*enable);
                return 0;
	}

	case VIDIOC_G_AUDIO:	
		return 0;

	case VIDIOC_S_AUDIO:	
		return 0;

	case GEODE_IOCTL_DEBUG_PHYS:
#ifndef NO_DECODER	
		debug_msg("v4l2_ioctl(): GEODE_IOCTL_DEBUG_PHYS\n");
		if( ((geode_ioctl_debug_phys_t*)arg)->control == GEODE_IOCTL_DEBUG_PHYS_READ )
		{
			tv_dbg_phys_read( 
				((geode_ioctl_debug_phys_t*)arg)->addr, 
				(unsigned long*)&((geode_ioctl_debug_phys_t*)arg)->data 
			);
		}
		else if( ((geode_ioctl_debug_phys_t*)arg)->control == GEODE_IOCTL_DEBUG_PHYS_WRITE )
		{
			tv_dbg_phys_write( 
				((geode_ioctl_debug_phys_t*)arg)->addr, 
				((geode_ioctl_debug_phys_t*)arg)->data 
			);
		}
#endif		
		break;
#ifdef DEBUG		
	case GEODE_IOCTL_DEBUG_PORTIO:
#ifndef NO_DECODER	
		if( ((geode_ioctl_debug_portio_t*)arg)->control == GEODE_IOCTL_DEBUG_PORTIO_READ_8 )
		{
			tv_dbg_portio_read_8( 
				((geode_ioctl_debug_portio_t*)arg)->addr, 
				(unsigned long*)&((geode_ioctl_debug_portio_t*)arg)->data 
			);
		}
		else if( ((geode_ioctl_debug_portio_t*)arg)->control == GEODE_IOCTL_DEBUG_PORTIO_READ_16 )
		{
			tv_dbg_portio_read_16( 
				((geode_ioctl_debug_portio_t*)arg)->addr, 
				(unsigned long*)&((geode_ioctl_debug_portio_t*)arg)->data 
			);
		}
		else if( ((geode_ioctl_debug_portio_t*)arg)->control == GEODE_IOCTL_DEBUG_PORTIO_READ_32 )
		{
			tv_dbg_portio_read_16( 
				((geode_ioctl_debug_portio_t*)arg)->addr, 
				(unsigned long*)&((geode_ioctl_debug_portio_t*)arg)->data 
			);
		}
		else if( ((geode_ioctl_debug_portio_t*)arg)->control == GEODE_IOCTL_DEBUG_PORTIO_WRITE_8 )
		{
			tv_dbg_portio_write_8( 
				((geode_ioctl_debug_portio_t*)arg)->addr, 
				((geode_ioctl_debug_portio_t*)arg)->data 
			);
		}
		else if( ((geode_ioctl_debug_portio_t*)arg)->control == GEODE_IOCTL_DEBUG_PORTIO_WRITE_16 )
		{
			tv_dbg_portio_write_16( 
				((geode_ioctl_debug_portio_t*)arg)->addr, 
				((geode_ioctl_debug_portio_t*)arg)->data 
			);
		}
		else if( ((geode_ioctl_debug_portio_t*)arg)->control == GEODE_IOCTL_DEBUG_PORTIO_WRITE_32 )
		{
			tv_dbg_portio_write_32( 
				((geode_ioctl_debug_portio_t*)arg)->addr, 
				((geode_ioctl_debug_portio_t*)arg)->data 
			);
		}
#endif	/* NO_DECODER */
		return 0;
		break;
#endif /* DEBUG */
				
#ifdef DEBUG	
	case GEODE_IOCTL_GET_DECODER_REG:
	{
#ifndef NO_DECODER 
		geode_ioctl_debug_i2c_t * dec = (geode_ioctl_debug_i2c_t*)arg;
		save_flags(flags);
		cli();
		dec->data=(unsigned char)tv_i2c_saa7114_read_byte_data(dec->addr);
		restore_flags(flags);
#endif
		return 0;	
	}
	case GEODE_IOCTL_SET_DECODER_REG:
	{
#ifndef NO_DECODER 
		geode_ioctl_debug_i2c_t * dec = (geode_ioctl_debug_i2c_t*)arg;
		save_flags(flags);
		cli();
		tv_i2c_saa7114_write_byte_data(dec->addr, dec->data);
		restore_flags(flags);
#endif
		return 0;	
	}
	
#endif	/* DEBUG */
	case VIDIOC_QUERYMENU:
		return 0;
	case VIDIOC_G_PARM:
	case VIDIOC_S_PARM:
	case VIDIOC_G_PERF:
	case VIDIOC_G_COMP:	
	case VIDIOC_S_COMP:	
	case VIDIOC_REQBUFS:
	case VIDIOC_QUERYBUF:
	case VIDIOC_QBUF:
	case VIDIOC_DQBUF:
	case VIDIOC_ENUM_FBUFFMT:	
	case VIDIOC_G_FBUF:
	case VIDIOC_S_FBUF:
	case VIDIOC_G_WIN:		
		return -EINVAL;
	default:
		return -ENOIOCTLCMD;
	}
	return 0;
}

/*---------------------------------------------------------------------
 * Description:
 *	memory map function for video device	
 * return:
 *	error code 	
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int v4l2_mmap(void *id, /* private pointer */
		     struct vm_area_struct *vma /* area to be mapped */
		     )
{
	return -ENODEV;
}

/*---------------------------------------------------------------------
 * Description:
 *	poll function for video device	
 * return:
 *	poll error	
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int v4l2_poll(void* id, struct file* file, poll_table* table )
{
	return POLLERR;
}

/*---------------------------------------------------------------------
 * Description:
 *	read function for video device	
 * returns:
 *      One frame (two fields) of data
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static long v4l2_read(void* id, char* buf, unsigned long count, int noblock)
{
	struct capture_device_open *o_dev = (struct capture_device_open *)id;
	ssize_t num_lines = o_dev->dev->clientfmt.height / 2; 
	ssize_t dest_offset  = 0;
	ssize_t source_offset =0;
	int i;
  	unsigned long odd_base;
	unsigned long even_base;

	//printk("V4L2 read\n");
        /* check if we have enough of buffer space */
        if (count < o_dev->dev->clientfmt.sizeimage) {
		err_msg("Not enough buffer space\n");
                return -EINVAL; /* we must have space for at-least one
                                   buffer
                                 */

        } 
	odd_base  = gfx_get_vip_base(1) + (unsigned long)gfx_virt_fbptr 
					- (unsigned long)gfx_phys_fbptr;
	even_base = gfx_get_vip_base(0) + (unsigned long)gfx_virt_fbptr
					- (unsigned long)gfx_phys_fbptr;

        for(i = 0; i < num_lines ; i ++) {
		copy_to_user((void*)(buf+dest_offset), (void*)(odd_base+ source_offset), 
			o_dev->dev->clientfmt.width*2);
		dest_offset += o_dev->dev->clientfmt.width * 2;
		copy_to_user((void*)(buf +dest_offset),(void*)( even_base + source_offset),
			o_dev->dev->clientfmt.width*2);
		source_offset += gfx_get_vip_pitch();
        }

        return o_dev->dev->clientfmt.sizeimage;
}

/*---------------------------------------------------------------------
 * Description:
 *	configures the video device. this function is called automatically 
 *	after the device is registered.	
 * return:
 *	zero	  	
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int v4l2_init_done(struct v4l2_device *v)
{
	struct capture_device *dev = (struct capture_device *)v;
	debug_msg("v4l2_init_done()\n");
	/*  Initialize video input array	*/
#ifndef NO_DECODER	
	strcpy(dev->source[VSOURCE_COMP].input.name, "Composite");
	dev->source[VSOURCE_COMP].input.index = VSOURCE_COMP;
	dev->source[VSOURCE_COMP].input.type = V4L2_INPUT_TYPE_CAMERA;
	dev->source[VSOURCE_COMP].input.capability = 0;
	dev->source[VSOURCE_COMP].input.assoc_audio = FALSE;
    
	/*  Initialize video control properties	*/
	dev->source[VSOURCE_COMP].control[VCTRL_BRIGHTNESS] = 
                                capture_control[VCTRL_BRIGHTNESS].default_value;
	dev->source[VSOURCE_COMP].control[VCTRL_CONTRAST] = 
                                capture_control[VCTRL_CONTRAST].default_value;
	dev->source[VSOURCE_COMP].control[VCTRL_SATURATION] = 
                                capture_control[VCTRL_SATURATION].default_value;
 	dev->source[VSOURCE_COMP].control[VCTRL_HUE] = 
                                capture_control[VCTRL_HUE].default_value; 
 	dev->source[VSOURCE_COMP].control[VCTRL_WHITENESS] = 
                                capture_control[VCTRL_WHITENESS].default_value; 

	strcpy(dev->source[VSOURCE_SVIDEO].input.name, "S-Video");
	dev->source[VSOURCE_SVIDEO].input.index = VSOURCE_SVIDEO;
	dev->source[VSOURCE_SVIDEO].input.type = V4L2_INPUT_TYPE_CAMERA;
	dev->source[VSOURCE_SVIDEO].input.capability = 0;
	/*  Initialize video control properties	*/
	dev->source[VSOURCE_SVIDEO].control[VCTRL_BRIGHTNESS] = 
                                capture_control[VCTRL_BRIGHTNESS].default_value;
	dev->source[VSOURCE_SVIDEO].control[VCTRL_CONTRAST] = 
                                capture_control[VCTRL_CONTRAST].default_value;
	dev->source[VSOURCE_SVIDEO].control[VCTRL_SATURATION] = 
                                capture_control[VCTRL_SATURATION].default_value;
 	dev->source[VSOURCE_SVIDEO].control[VCTRL_HUE] = 
                                capture_control[VCTRL_HUE].default_value; 
 	dev->source[VSOURCE_SVIDEO].control[VCTRL_WHITENESS] = 
                                capture_control[VCTRL_WHITENESS].default_value; 

	strcpy(dev->source[VSOURCE_TUNER].input.name, "Tuner");
	dev->source[VSOURCE_TUNER].input.index = VSOURCE_TUNER;
	dev->source[VSOURCE_TUNER].input.type = V4L2_INPUT_TYPE_TUNER;
	dev->source[VSOURCE_TUNER].input.capability = V4L2_INPUT_CAP_AUDIO;
	dev->source[VSOURCE_TUNER].input.assoc_audio = TRUE;
	/*  Initialize video control properties	*/
	dev->source[VSOURCE_TUNER].control[VCTRL_BRIGHTNESS] = 
                                capture_control[VCTRL_BRIGHTNESS].default_value;
	dev->source[VSOURCE_TUNER].control[VCTRL_CONTRAST] = 
                                capture_control[VCTRL_CONTRAST].default_value;
	dev->source[VSOURCE_TUNER].control[VCTRL_SATURATION] = 
                                capture_control[VCTRL_SATURATION].default_value;
 	dev->source[VSOURCE_TUNER].control[VCTRL_HUE] = 
                                capture_control[VCTRL_HUE].default_value; 
 	dev->source[VSOURCE_TUNER].control[VCTRL_WHITENESS] = 
                                capture_control[VCTRL_WHITENESS].default_value; 
#endif
	strcpy(dev->source[VSOURCE_MPEG_DEC].input.name, "MPEG Dec.");
        dev->source[VSOURCE_MPEG_DEC].input.index = VSOURCE_MPEG_DEC;
        dev->source[VSOURCE_MPEG_DEC].input.type = V4L2_INPUT_TYPE_CAMERA;
        dev->source[VSOURCE_MPEG_DEC].input.capability = 0;
        dev->source[VSOURCE_MPEG_DEC].input.assoc_audio = FALSE;
	/*  Initialize the video decoder hardware	*/
#ifndef NO_DECODER	
	dev->videc.initialize(dev);
        g_video_buf.width = dev->videc.width[VIN_STD];
        g_video_buf.height = dev->videc.height[VIN_STD]*2;
	switch(VIN_STD)
        {
                case 1: /* NTSC */
                dev->videc.set_standard(dev, V4L2_STD_NTSC);
                break;
                case 2: /* PAL */
                dev->videc.set_standard(dev, V4L2_STD_PAL);
                break;
        }
	set_video_input(dev, VSOURCE_COMP);
#endif
	/*  Capture parameters  */
	dev->capture.capability = 0;
	dev->capture.capturemode = 0;
	dev->capture.extendedmode = 0;
	dev->capture.timeperframe = 0;

	/* Default capture dimensions */
	dev->clientfmt.width = 160;
	dev->clientfmt.height = 120;
	dev->clientfmt.depth = 16;
	dev->clientfmt.pixelformat = V4L2_PIX_FMT_UYVY;
	dev->clientfmt.flags = 0;
	dev->clientfmt.bytesperline = 0;
	dev->clientfmt.sizeimage = 0;

#ifndef EXCLUDE_TUNER
	dev->tuner.input = 0;
#ifdef CYGNUS
#ifdef CYGNUS_PAL_TUNER
	strcpy(dev->tuner.name,"Alps TMDZ2");
	strcpy(dev->tuner.std.name,"PAL");
	dev->tuner.std.framerate.numerator = 25;
	dev->tuner.std.framerate.denominator = 1;
	dev->tuner.std.framelines = 625;
	dev->tuner.std.colorstandard = V4L2_COLOR_STD_PAL;
	dev->tuner.std.colorstandard_data.ntsc.colorsubcarrier = V4L2_COLOR_SUBC_PAL;
	dev->tuner.std.transmission = 0;
	dev->tuner.std.reserved2 = 0;
	dev->tuner.capability = V4L2_TUNER_CAP_NORM
					| V4L2_TUNER_CAP_STEREO;	
        // TODO: Fill in the rest of the structures.                                
#else /* CYGNUS_NTSC_TUNER */
	strcpy(dev->tuner.name,"Alps TSBH1");
	strcpy(dev->tuner.std.name,"NTSC-M");
	dev->tuner.std.framerate.numerator = 30;
	dev->tuner.std.framerate.denominator = 1;
	dev->tuner.std.framelines = 525;
	dev->tuner.std.colorstandard = V4L2_COLOR_STD_NTSC;
	dev->tuner.std.colorstandard_data.ntsc.colorsubcarrier = V4L2_COLOR_SUBC_NTSC;
	dev->tuner.std.transmission = 0;
	dev->tuner.std.reserved2 = 0;
	dev->tuner.capability = V4L2_TUNER_CAP_NORM
					| V4L2_TUNER_CAP_STEREO;	
        // TODO: Fill in the rest of the structures.                                
#endif
#else   
	strcpy(dev->tuner.name,"Philips FM1236");
	strcpy(dev->tuner.std.name,"NTSC-M");
	dev->tuner.std.framerate.numerator = 30;
	dev->tuner.std.framerate.denominator = 1;
	dev->tuner.std.framelines = 525;
	dev->tuner.std.colorstandard = V4L2_COLOR_STD_NTSC;
	dev->tuner.std.colorstandard_data.ntsc.colorsubcarrier = V4L2_COLOR_SUBC_NTSC;
	dev->tuner.std.transmission = 0;
	dev->tuner.std.reserved2 = 0;
	dev->tuner.capability = V4L2_TUNER_CAP_NORM
					| V4L2_TUNER_CAP_STEREO;	

	dev->tuner.rangelow = 0x0010;/*BUGBUG: check data sheet */
	dev->tuner.rangehigh = 0xFFFF;/*BUGBUG: check data sheet */
	dev->tuner.rxsubchans = 0;
	dev->tuner.signal = 0;
	dev->tuner.afc = 0;
#endif /* CYGNUS */
#endif /* EXCLUDE_TUNER */
	return 0;
}


/*---------------------------------------------------------------------
 * Description:
 *	configures the capture device.
 * return:
 *	TRUE	
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int __init config_capture_device(struct capture_device *dev)
{
	sprintf( dev->shortname, "capture0" );

	sprintf(dev->v.name, "NSC Geode V4L2 Capture driver");
	dev->v.type = V4L2_TYPE_CAPTURE;
	dev->v.minor = VIDEO_MINOR;
	dev->v.open = v4l2_open;
	dev->v.close = v4l2_close;
	dev->v.read = v4l2_read;
	dev->v.write = v4l2_write;
	dev->v.ioctl = v4l2_ioctl;
	dev->v.mmap = v4l2_mmap;
	dev->v.poll = v4l2_poll;
	dev->v.initialize = v4l2_init_done;
	dev->v.priv = NULL;

	dev->videc.initialize = geode_capture_initialize;
	dev->videc.set_input = geode_capture_set_input;
	dev->videc.set_standard = geode_capture_set_standard;
	dev->videc.set_vcrmode = geode_capture_set_vcrmode;
	dev->videc.is_stable = geode_capture_is_stable;
	dev->input = 0;

	return TRUE;
}

/*---------------------------------------------------------------------
 * Description:
 *	Initialize the durango pointers
 * return:
 *  0 on success & -1 on faliure
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
int init_gfx(void)
{
        gfx_detect_cpu();
	gfx_detect_video();

        //      Map Durango pointers.
        gfx_virt_regptr = (unsigned char *)ioremap(0x40000000, 0x10000);
        if( gfx_virt_regptr == 0 )
        {
                err_msg( "failed to map gfx_virt_regptr\n" );
                return -1;
        }

        gfx_virt_vidptr = (unsigned char *)ioremap(0x40010000, 0x10000);
        if( gfx_virt_vidptr == 0 )
        {
                err_msg( "failed to map gfx_virt_vidptr\n" );
                return -1;
        }
	/* 4MB frame-buffer */
	gfx_virt_fbptr = (unsigned char *)ioremap(0x40800000, 0x400000);
        if( gfx_virt_fbptr == 0 )
        {
                err_msg( "failed to map gfx_virt_fbptr\n" );
                return -1;
        }
        gfx_virt_vipptr = ioremap( 0x40015000, 0x10000 );
        if( gfx_virt_vipptr == 0 )
        {
                err_msg( "failed to map gfx_virt_vipptr\n" );
                return -1;
        }
        return 0;
}


/*---------------------------------------------------------------------
 * Description:
 *	Cleanup the durango pointers. ( Un-map)
 * return:
 *  void
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
void cleanup_gfx(void)
{
        if( gfx_virt_regptr )
        {
                iounmap( (void*)gfx_virt_regptr );
                gfx_virt_regptr = NULL;
        }

        if( gfx_virt_vidptr )
        {
                iounmap( (void*)gfx_virt_vidptr );
                gfx_virt_vidptr = NULL;
        }

        if( gfx_virt_vipptr )
        {
                iounmap( (void*)gfx_virt_vipptr );
                gfx_virt_vipptr = NULL;
        }
	if( gfx_virt_fbptr )
        {
                iounmap( (void*)gfx_virt_fbptr );
                gfx_virt_fbptr = NULL;
        }

}


/*---------------------------------------------------------------------
 * Description:
 *	Initialize the VideoLib
 * return:
 *	Zero on success, -1 on failure.
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
static int init_videolib()
{
	if(g_VideoLib_initialized)return 0;
	debug_msg("Trying to VideoInit...");
        if( VL_VideoInit( &g_vs ) )
	{	
                err_msg("Error on VideoInit\n");
		return -1;
	}	
	debug_msg("ok\n");	
    gfx_set_display_pitch(g_GFX_pitch);
    gfx_set_compression_enable( 0 );
	VL_VideoSetDeinterlacingMethod(VL_VIDEO_WINDOW_FULL,
                                 DEINTERLACE);
    VL_VideoSetDeinterlacingMethod(VL_VIDEO_WINDOW_SCALED,
                                 DEINTERLACE);
	g_VideoLib_initialized = TRUE;
#if SIGMA_DIRECT
	if(!NO_DVIP)
	VL_VideoSetFullScreenSource(VL_VIDEO_SOURCE_DVIP_ONLY,
			get_genlock(0,VOUT_STD));	
#endif	
	VL_VideoSetWindow(&g_video_buf,0,0,0,0); 
#ifdef DEFAULT_OVERLAY_ALWAYS_ON        
	VL_VideoSetOverlay(VL_VIDEO_OVERLAY_ALWAYS_ON);		
        g_overlay_mode = VL_VIDEO_OVERLAY_ALWAYS_ON;
#else         
	VL_VideoSetOverlay(VL_VIDEO_OVERLAY_COLOR_KEY);		
        g_overlay_mode = VL_VIDEO_OVERLAY_COLOR_KEY;
#endif        
	VL_VideoEnable(1);
	VL_VideoPause();
	return 0;
}

/*---------------------------------------------------------------------
 * Description:
 *	Init module for the driver
 * return:
 *	Zero on success, -1 on failure.
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
int init_module(void)
#else
int __init geode_init_module(void)
#endif
{
	OS_Interface_init();
	err_msg(DESC_STR"\n");
	memset( &vbi_dev, 0, sizeof(struct vbi_device) );
	memset( &capture_dev, 0, sizeof(struct capture_device) );
	memset( &vout_dev, 0, sizeof(struct vout_device) );
#ifndef NO_DECODER	
	save_flags(flags);
	cli();
	if(tv_i2c_initialize())
	{	
		err_msg("couldn't init I2C bus\n");
		goto err_exit;
	}	
	info_msg("I2C Bus initialized\n");
	restore_flags(flags);
#endif	
	if(init_gfx())
	{
		err_msg("error on init_gfx\n");
		goto undo_i2c;
	}
	info_msg("Durango initialized\n");
	err_msg("Video Library Version %s\n", VL_Version());	
	err_msg("Durango version 2.33.01\n");
#ifndef NO_DECODER	
	save_flags(flags);
	cli();
	tv_saa7114_initialize_ntsc();
	
	restore_flags(flags);
#endif	
	g_GFX_pitch = gfx_get_display_pitch();
        switch(VOUT_STD)
        {
                case VSOURCE_STD_NTSC:
        		g_vs.TVStandard     = VL_TV_STANDARD_NTSC;
        		g_video_buf.address = NTSC_START_ADDRESS;
               	 	break;
                case VSOURCE_STD_PAL: 
                	g_vs.TVStandard = VL_TV_STANDARD_PAL;
                	g_video_buf.address = PAL_START_ADDRESS;
	                break;
                default:
                	info_msg("Not changing current video timings\n");
                	g_vs.TVStandard = VL_TV_STANDARD_NONE;
                	g_video_buf.address = OTHERS_START_ADDRESS;
			NO_DVIP = 1; 
        }
        g_vs.CRTEnable           = USE_CRT;
        g_vs.TimerIRQ            = IRQ;
        g_vs.SignalMonitorEnable = FALSE;
	g_vs.FullScreenSource = VL_VIDEO_SOURCE_MEMORY;

	switch(TV_OUT) {
		default: /* default to composite */
		case 1:
        	g_vs.TVOutput            = VL_TV_OUTPUT_COMPOSITE;break;
		case 2:
        	g_vs.TVOutput            = VL_TV_OUTPUT_S_VIDEO;break;
		case 3:
        	g_vs.TVOutput            = VL_TV_OUTPUT_YUV;break;
		case 4:
        	g_vs.TVOutput            = VL_TV_OUTPUT_SCART;break;
	}		
	switch(TV_ASPECT_RATIO) {
		case 1:
        	g_vs.TVAspectRatio = VL_TV_ASPECT_RATIO_GRAPHICS_ONLY; break;
		case 2:
        	g_vs.TVAspectRatio = VL_TV_ASPECT_RATIO_GRAPHICS_AND_VIDEO; break;
		case 3:
		default:
        	g_vs.TVAspectRatio = VL_TV_ASPECT_RATIO_VIDEO_ONLY;
	}
	if(!NO_DVIP)
		g_vs.GenlockDelay = get_genlock(VSOURCE_COMP,VOUT_STD);
	debug_msg("Genlock: %x\n",get_genlock(VSOURCE_COMP, VOUT_STD));
#ifdef SIGMA_DIRECT	
	g_vs.FieldPolarityInvert = TRUE;
	if(!NO_DVIP)
		g_vs.FullScreenSource = VL_VIDEO_SOURCE_DVIP_ONLY;
	g_vs.GenlockDelay = get_genlock(VSOURCE_COMP,VOUT_STD);
#else
	g_vs.FieldPolarityInvert = FALSE;
#endif		
	/* Configure capture device before VideoInit */
	config_capture_device( &capture_dev );
        if( init_videolib() )
	{	
                err_msg("Error on init_videolib\n");
		goto undo_gfx;
	}	

	config_vout_device( &vout_dev );
	config_vbi_device( &vbi_dev );
	vbi_dev.irq = VBI_IRQ;
	if( v4l2_register_device( (struct v4l2_device*)&vbi_dev ) != 0 )
	{
		info_msg("init_module():failed to register vbi dev\n");
		goto unreg_vbi;
	}
	vbi_dev.is_registered = 1;
	if( v4l2_register_device( (struct v4l2_device *)&capture_dev ) != 0 )
	{
		info_msg("init_module():failed to register capture dev\n");
		goto unreg_capture;
	}
	capture_dev.is_registered = 1;
	if( v4l2_register_device( (struct v4l2_device*)&vout_dev ) != 0 )
	{
		info_msg("init_module():failed to register vout dev\n");
		goto unreg_vout;
	}
	vout_dev.is_registered = 1;
	/* register the timer handler */
#ifndef NO_DECODER	/* We dont need signal monitoring when there is no
			  decoder */
	init_timer(&geode_timer);
	geode_timer.function = timer_handler;
	geode_timer.data = 0;
	geode_timer.expires = jiffies + HZ; /* one second */
#ifndef EXCLUDE_TUNER	
        tda9851_set_mode( TV_TDA9851_CTRL_STEREO );
#endif	
#endif	/* NO_DECODER */
	return 0;

unreg_vout:
	v4l2_unregister_device( (struct v4l2_device *)&vout_dev );
unreg_capture:
	v4l2_unregister_device( (struct v4l2_device *)&capture_dev );
unreg_vbi:
	v4l2_unregister_device( (struct v4l2_device *)&vbi_dev );
undo_gfx:
	cleanup_gfx();
undo_i2c:
#ifndef NO_DECODER	
	tv_i2c_cleanup();
#endif	
err_exit:	
	OS_Interface_close();
	return -1;
}

/*---------------------------------------------------------------------
 * Description:
 *	cleanup module for the driver
 * return:
 *  void
 * side-effects:
 * Notes:
 *----------------------------------------------------------------------*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
void cleanup_module(void)
#else
void __exit geode_cleanup_module(void)
#endif
{
#ifndef EXCLUDE_TUNER	
	tda9851_set_mode( TV_TDA9851_CTRL_MUTE );
#endif	
	signalISR(); /* To return to normal timings */
	VL_VideoEnd();
	//VL_VbiEnd();
	vbi_decode_cleanup();
#ifndef NO_DECODER	
	tv_i2c_cleanup();
#endif	
	g_timer_activated = 0;
#ifndef NO_DECODER	
	del_timer(&geode_timer);
#endif	
	OS_Interface_close();
	gfx_set_color_space_YUV(0);
	gfx_set_crt_enable(1);
	cleanup_gfx();
	v4l2_unregister_device( (struct v4l2_device *)&capture_dev );
	info_msg( "Removed device %s\n", capture_dev.shortname );
	v4l2_unregister_device( (struct v4l2_device *)&vbi_dev );
	info_msg( "Removed device %s\n", vbi_dev.shortname );
	v4l2_unregister_device( (struct v4l2_device *)&vout_dev );
	info_msg( "Removed device %s\n", vout_dev.shortname );
	memset( &vbi_dev, 0, sizeof(struct vbi_device) );
	info_msg( "cleanup_module(): done\n" );
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
/* Linux 2.4 specific macros which makes this code compilable into the kernel*/
module_init(geode_init_module);
module_exit(geode_cleanup_module);
#endif

