/*
 *
 * National Semiconductor Corporation Open Source License for 
 *
 *    Geode Video4linux2 driver
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c)  2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 *   * Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer. 
 *
 *   * Redistributions in binary form must reproduce the above 
 *     copyright notice, this list of conditions and the following 
 *     disclaimer in the documentation and/or other materials provided 
 *     with the distribution. 
 *
 *   * Neither the name of the National Semiconductor Corporation nor 
 *     the names of its contributors may be used to endorse or promote 
 *     products derived from this software without specific prior 
 *     written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER 
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE,
 * INTELLECTUAL PROPERTY INFRINGEMENT, OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF 
 * YOUR JURISDICTION. It is licensee's responsibility to comply with 
 * any export regulations applicable in licensee's jurisdiction. Under 
 * CURRENT (2001) U.S. export regulations this software 
 * is eligible for export from the U.S. and can be downloaded by or 
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed 
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran, 
 * Syria, Sudan, Afghanistan and any other country to which the U.S. 
 * has embargoed goods and services. 
 *
 */
 /* File Name    : geodedrv.h
 * File Contents: Structure definitions and function declarations for vout
 * and video device.
 *
 * Project      : GeodeV4l2 linux driver
 * Author       : Sojan James
 * Date         :
 * ---------------------------------------------------------------------*/
#ifndef _GEODEDRV_H_
#define _GEODEDRV_H_

/* System includes */
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/poll.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/videodev.h>
#include <linux/version.h>
#include <asm/uaccess.h>
#include <linux/init.h>

/* The ioctl definitions and shared structure definitions */
#include "ioctl_geode_v.h"

/* TVlib includes */
#ifndef NO_DECODER
#include "tv_debug.h"
#include "tv_tuner.h"
#include "tv_saa7114.h"
#include "tv_i2c.h"
#include "tv_tda9851.h"
#include "tv_os_geode.h"
#endif /* NO_DECODER */

/* Durango includes */
#include "gfx_rtns.h"
#include "gfx_defs.h"

/* VideoLib includes */
#include <vl_rtns.h>
#include <vl_defs.h>


#define PKMOD "geode_v: "
#ifdef  DEBUG
#define debug_msg(fmt,arg...) printk(KERN_ERR PKMOD fmt,##arg)
#else
#define debug_msg(fmt,arg...)
#endif
#if 1
#define err_msg(fmt,arg...) printk(KERN_ERR PKMOD fmt,##arg)
#else
#define err_msg(fmt,arg...)
#endif
#if 1
#define info_msg(fmt,arg...) printk(KERN_INFO PKMOD fmt,##arg)
#else
#define info_msg(fmt,arg...)
#endif

#define GEODE_IOCTL_DEBUG_PHYS_READ		0x00000000
#define GEODE_IOCTL_DEBUG_PHYS_WRITE		0x00000001
#define GEODE_IOCTL_DEBUG_I2C_READ		0x00000000
#define GEODE_IOCTL_DEBUG_I2C_WRITE		0x00000001
#define GEODE_IOCTL_DEBUG_PORTIO_READ_8		0x00000000
#define GEODE_IOCTL_DEBUG_PORTIO_READ_16	0x00000001
#define GEODE_IOCTL_DEBUG_PORTIO_READ_32	0x00000002
#define GEODE_IOCTL_DEBUG_PORTIO_WRITE_8	0x00000003
#define GEODE_IOCTL_DEBUG_PORTIO_WRITE_16	0x00000004
#define GEODE_IOCTL_DEBUG_PORTIO_WRITE_32	0x00000005


/*	Values for type field		*/
#define DEVICE_TYPE_0			0
/*	Extreme video dimensions	*/
#define MIN_WIDTH			32
#define MIN_HEIGHT			24
#define MAX_WIDTH			720
#define MAX_HEIGHT			576
#define MAX_FRAME_AGE			200

#define MAX_OPENS			5

typedef enum VideoControls_t {
	VCTRL_BRIGHTNESS,	
	VCTRL_CONTRAST,	
	VCTRL_SATURATION,
	VCTRL_HUE,	
	VCTRL_WHITENESS,
	MAXCONTROLS} VideoControls_t;	

/* forward references */
struct vbi_device;
struct capture_device;
struct vout_device;

struct video_decoder
{
	int	is_initialized;

	int 	hskip[VSOURCE_STD_MAX];
	int	vskip[VSOURCE_STD_MAX];
	int	width[VSOURCE_STD_MAX];
	int	height[VSOURCE_STD_MAX];

	int     vbi_hskip[VSOURCE_STD_MAX];
	int     vbi_vskip[VSOURCE_STD_MAX];
	int 	vbi_width[VSOURCE_STD_MAX];
	int 	vbi_height[VSOURCE_STD_MAX];	
	
	int	preferred_field;

	int	num_inputs;
	int	input;
	__u32	standards;
	__u32	standard;
	__u32	frame_period;
	int	decoder_is_stable;
	__u32	decoder_stable_time;

	int	(*initialize)(struct capture_device *dev);
	int	(*set_input)(struct capture_device *dev, int x);
	int	(*set_standard)(struct capture_device *dev, int x);
	int	(*set_vcrmode)(struct capture_device *dev, int x);
	int	(*is_stable)(struct capture_device *dev);
};

struct video_source
{
	struct v4l2_input	input;
	int			control[MAXCONTROLS];
	struct v4l2_tuner	tuner;
	int			vcrmode;
};


struct capture_device_open
{
	int			isopen;
	int			noncapturing;
	struct capture_device	*dev;
};

struct vout_device_open
{
	int			isopen;
	int			noncapturing;
	struct vout_device	*dev;
};

struct capture_device
{
	struct v4l2_device	v;	/*  Must be first */
	char			shortname[16];
	int			is_registered;
	int			open_count;
	struct capture_device_open	open_data[MAX_OPENS];
	int 			capturing_opens;
/*	Per-bus index number for each device	*/
	int			index;
/*	General type of device  */
	int			type;

/*	Video decoder stuff		*/
	struct video_decoder	videc;
	struct video_source	source[VSOURCE_COUNT];
	int			source_width;
	int			source_height;


/*	Client capture format and capture modes	*/
	struct v4l2_pix_format	clientfmt;
	struct v4l2_captureparm	capture;
#ifndef EXCLUDE_TUNER
	struct v4l2_tuner		tuner;
#endif
	int			input;/* which video source is selected */
};

struct vout_device
{
	struct v4l2_device	v;	/*  Must be first */
	char			shortname[16];
	int			is_registered;
	struct vout_device_open	open_data[MAX_OPENS];
	int 			capturing_opens;
	int			open_count;

	struct v4l2_pix_format	clientfmt;
	__u32			standards;
	__u32			standard;
	
};


#endif	// GEODEDRV_H

