/* NSC_LIC_BSD
 *
 * Revision 1.2
 *
 * National Semiconductor Corporation Open Source License for
 *
 *    TVLIB
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c) July 19, 2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use by Licensee of this Software in source and
 * binary forms, with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   * Redistributions in binary form must either:
 *     (a) reproduce the above copyright notice, this list of conditions
 *         and the following disclaimer in the documentation and/or
 *         other materials provided with the distribution, or
 *     (b) if Licensee elects to redistribute any Software not in
 *         accordance with (a), then Licensee shall indemnify National
 *         Semiconductor Corporation and all copyright holders and
 *         contributors for any claims related to such Software (and pay
 *         all costs and damages, including all attorney fees)
 *
 *   * Neither the name of the National Semiconductor Corporation nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF
 * YOUR JURISDICTION. It is Licensee's responsibility to comply with
 * any export regulations applicable in Licensee's jurisdiction. Under
 * CURRENT (July 19, 2001) U.S. export regulations this software
 * is eligible for export from the U.S. and can be downloaded by or
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran,
 * Syria, Sudan, Afghanistan and any other country to which the U.S.
 * has embargoed goods and services.
 *
 * END_NSC_LIC_BSD */

/*
 * This file contains routines to use the I2C bus.
 *
 * I2C base address detection, initialization, and cleanup functions:
 * 
 *    i2c_get_acb_base_addresses
 *    tv_i2c_initialize
 *    tv_i2c_cleanup
 *
 * Linux system (kernel I2C support) I2C functions:
 *
 *    tv_i2c_attach_adapter
 *    tv_i2c_detach_client
 *    tv_i2c_detect_client
 *
 * I2C access functions:
 *
 *    tv_i2c_scan
 *    tv_i2c_cx241xx_read_data
 *    tv_i2c_cx241xx_write_data
 *    tv_i2c_scart_write_data
 *    tv_i2c_saa7114_read_byte_data
 *    tv_i2c_saa7114_write_byte_data
 *    tv_i2c_tuner_read_byte
 *    tv_i2c_tuner_write_byte
 *    tv_i2c_tuner_write_byte_data
 *    tv_i2c_tuner_write
 *    tv_i2c_tda9851_read_byte
 *    tv_i2c_tda9851_write_byte
 *    tv_i2c_initialize
 *    tv_i2c_cleanup
 *
 * I2C utility functions:
 *
 *    acb_start_i2c_clock
 *    acb_stop_i2c_clock
 *    acb_write_bus_i2c 
 *    acb_write_i2c
 *    acb_start 
 *    acb_stop  
 *    acb_write_byte_i2c
 *    acb_reset_i2c
 *    acb_request_master
 *    acb_send_address
 *    timeout          
 *    start_timer     
 *    acb_abort_data
 *    acb_bus_recovery
 *    acb_stall_after_start
 *    acb_ack
 *    acb_read_byte_i2c 
 *    acb_read_i2c
 *    acb_read_bus_i2c
 *
 * Project: TVLIB
 * Author:  Peter Lohmann
 * Date:    April 10, 2001
 *---------------------------------------------------------------------*/


#ifndef TV_I2C_H
#define TV_I2C_H

#ifdef LINUX
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif
#define __NO_VERSION__ 
#include <linux/module.h>
#include <linux/version.h>
#include <asm/io.h>
#elif defined(WINCE)
#include <windows.h>
#include <stdlib.h>
#elif defined(WIN98)
#elif defined(VXWORKS)
#include <vxWorks.h>
#include <intLib.h>
#include <iv.h>
#include <drv/pci/pciConfigLib.h>
#include <sysLib.h>
#include <rngLib.h>
#endif

#include "tv_os_geode.h"

#ifdef __cplusplus
extern "C" {
#endif


#ifdef CYGNUS
#ifdef XPRESSROM
#define GEODE_ACB2_BASE_ADDR	0x840
#define GEODE_ACB_BASE_ADDR	0x820
#else
#define GEODE_ACB2_BASE_ADDR	0x820
#define GEODE_ACB_BASE_ADDR	0x810
#endif
#else
#define GEODE_ACB2_BASE_ADDR	0x820
#define GEODE_ACB_BASE_ADDR	GEODE_ACB2_BASE_ADDR
#endif

#define GEODE_ACBSDA			0
#define GEODE_ACBST				1
#define GEODE_ACBCST			2
#define GEODE_ACBCTL1			3
#define GEODE_ACBADDR			4
#define GEODE_ACBCTL2			5

#define DEC_WRITE_ADR			0x42
#define DEC_READ_ADR			0x43

#define CX241XX_WRITE_ADR		0xAA
#define CX241XX_READ_ADR			0xAB

#define TDA9851_WRITE_ADR		0xB6
#define TDA9851_READ_ADR			0xB7

#define SCART_WRITE_ADR			0x96
#define SCART_READ_ADR			0x97

void tv_i2c_get_acb_base_addresses( WORD* acb1, WORD* acb2 );

BYTE tv_i2c_scart_write_data( BYTE ucVal1, BYTE ucVal2 );
int tv_i2c_cx241xx_read_data( BYTE ucReg, BYTE* pucData );
//BYTE tv_i2c_cx241xx_read_data( BYTE ucReg );
int tv_i2c_cx241xx_write_data( BYTE ucReg, BYTE ucVal );
//BYTE tv_i2c_cx241xx_write_data( BYTE ucReg, BYTE ucVal );
BYTE tv_i2c_saa7114_read_byte_data( BYTE reg );
BYTE tv_i2c_saa7114_write_byte_data( BYTE reg, BYTE value );
BYTE tv_i2c_tuner_read_byte( BYTE tuner_i2c_read_addr );
BYTE tv_i2c_tuner_write_byte( WORD value );
BYTE tv_i2c_tuner_write_byte_data( BYTE reg, WORD value );
BYTE tv_i2c_tuner_write( BYTE tuner_i2c_write_addr, WORD divider, WORD control_data );
BYTE tv_i2c_tda9851_read_byte();
BYTE tv_i2c_tda9851_write_byte( BYTE value );


int tv_i2c_initialize();
int tv_i2c_cleanup();

#ifdef USE_SYSTEM_I2C  

#include <linux/i2c.h>

extern struct i2c_client*	g_client_tuner; 
extern struct i2c_client*	g_client_decoder; 
extern struct i2c_client*	g_client_tda9851; 

#else	// !USE_SYSTEM_I2C

extern WORD	geode_acb1_base_addr;
extern WORD	geode_acb2_base_addr;

extern unsigned long	tv_i2c_timeout_cnt;


#define	MAX_COMMAND_TIMEOUT	( 5*1000)
extern ULONG	start_time;

void tv_i2c_scan();

void	acb_start_i2c_clock( WORD );
void	acb_stop_i2c_clock( WORD );

int 	acb_write_bus_i2c( WORD, BYTE, BYTE );
int	acb_write_i2c_verify( WORD, BYTE, BYTE, BYTE );
int	acb_write_i2c( WORD, BYTE, BYTE, BYTE );

void	acb_start( WORD );
void	acb_stop( WORD );
int	acb_write_byte_i2c_verify( WORD, BYTE );
void	acb_write_byte_i2c( WORD, BYTE );
void	acb_reset_i2c( WORD );
int	acb_request_master( WORD );		
int	acb_send_address_verify( WORD, BYTE );
void	acb_send_address( WORD, BYTE cData );	

int	acb_timeout();
void acb_start_timer();
void acb_abort_data( WORD );
void acb_bus_recovery( WORD );
void acb_stall_after_start( WORD, int state );
int acb_ack( WORD, int fPut, int negAck );
BYTE acb_read_byte_i2c( WORD );
int acb_read_i2c_verify( WORD bus, BYTE chipadr, BYTE subadr, BYTE* pucData );
BYTE acb_read_i2c( WORD, BYTE chipadr, BYTE subadr );
BYTE acb_read_bus_i2c( WORD, BYTE chipadr );

#endif

#ifdef __cplusplus
}
#endif


#endif	// TV_I2C_H




