/*
 *
 * National Semiconductor Corporation Open Source License for 
 *
 * National Semiconductor IR Drivers
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c)  2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 *   * Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer. 
 *
 *   * Redistributions in binary form must reproduce the above 
 *     copyright notice, this list of conditions and the following 
 *     disclaimer in the documentation and/or other materials provided 
 *     with the distribution. 
 *
 *   * Neither the name of the National Semiconductor Corporation nor 
 *     the names of its contributors may be used to endorse or promote 
 *     products derived from this software without specific prior 
 *     written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER 
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE,
 * INTELLECTUAL PROPERTY INFRINGEMENT, OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF 
 * YOUR JURISDICTION. It is licensee's responsibility to comply with 
 * any export regulations applicable in licensee's jurisdiction. Under 
 * CURRENT (2001) U.S. export regulations this software 
 * is eligible for export from the U.S. and can be downloaded by or 
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed 
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran, 
 * Syria, Sudan, Afghanistan and any other country to which the U.S. 
 * has embargoed goods and services. 
 *
 * File Name	  : nscir.c						*
 *									*
 * Project Name	  : ASTB						*
 *									*
 * Module Name 	  : NSCIR						*
 *									*
 * Author Name	  : Arun Kumar, Niraj Kumar				*
 *									*
 * Language	  : C							*
 *									*
 * Date		  : 08/10/01						*
 *									*
 * Description	  : This file contains the basic IR driver, which will  *
 * 		    receive the raw IR data and call the specific       *
 * 		    protocol decoding for that.				*
 *									*
 ************************************************************************/


#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/io.h>			// inb and outb
#include <linux/tqueue.h>	// for bottom-half
#include <linux/interrupt.h>
#include <linux/sched.h>	// for current
#include <linux/poll.h>		// poll method

#include "nscir.h"	// ioctl codes and major number

#define VERSION "1.0"

#define TWIRP_PROTOCOL 1
#define SEJIN_PROTOCOL 2
#define RC5_PROTOCOL 3
#define BYTE unsigned char
#define MAX_PROTO 8

// Various types of debug information
#define REPORT_RAW			0
#define REPORT_EDGES 			0
#define REPORT_RAW_PACKET		0
#define REPORT_STATUS			0
#define REPORT_TWIRP_DECODE		0
#define REPORT_COOKED_DATA		0


typedef int count_t;

#define TRUE 	1
#define FALSE	0


int NscIrUnRegProto(int);
int NscIrRegProto(int,int (*)(),int (*)(),int (*)());

EXPORT_SYMBOL_NOVERS(NscIrRegProto);
EXPORT_SYMBOL_NOVERS(NscIrUnRegProto);


typedef struct  
{	int 	count;
 	BYTE	protoMap;
	int (*proto_twirp_write)();
	int (*proto_init[MAX_PROTO])();
	int (*proto_decode[MAX_PROTO])();
	
}NscIrTest;
static NscIrTest NscIrReg;


/*
   While hardcoding the IO_BASE may seem a poor idea, the SuperI/O
   overloads IR with COM3 (base=0x3e8, IRQ4), so there is no choice
   for the user.

   3f8, irq4:  ttys0
   2f8, irq3:  ttys1
   3e8, irq4:  ttys2
   2e8, irq3:  ttys3

   Setup in the BIOS.
*/
#define IO_BASE 0x2e8
#define IO_IRQ  7

int PROTOCOL_ID;
MODULE_PARM(PROTOCOL_ID,"i");
//--------------for sejin-------
#define MAX_SJ_BUF 1024
static unsigned char sejinRawBuf[MAX_SJ_BUF];
static int    sjnWIndex=0, sjnRIndex=0;
static int   sjnDataCount=0  ;

//--------------for RC5-------
#define MAX_RC5_BUF 1024
static unsigned char rc5RawBuf[MAX_RC5_BUF];
static int    rc5WIndex=0, rc5RIndex=0;
static int   rc5DataCount=0  ;
//------------------------------

/* some offsets of import */
const int RXD_REG = IO_BASE + 0x00;	// Receive data
const int IER_REG = IO_BASE + 0x01;	// Interrupt enable register
const int FCR_REG = IO_BASE + 0x02;	// FIFO control register offset
const int BSR_REG = IO_BASE + 0x03;	// bank select register offset
const int MCR_REG = IO_BASE + 0x04;	// Modem/Mode control register
const int LSR_REG = IO_BASE + 0x05;	// Link status register
const int ASCR_REG = IO_BASE + 0x07;	// auxiliary status/control register

// Bank 2 Registers
const int BGDL_REG = IO_BASE + 0x00;	// Low byte of Baud Rate divisor
const int BGDH_REG = IO_BASE + 0x01;	// High byte of Baud Rate divisor
const int EXCR1_REG = IO_BASE + 0x02;	// Extended Control Register 1
const int EXCR2_REG = IO_BASE + 0x04;	// Extended Control Register 2

// Bank 3 Registers
const int MID_REG = IO_BASE + 0x00;	// Module ID register

// Bank 5 Registers
const int IRCR2_REG = IO_BASE + 0x04;

// Bank 7 Registers
const int IRRXDC_REG = IO_BASE + 0x00;	// Receiver Demodulator Control
const int RCCFG_REG = IO_BASE + 0x02;	// RC configuration
const int IRCFG4    = IO_BASE + 0x07;
const int IRCFG1    = IO_BASE + 0x04;

static char cur_bank = 0;

#define	PRESCALER_13		0x00
#define PRESCALER_1_625		0x01
#define PRESCALER_1		0x11



/* local routines */
static void init_superio (void);		/* init SuperI/O */
static void set_extended_mode (void);
static void stop_reception (void);
static void set_bank (unsigned char bank);
static void set_baudrate (unsigned char BaudRatePrescale, 
	unsigned short BaudRateDivisor);

static void interrupt_handler (int irq, void *dev_id, struct pt_regs *regs);
static void install_handler (void);
static void remove_handler (void);

static void set_demodulator (unsigned char freq_code, unsigned char bw_code);
static void set_fifo_size (void);
static void clear_rx_fifo (void);
static void setup_ceir (void);

void switch_proto( int);
static void stop_and_flush (void);
void nscir_protocol_init(void);


static char proto_name[4][8]={"TWIRP","SEJIN","RC5","NULL"};
// number of minor devices open, controls handler registration
static unsigned char open_count = 0;


static struct tq_struct twirp_task,sejin_task,rc5_task;

int nscir_init(void);
static unsigned int iii=0;



int NscIrUnRegProto(int protocol_id)
{
	int i;


	if( NscIrReg.protoMap & (1<<(protocol_id-1)) )
	{
		if(PROTOCOL_ID  == protocol_id )
		{
			if((i=get_next_proto())!=0)
				switch_proto(i);
		}

		NscIrReg.count--;
		
		if(NscIrReg.count==0)
		{
			PROTOCOL_ID =0;
			remove_handler();

		}
		NscIrReg.protoMap &= (~(1<<(protocol_id-1))); // mask ,disable
		NscIrReg.proto_init[protocol_id-1] = 0 ;
		NscIrReg.proto_decode[protocol_id-1] = 0 ;
		return 1;	
	}
	else return 0;

}

int NscIrRegProto(int protocol_id, int (*init_proto)(),int (*decode_proto)(),int (*spl_proto)())
{
        
	if( NscIrReg.protoMap & (1<<(protocol_id-1)))
	{  printk (" Protocol is already installed \n");
	   return 0;
	}
	if((protocol_id > MAX_PROTO) || (protocol_id<=0))
	{
		printk(" Unknown protocol ");
		return 0;
	}
	

	if( ( init_proto != 0 ) && ( decode_proto != 0 ) )
	{
		
		NscIrReg.protoMap |= ( 0x01 << (protocol_id-1) );
		NscIrReg.proto_init[protocol_id-1] = init_proto ;
		NscIrReg.proto_decode[protocol_id-1] = decode_proto ;

		if(protocol_id == 1)
		 {	// because twirp implementation different
			  NscIrReg.proto_twirp_write = spl_proto;
		  
		 }
		switch_proto( protocol_id );
		NscIrReg.count++;
		return 1;
	}
        printk("Illegal Registartion \n"); 
	return 0;
}










int get_next_proto()
{
	int i=0,prev=0;
	unsigned char	gen;
	
	if(PROTOCOL_ID ==0) return 0;
        gen = NscIrReg.protoMap;
	gen = gen >> PROTOCOL_ID; 
	for(i=PROTOCOL_ID;i<MAX_PROTO;i++)
		{
		   	if(gen  & 0x01) return  i+1;
			gen >>= 1;
		}
		
        gen = NscIrReg.protoMap;
	gen  <<= (MAX_PROTO-PROTOCOL_ID)+1;
	
	for(i=0; i<PROTOCOL_ID;i++)
		{
			if( gen  & 0x80) prev =i+1;
			gen <<= 1;
		}

	if(prev) 	return (PROTOCOL_ID -(prev));
		else 	return 0;	
}




/************************************************************************
 * Syntax	: nscir_open                                            *
 *                                                                     	*
 * Description	: This function opens the basci IR driver.             	*
 *                                                                     	*
 * Parameters	: inode                                                	*
 *		    - The inode structure for the device special file.  *
 *                file                                                 	*
 *		    - The device-special file.                         	*
 *                                                                     	*
 * Return 	: 0 on sucess, error value on failure.                 	*
 *                                                                     	*
 * Comment	: During open, the queues for the appropriate device   	*
 *		  are reset.                                           	*
 *                                                                     	*
 ************************************************************************/

static int nscir_open (struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR(file -> f_dentry -> d_inode -> i_rdev);

	MOD_INC_USE_COUNT;

	if (open_count == 0)
	{
//		install_handler ();
		printk("National IR Driver: %s\n", VERSION);
	}
	open_count++;
//	printk("\n Mask Value : %x %x \n", inb(0x21),inb(0x20));
	return 0;
}



/************************************************************************
 * Syntax	: nscir_release                                         *
 *                                                                     	*
 * Description	: This function close the basci IR driver.             	*
 *                                                                     	*
 * Parameters	: inode                                                	*
 *		    - The inode structure for the device special file. 	*
 *                file                                                 	*
 *		    - The device-special file.                         	*
 *                                                                     	*
 * Return 	: 0 on sucess, error value on failure.                 	*
 *                                                                     	*
 * Comment	: None.	   						*
 *                                                                     	*
 ************************************************************************/

static int nscir_release (struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;

	open_count--;

	return 0;
}

/************************************************************************
 * Syntax	: nscir_ioctl	                                        *
 *                                                                     	*
 * Description	: Ioctls for the device.		.             	*
 *                                                                     	*
 * Parameters	: inode                                                	*
 *		    - The inode structure for the device special file. 	*
 *                file                                                 	*
 *		    - The device-special file.                         	*
 *		  cmd							*
 *		    - Command code.					*
 *		  arg							*
 *		    - argument to command.				*
 *                                                                     	*
 * Return 	: 0 on sucess, error value on failure.                 	*
 *                Error values are,					*
 *		    EINVAL: ioctl command invalid, or NULL passed	*
 *		    EFAULT: arg contains a bad address.			*
 *                                                                     	*
 * Comment	: None.	   						*
 *                                                                     	*
 ************************************************************************/

static int nscir_ioctl (struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
       return 0;
}

/************************************************************************
 * Syntax	: nscir_poll                                            *
 *                                                                     	*
 * Description	: 					             	*
 *                                                                     	*
 * Parameters	: file                                                	*
 *		    - The device-special file.  			*
 *                wait                                                 	*
 *		    - The poll table.	                         	*
 *                                                                     	*
 * Return 	: Bitmask of permitted I/O operations.                 	*
 *		  ENODEV if file points to an invalid minor number.	*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/

static unsigned int nscir_poll (struct file *file, poll_table *wait)
{
	return (-ENODEV);
}



/***************************************************************
 * File operations table, contains entry points into driver.
 ***************************************************************/
static struct file_operations nscir_fops =
{
	poll:		nscir_poll,
	ioctl:		nscir_ioctl,
	open:		nscir_open,
	release:	nscir_release,
};



#ifdef MODULE

/************************************************************************
 * Syntax	: init_module                                           *
 *                                                                     	*
 * Description	: This function loads the driver as a module in the    	*
 *		  Linux kernel.						*
 *                                                                     	*
 * Parameters	: None.                                                	*
 *                                                                     	*
 * Return 	: 0 on success, errno on failure.	               	*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/

int init_module (void)

{
	int i;

/*	if( (PROTOCOL_ID < 1) || (PROTOCOL_ID >2) )
	{ 
		PROTOCOL_ID = 1;
		printk(" \n specified  PROTOCOL_ID is not supported \n");
	}
	printk("\n Activated with '%s' IR keyboard \n",
                                              proto_name[PROTOCOL_ID-1]);

*/
	PROTOCOL_ID=0;

	// initialization of the protocol registration structure begins
	NscIrReg.count=0;
	NscIrReg.protoMap=0;
	
	for(i=0;i<MAX_PROTO;i++)
	{
		NscIrReg.proto_init[i]= 0;
		NscIrReg.proto_decode[i]=0;
	}
	
	// initialization of the protocol registration structure ends
	
	outb(07, 0x015c)  ; outb(02,0x015d);

	return (nscir_init ());
}


/************************************************************************
 * Syntax	: cleanup_module                                        *
 *                                                                     	*
 * Description	: This function unloads the driver module from the    	*
 *		  Linux kernel.						*
 *                                                                     	*
 * Parameters	: None.                                                	*
 *                                                                     	*
 * Return 	: None.	               					*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/

void cleanup_module (void)
{
	if (unregister_chrdev (NSCIR_MAJOR, "nscir"))
	{
		printk ("nscir:  error unregistering driver\n");
	}
	else
		remove_handler ();
}

#endif



/************************************************************************
 * Syntax	: nscir_init		                                *
 *                                                                     	*
 * Description	: This function initialize the driver, called during  	*
 *		  driver load.						*
 *                                                                     	*
 * Parameters	: None.                                                	*
 *                                                                     	*
 * Return 	: 0 on success, -EIO on error.				*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/

int nscir_init(void)
{
	// register the character device
	if (register_chrdev (NSCIR_MAJOR, "nscir", &nscir_fops))
	{
		printk ("nscir:  Fail to register IR device\n");
		return -EIO;
	}
	nscir_protocol_init();
	return 0;
}


/************************************************************************
 * Syntax	: nscir_protocol_init	                                *
 *                                                                     	*
 * Description	: This function initialize the specific protocol  	*
 *		  docoding based on the protocol id.			*
 *                                                                     	*
 * Parameters	: None.                                                	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/


void	nscir_protocol_init(void)
{
	// initialize National's super i/o
	init_superio ();
	// empty IR input queue

	switch(PROTOCOL_ID)
	{
	case   TWIRP_PROTOCOL:
	//	raw_ir_q_init(); 
		NscIrReg.proto_init[0]();
		break;
	case   SEJIN_PROTOCOL:
		//raw_ir_q_init(); 
		// NscCirInitializeDetection();
		NscIrReg.proto_init[1]();
		break;
	case   RC5_PROTOCOL:
		//raw_ir_q_init(); 	
		//NscCirInitializeDetectionRC5();	
		NscIrReg.proto_init[2]();
		break;	
	default: break;
	}
//	install_handler ();
}	



/************************************************************************
 * Syntax	: switch_proto		                                *
 *                                                                     	*
 * Description	: This function is used to make the control to switch 	*
 *		  between differnets protocols.			  	*
 *                                                                     	*
 * Parameters	: proto_id                                             	*
 *	 	    - Protocol id (To identify the different protocols).*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/


void switch_proto( int proto_id)
{
	printk("\n  Switching to %s protocol\n",proto_name[proto_id-1]);
	
	if(NscIrReg.count) remove_handler();
		
	PROTOCOL_ID = proto_id;
	nscir_protocol_init();

	install_handler();
	 
}



/************************************************************************
 * Syntax	: init_superio		                                *
 *                                                                     	*
 * Description	: This function is used to initialize the registrers in	*
 *		  the Super I/O.					*
 *                                                                     	*
 * Parameters	: None.                                             	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/

static void init_superio (void)
{
	set_bank (0);	// force the UART into bank 0

	set_extended_mode ();
	stop_reception ();
	clear_rx_fifo ();

	// Set the prescaler to 13, and the divisor is set to 1.  
	// This yields 115200bps with a 0.16% error.  
	// This is around a 3x oversampling of the IR signal.
	// Note: The sampling rate seems to depend solely on the divisor, and
	// not on the prescaler.  Thus, a divisor of 1 is the best we are
	// likely to do.
//	set_baudrate (PRESCALER_1_625, 1);  // baud 7.2k
	switch(PROTOCOL_ID)
	{
	case TWIRP_PROTOCOL:
		set_baudrate (PRESCALER_13, 1);  // baud 115.2k
		break;
	case SEJIN_PROTOCOL:
		set_baudrate (PRESCALER_13, 3);  // baud 115.2k
		break;
	case RC5_PROTOCOL:
		set_baudrate (PRESCALER_13, 32);	
		break;	
	default :
		break;
	}
	set_fifo_size ();	// reset FIFOs to 16 deep

	// The internal demodulator isn't used.  The following call sets
	// the freq to 35.7-39.5kHz, if the demodulator is later used.
//	 set_demodulator (0xb, 0x1);
//	 set_demodulator (0xb, 0x2);

	setup_ceir();
	stop_reception ();
}



/************************************************************************
 * Syntax	: set_extended_mode	                                *
 *                                                                     	*
 * Description	: Set the UART into extended mode, turns off loopback	*
 *		  and sets CIR mode..					*
 *                                                                     	*
 * Parameters	: None.                                             	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/

static void set_extended_mode (void)
{
	unsigned char value;

	if (cur_bank != 2)
		set_bank (2);

	value = inb (EXCR1_REG);
	value |= 0x01;			// set extended mode
	value &= ~0x10;			// turn off loopback
	outb (value, EXCR1_REG);

	// set CEIR mode
	if (cur_bank != 0)
		set_bank (0);
	outb (0xc0, MCR_REG);
}




/************************************************************************
 * Syntax	: stop_reception	                                *
 *                                                                     	*
 * Description	: This function stops sampling the receiver [sets the 	*
 *		  RXACT bit in the ASCR (bank 0,register 7, bit 5) to 0,*
 *		  disabling the sampling]. 				*
 *                                                                     	*
 * Parameters	: None.                                             	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: Called during initialization and at the end of a 	*
 *		  packet. This shuts off the sampling process until IR  *
 *		  energy is detected again.				*
 *                                                                     	*
 ************************************************************************/

static void stop_reception (void)
{
	const unsigned char RXACT = 0x20;
	unsigned char value;

	if (cur_bank != 0)
		set_bank (0);

	// set the RXACT bit in the ASCR
	value = inb (ASCR_REG);

	value |= RXACT;
	outb (value, ASCR_REG);
}



/************************************************************************
 * Syntax	: clear_rx_fifo		                                *
 *                                                                     	*
 * Description	: This function clears the receive FIFO [ Sets the RXSR *
 *		  bit in the FCR (bank 0, register 2, bit 1) to 1,	*
 *		  resetting the receiver.				*
 *                                                                     	*
 * Parameters	: None.                                             	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: Performs a soft reset on the receiver, clearing the 	*
 *		  FIFO. Also resets the threshold to 1 for both FIFOs.	*
 *                                                                     	*
 ************************************************************************/

static void clear_rx_fifo (void)
{
	const unsigned char RXSR = 0x02;

	if (cur_bank != 0)
		set_bank (0);

	// this register is write-only, so just set the desired bit
	outb (RXSR, FCR_REG);
}




/************************************************************************
 * Syntax	: set_baudrate		                                *
 *                                                                     	*
 * Description	: This function  sets the baudrate prescale and divisor *
 *		  value. 						*
 *                                                                     	*
 * Parameters	: BaudRatePrescale                                      *
 *			- Encoded prescaler select values, placed in 	*
 *			  bits 5 & 4 of the EXCR2 register.		*
 *				0x00 : 13.0				*
 *				0x01 : 1.625				*
 *				0x11 : 1.0				*
 *		  BaudRateDivisor					*
 *			- This is 16 bit value, which together with the *
 *			  Prescaler, this determines the baud rate.	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: The BGD registers are in bank 2 (register 0 & 1) and	*
 *		  the EXCR2 registers is in Bank 2 (register 2).	*
 *                                                                     	*
 ************************************************************************/

static void set_baudrate (unsigned char BaudRatePrescale, unsigned short BaudRateDivisor)
{
	unsigned char value;
     //  printk(" set baud \n");

	if (cur_bank != 2)
		set_bank (2);

	value = inb (EXCR2_REG);
	// Clear the following bits:
	//	4-5:  clear prescaler select
	//	6:    reserved, set to 0
	//	7:    lock bit, set to 0 when extended mode selected
	value &= 0x0F;
	value |= (BaudRatePrescale << 4);
	outb (value, EXCR2_REG);
        

	// Load the Baud Rate Divisor
	value = BaudRateDivisor >> 8;	// high byte of divisor
	outb (value, BGDH_REG);
	 outb (BaudRateDivisor & 0xff, BGDL_REG);
//	 outb (128, BGDL_REG);   // this sets to 1800 bps , for test sejin
//	 outb (3, BGDL_REG);   // this sets to 1800 bps , for test sejin
}



/************************************************************************
 * Syntax	: set_fifo_size		                                *
 *                                                                     	*
 * Description	: This function set the depth of the FIFOs (The EXCR2	*
 *		  register is in Bank 2, register 2).			*
 *                                                                     	*
 * Parameters	: None.                                             	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: The FIFO depth is set to 16. The FIFOs are always 	*
 *		  enabled in CEIR mode. So there is no need to mess with*
 *		  FCR register.						*
 *                                                                     	*
 ************************************************************************/

static void set_fifo_size (void)
{
	unsigned char value;

	if (cur_bank != 2)
		set_bank (2);
	switch(PROTOCOL_ID)
	{
	case TWIRP_PROTOCOL:
		value = inb (EXCR2_REG);
		value &= 0x0F;		// reset both FIFOs to 16 deep
		outb (value, EXCR2_REG);
		break;
	case SEJIN_PROTOCOL:
		value = inb (EXCR2_REG);
		value |= 0x04;		// reset both FIFOs to 16 deep
		outb (value, EXCR2_REG);
		break;
	case RC5_PROTOCOL:
		value = inb (EXCR2_REG);	
		value &= 0x0F;		// reset both FIFOs to 16 deep	
		outb (value, EXCR2_REG);	
		break;	
	default:
		break;
	}

}


/************************************************************************
 * Syntax	: set_demodulator	                                *
 *                                                                     	*
 * Description	: This function sets the frequency and bandwidth	*
 *		  sampled.(The receiver demodulator control register	*
 *		  IRRXDC is in bank 7, register 0)			*
 *                                                                     	*
 * Parameters	: frq_code                                             	*
 *		    - encoded frquency to watch for			*
 *		  bw_code						*
 *		    - encoded bandwidth					*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: Performs a soft reset on the receiver, clearing the 	*
 *		  FIFO. Also resets the threshold to 1 for both FIFOs.	*
 *                                                                     	*
 ************************************************************************/

static void set_demodulator (unsigned char freq_code, unsigned char bw_code)
{
	unsigned char value = (bw_code << 5) | freq_code;

	if (cur_bank != 7)
		set_bank (7);

	outb (value, IRRXDC_REG);
}


/************************************************************************
 * Syntax	: setup_ceir		                                *
 *                                                                     	*
 * Description	: This function performs some UART settings and SC2200  *
 *		  specific initialization.				*
 *                                                                     	*
 * Parameters	: None.                                             	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: None.						 	*
 *                                                                     	*
 ************************************************************************/

static void setup_ceir (void)
{
	unsigned long pmr;
	unsigned long mcr;
	unsigned char value;
   
     //   printk(" set up CIR \n");

	// Sets the IR port signals for SP3.
	pmr = inl (0x9030);		// read PMR
	pmr &= 0xffffffbf;		// clear bit 6, SP3SEL
	outl (pmr, 0x9030);

	// Set MCR bit 4 = 1 (irtxen = IRTX enabled)
	mcr = inl (0x9034);		// read the Misc Configuration Reg
	mcr |= 0x10;			// set IRTX/SOUT3 is enabled
	outl (mcr, 0x9034);

	if (cur_bank != 7)
		set_bank (7);

	// setup RCCFG register
	// old:  outb (0xd0, IO_BASE + 0x02);
	// for some reason SolutioNets did these as separate writes...

	// clear out RCCFG
	outb (0x0, RCCFG_REG);
   
	// enable RLE
	value = inb (RCCFG_REG);
	value |= 0x80;
	outb (value, RCCFG_REG);

	// disable internal demodulator
	value = inb (RCCFG_REG);
	value |= 0x10;
	outb (value, RCCFG_REG);

// ===================================* /
	set_bank (5);
	value = inb (IRCR2_REG);
	value &= ~0x02;		// MSR register and MS interrupt in UART mode
	value &= ~0x01;			// set half duplex mode
	outb (value, IRCR2_REG);

//==============sejin dongle mode ===
        set_bank (7);
         outb(0x48,IRCFG4);
	 outb(0x06,IRCFG1);
        set_bank(0);
//================================
}


/************************************************************************
 * Syntax	: set_bank		                                *
 *                                                                     	*
 * Description	: This function sets the active register bank for the   *
 *		  UART.							*
 *                                                                     	*
 * Parameters	: bank                                             	*
 *		    - index of bank (0 to 7)				*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: None.						 	*
 *                                                                     	*
 ************************************************************************/

static void set_bank (unsigned char bank)
{
	unsigned char val;

	static const char LINK_CONTROL = 0x00;
	static const char BANK_SELECT = 0x80;

	// the following values for the LCR are from p18 of the 87108 datasheet
	static const char CHARACTER_8_BITS = 0x03;
	static const char ONE_STOP_BIT = 0x00;
	static const char NO_PARITY = 0x00;
	static const char NO_BREAK = 0x00;

	// writes into the first two banks may upset LCR values, so reload
	static const char DEFAULT_PARAMS =
		CHARACTER_8_BITS | ONE_STOP_BIT | NO_PARITY | NO_BREAK;

	// The following values are from Table 4 in the 87108 datasheet, p19
	static const char code[8] =
	{
		LINK_CONTROL | DEFAULT_PARAMS,		// Bank 0
		BANK_SELECT | DEFAULT_PARAMS,		// Bank 1
		0xe0,								// Bank 2
		0xe4,								// Bank 3
		0xe8,								// Bank 4
		0xec,								// Bank 5
		0xF0,								// Bank 6
		0xF4								// Bank 7
	};

	if (bank > 7)
	{
		//printk ("nscir: set_bank:  bank %d out of range\n", bank);
		return;
	}

	val = code[bank];
	outb (val, BSR_REG);

	cur_bank = bank;
}



/************************************************************************
 * Syntax	: interrupt_handler	                                *
 *                                                                     	*
 * Description	: This is the top-half of the interrupt handler, called *
 *		  when the Rx FIFO threshold is reached. This handler	*
 *		  moves data from the FIFO to the IR input queue, and 	*
 *		  schedules the bottom half for immediate execution.	*
 *                                                                     	*
 * Parameters	: irq                                             	*
 *		    - Interrupt number (0 -15)				*
 *		  dev_id						*
 *		    - A unique indentification for sharing interrupts	*
 *		  regs							*
 *		    - Processor registers before the interrupt.		*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: None.						 	*
 *                                                                     	*
 ************************************************************************/

static void interrupt_handler (int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned char lsr;
	unsigned char byte;
	static int ff_cnt =0;
        static unsigned char pre_val =0;

	if (cur_bank != 0) set_bank (0);

	// read all data from the Rx FIFO
	lsr = inb (LSR_REG);
	while (lsr & 0x01)
	{
		byte = inb (RXD_REG);
		switch(PROTOCOL_ID)
		{
		case TWIRP_PROTOCOL :
			if ((byte & 0x7f) == 0x7f)
				byte |= 0x80;
			// enqueue data, and schedule bh
        	        // raw_ir_q_write(byte);

			NscIrReg.proto_twirp_write(byte);
			
			queue_task (&twirp_task, &tq_immediate);
	 		mark_bh (IMMEDIATE_BH);
		        break;

		case SEJIN_PROTOCOL :	
        		 sjnWIndex= (sjnWIndex ) % MAX_SJ_BUF;
	        	 sejinRawBuf[sjnWIndex] = byte;
	        	 sjnDataCount ++; 
		         sjnWIndex ++ ;
			 queue_task(&sejin_task,&tq_immediate);
	 		mark_bh (IMMEDIATE_BH);
	  		 break;

	  	case RC5_PROTOCOL :	
	  		 rc5WIndex= (rc5WIndex ) % MAX_RC5_BUF;	
	  		 rc5RawBuf[rc5WIndex] = byte;	
	  		 rc5DataCount ++; 	
	  	         rc5WIndex ++ ;	
	  		 queue_task(&rc5_task,&tq_immediate);	
	 		mark_bh (IMMEDIATE_BH);
	  		 break;	
	  		  		
	      	default :
			 break;
		}
	       
		lsr = inb (LSR_REG);
	}

	return;
}



/************************************************************************
 * Syntax	: stop_and_flush	                                *
 *                                                                     	*
 * Description	: This function turns off the IR receiver, and flushes 	*
 *		  the queue. This allows quicker shutdown of the 	*
 *		  receiver, and reduces the number of samples which has *
 *		  no IR energy.						*
 *                                                                     	*
 * Parameters	: None.                                             	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: None.						 	*
 *                                                                     	*
 ************************************************************************/

static void stop_and_flush (void)
{
	stop_reception ();
        // raw_ir_q_init();
	NscIrReg.proto_init[0]();
}


/************************************************************************
 * Syntax	: raw_ir_in		                                *
 *                                                                     	*
 * Description	: This is the bottom half of the interrupt handler, and *
 *		  is ran with interrupt enabled. The top half has 	*
 *		  received data, and placed it in a circular queue. This*
 *		  half will pull the data from the queue and process it *
 *		  for TWIRP protocol decoding.				*
 *                                                                     	*
 * Parameters	: unused			                       	*
 *		    - An unused parameter.				*
 *		      ( data is taken from the IR input queue ir_queue	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: Time isn't as critical here, as the top half runs 	*
 *		  independantly. As long as the queue doesn't overflow 	*
 *		  things will be OK.  It is hard to imagine IR data 	*
 *		  coming in so fast that it cannot be decoded before 	*
 *		  data is dropped.					*
 *                                                                     	*
 ************************************************************************/

static void raw_ir_in (void * unused)
{
	int ret_val ;
	int i;
//     if( (ret_val=NscIrRawDataTwirp(1))==1 )
     if( (ret_val=NscIrReg.proto_decode[0](1))==1 )
	{
   		stop_and_flush ();
		// ret_val=NscIrRawDataTwirp(2);
		ret_val=NscIrReg.proto_decode[0](2);
	}

	if(ret_val == 0xfff)
	{
		if((i=get_next_proto())!=0)
		switch_proto(i);
	}

}



/************************************************************************
 * Syntax	: raw_ir_in1		                                *
 *                                                                     	*
 * Description	: This is the bottom half of the interrupt handler, and *
 *		  is ran with interrupt enabled. The top half has 	*
 *		  received data, and placed it in a circular queue. This*
 *		  half will pull the data from the queue and process it *
 *		  for SEJIN protocol decoding.				*
 *                                                                     	*
 * Parameters	: unused			                       	*
 *		    - An unused parameter.				*
 *		      ( data is taken from the IR input queue ir_queue	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: Time isn't as critical here, as the top half runs 	*
 *		  independantly. As long as the queue doesn't overflow 	*
 *		  things will be OK.  It is hard to imagine IR data 	*
 *		  coming in so fast that it cannot be decoded before 	*
 *		  data is dropped.					*
 *                                                                     	*
 ************************************************************************/

static void raw_ir_in1 (void * unused)
{
	unsigned int i,j;
	while( sjnDataCount > 0 )
	{   sjnDataCount  --;
//		if(0xfff==NscIrRawDataSejin(sejinRawBuf[sjnRIndex]))
		if(0xfff==NscIrReg.proto_decode[1](sejinRawBuf[sjnRIndex]))
		{ 	//printk(" switching needed \n");
			if((i = get_next_proto()))
				switch_proto(i);
		}
		else
		{
         
	//	if(sejinRawBuf[sjnRIndex]!= 0xff)
	//	printk("%x ",sejinRawBuf[sjnRIndex]);

                sjnRIndex = sjnRIndex + 1 ;
         	sjnRIndex= (sjnRIndex ) % MAX_SJ_BUF;
		}
	};
}


/************************************************************************
 * Syntax	: raw_ir_in2		                                *
 *                                                                     	*
 * Description	: This is the bottom half of the interrupt handler, and *
 *		  is ran with interrupt enabled. The top half has 	*
 *		  received data, and placed it in a circular queue. This*
 *		  half will pull the data from the queue and process it *
 *		  for RC5 protocol decoding.				*
 *                                                                     	*
 * Parameters	: unused			                       	*
 *		    - An unused parameter.				*
 *		      ( data is taken from the IR input queue ir_queue	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: Time isn't as critical here, as the top half runs 	*
 *		  independantly. As long as the queue doesn't overflow 	*
 *		  things will be OK.  It is hard to imagine IR data 	*
 *		  coming in so fast that it cannot be decoded before 	*
 *		  data is dropped.					*
 *                                                                     	*
 ************************************************************************/

static void
raw_ir_in2 (void * unused)
{
	unsigned int i,j;
	while( rc5DataCount > 0 )
	{   rc5DataCount  --;

	//	if(0xfff==NscIrRawDataRC5(rc5RawBuf[rc5RIndex]))
		if(0xfff==NscIrReg.proto_decode[2](rc5RawBuf[rc5RIndex]))
		{ // printk(" switching is required \n"); 
			if((i=get_next_proto())!=0)
				switch_proto(i);
		}
		else
		{

//printk("<%x>  \n",rc5RawBuf[rc5RIndex]);
                	rc5RIndex = rc5RIndex + 1 ;
	         	rc5RIndex= (rc5RIndex ) % MAX_RC5_BUF;
		}
	};
}


/************************************************************************
 * Syntax	: install_handler	                                *
 *                                                                     	*
 * Description	: Hooks a handler for IO_IRQ. Interrupts on this IRQ	*
 *		  will be directed to the interrupt_handler(), the top  *
 *		  half of the handler.					*
 *                                                                     	*
 * Parameters	: None.                                             	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/

static void install_handler (void)
{
	int result;

	result = request_irq (IO_IRQ, interrupt_handler, SA_INTERRUPT,
		"nscir", NULL);
	if (result)
	{
		printk ("nscir:  can't get assigned irq %d\n", IO_IRQ);
		return;
	}
//	else printk ("nscir: interrupt %d hooked\n", IO_IRQ);

	// initialize task for bottom half of interrupt handler
	twirp_task.routine = raw_ir_in;
	sejin_task.routine = raw_ir_in1;
	rc5_task.routine = raw_ir_in2;


	// initialize queue indexes
        // raw_ir_q_init();

	if (cur_bank != 0)
		set_bank (0);
	// enable Rx interrupts
	outb (0x1, IER_REG);
}



/************************************************************************
 * Syntax	: remove_handler	                                *
 *                                                                     	*
 * Description	: This function unhooks the interrupt handler ( The IRQ	*
 *		  is unhooked, and interrupts are disabled).		*
 *                                                                     	*
 * Parameters	: None.                                             	*
 *                                                                     	*
 * Return 	: None.							*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/

static void remove_handler (void)
{
	outb (0x0, IER_REG);		// disable interrupts
	free_irq (IO_IRQ, NULL);	// release IRQ

//	printk ("nscir:  unhooked interrupt\n");
}

