/************************************************************************
 * File Name	  : twirp.c						*
 *									*
 * Project Name	  : ASTB						*
 *									*
 * Module Name 	  : NSCIR						*
 *									*
 * Author Name	  : Arun Kumar, Niraj Kumar				*
 *									*
 * Language	  : C							*
 *									*
 * Date		  : 08/10/01						*
 *									*
 * Description	  : This file contains the twirp protocol decoding part.*
 * 		    Once the decoding for the protocol is done then the *
 *		    received scan code will be interfaced with system	*
 *		    keyboard and busmouse drivers.			*
 *									*
 ************************************************************************/

#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/io.h>			// inb and outb
#include <linux/tqueue.h>	// for bottom-half
#include <linux/interrupt.h>
#include <linux/sched.h>	// for current
#include <linux/poll.h>		// poll method


#define VERSION "1.0"

// Various types of debug information
#define REPORT_RAW			0
#define REPORT_EDGES 			0
#define REPORT_RAW_PACKET		0
#define REPORT_STATUS			0
#define REPORT_TWIRP_DECODE		0
#define REPORT_COOKED_DATA		0
#define TWIRP_PROTO  1
// TWIPR registration values for supported devices (from SolutioNet)
#define TWIRP_REMOTE		0x0d
#define TWIRP_KEYBOARD		0x0e
#define	TWIRP_MOUSE		0x04


typedef int count_t;

#define TRUE 	1
#define FALSE	0

// The lower limit for a single pulse
#define PULSE_LIMIT		106
#define MAX_RLE			8


/* TWIRP constants */
const unsigned long TWIRP1X_PULSE_WIDTH = 212;		// 212us
const unsigned long TWIRP1X_DET_SETTLING = 758;		// 758us
const unsigned long TWIRP1X_PPM_WIDTH = 137;		// 137us
const unsigned long TWIRP1X_CLIPPING = 1;			// 13us
const unsigned long TWIRP1X_38KHZ_CARRIER = 26;		// 26us
const unsigned long DETECTOR_STRETCH = 250;			// 200us
const unsigned long DETECTOR_SHRINK = 100;			// 100us

#define MIN_EDGE_SEPARATION \
	 (TWIRP1X_PULSE_WIDTH + TWIRP1X_DET_SETTLING - (TWIRP1X_PPM_WIDTH))
#define MAX_EDGE_SEPARATION \
	(TWIRP1X_PULSE_WIDTH + TWIRP1X_DET_SETTLING + (15 * TWIRP1X_PPM_WIDTH) \
	+ (TWIRP1X_PPM_WIDTH))
#define MIN_PULSE_WIDTH \
	(TWIRP1X_PULSE_WIDTH - (2 * TWIRP1X_38KHZ_CARRIER) - DETECTOR_SHRINK)
#define MAX_PULSE_WIDTH \
	(TWIRP1X_PULSE_WIDTH + (2 * TWIRP1X_38KHZ_CARRIER) + DETECTOR_SHRINK)
#define MAX_INTER_PULSE \
	((TWIRP1X_DET_SETTLING + (16 * TWIRP1X_PPM_WIDTH)) * 1.2)

#define NUM_PULSES_PER_TWIRP_PACKET 	9
#define TWIRP1X_NUM_PACKET_BYTES 		4

#define IR_DETECTOR_TUNING_FIRST	(-45)
#define IR_DETECTOR_TUNING_REST		(-15)

unsigned short aPulseOffsets[NUM_PULSES_PER_TWIRP_PACKET];
unsigned short aWaveformOffsets[NUM_PULSES_PER_TWIRP_PACKET];
unsigned short packet[TWIRP1X_NUM_PACKET_BYTES];

unsigned char TWIRPWaveformReady = FALSE;
unsigned char TWIRPPacketReady = FALSE;

int  NscIrRawDataTwirp(int);


// TWIRP routines
static void TwirpDecodeIR (unsigned short edge_width, unsigned char is_on_edge);
static void process_waveform (unsigned short *waveform);
static void process_packet (unsigned short *packet);

#if REPORT_TWIRP_DECODE
static void print_modifier (char mod);
#endif

//==========Raw Queue management
 void raw_ir_q_init( void );
 unsigned char  raw_ir_q_read( void );
 void raw_ir_q_write(unsigned char);

// EXPORT_SYMBOL_NOVERS( NscIrRawDataTwirp);
//EXPORT_SYMBOL_NOVERS( raw_ir_q_init );
//EXPORT_SYMBOL_NOVERS( raw_ir_q_write );
int reset_counter =0;

static void enqueue_mouse (unsigned char lb, unsigned char rb, 
	unsigned char x_pos, unsigned char y_pos);


static struct tq_struct twirp_task;

/*************************************************************************
 * Queues                                                                *
 *                                                                       *
 * IR-Queue                                                              *
 *		Used to move data from the top-half and the bottom-half  *
 *		of the driver.  Data is enqueued by the top-half, and the*
 *		bottom-half is scheduled to run at a later time.         *
 *                                                                       *
 *************************************************************************/
#define IR_QUEUE_LEN	100
static unsigned char ir_queue[IR_QUEUE_LEN];
static unsigned short ir_in_index;
static unsigned short ir_out_index;

static unsigned char std_map[112]=
{
/* 
0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,
*/

00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
48,  30,  16,  44,0x00,  59,  02,0x00,  15,  31,  17,  45,  01,  41,  03,  58,
77,  32,  18,  46,0x00,0x00,  04,0x00,  57,  33,  19,  47,  06,  34,  05,  20,
40,  36,  22,  50,  07,  35,   8,  21,0x00,0x00,0x00,0x00, 126,0x00,0x00,0x00,
12,  37,  23,  51,0x00,  60,   9, 120,  26,  38,  24,  52, 121, 125,  10, 124,
28,  39,  25,  53,  13,  43,  11, 122, 127, 123,  27,0x00,  14,  72,  80,  75,
75,  00,0x00,  49,  73,  81,0x00,  77,0x00,0x00,0x00,0x00,0x00,0x00,  71,0x00
};

static unsigned char rmt_kbd[112]=
{
/* 
0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,
*/
  00,0x00,0x58,0x3c,0x20,0x00,0x00,0x00,0x00,0x00,0x69,0x5e,0x60,0x5d,0x00,0x00,
0x50,0x00,0x62,0x00,0x00,0x00,0x64,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x56,0x2c,
  00,0x00,0x00,0x00,0x00,0x00,0x4e,0x2e,0x00,0x00,0x00,0x00,0x00,0x00,0x46,0x26,
  00,0x00,0x00,0x00,0x00,0x00,0x36,0x1e,0x00,0x00,0x00,0x00,0x00,0x00,0x34,0x16,
  00,0x1c,0x6e,0x17,0x00,0x00,0x47,0x4c,0x57,0x59,0x4f,0x4d,0x65,0x00,0x00,0x00,
  00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
};



// EXPORT_SYMBOL_NOVERS( NscIrRawDataTwirp);
//EXPORT_SYMBOL_NOVERS( raw_ir_q_init );
//EXPORT_SYMBOL_NOVERS( raw_ir_q_write );

EXPORT_NO_SYMBOLS;


int init_module(void)
{

	if(0==NscIrRegProto(TWIRP_PROTO, raw_ir_q_init,NscIrRawDataTwirp,raw_ir_q_write ))
	{
		printk(" not able to load the twirp module \n"); 
	}
	return 0;
}



/************************************************************************
 * Syntax	: cleanup_module                                        *
 *                                                                     	*
 * Description	: This function clears the driver module from the Linux	*
 *		  kernel.						*
 *                                                                     	*
 * Parameters	: None.                                                	*
 *                                                                     	*
 * Return 	: None.	               					*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/

void cleanup_module(void)
{
	if(0==NscIrUnRegProto(TWIRP_PROTO))
	{
		printk(" not able to load the twirp module \n"); 
	}
}





/************************************************************************
 * Syntax	: raw_ir_q_init                                         *
 *                                                                     	*
 * Description	: This function initialises the raw ir queue. 		*
 *                                                                     	*
 * Parameters	: None                                                	*
 *                                                                     	*
 * Return 	: None				                 	*
 *                                                                     	*
 * Comment	: This is an exported function.			   	*
 *                                                                     	*
 ************************************************************************/

void raw_ir_q_init()
{
   ir_in_index =0;
   ir_out_index =0;
}


/************************************************************************
 * Syntax	: raw_ir_q_write                                        *
 *                                                                     	*
 * Description	: This function writes the data into the raw ir queue. 	*
 *                                                                     	*
 * Parameters	: None                                                	*
 *                                                                     	*
 * Return 	: None				                 	*
 *                                                                     	*
 * Comment	: This is an exported function.			   	*
 *                                                                     	*
 ************************************************************************/

void raw_ir_q_write( unsigned char byte)
{
	ir_queue[ir_in_index++] =  byte;
	if(ir_in_index == IR_QUEUE_LEN )
	ir_in_index =0;
}


/************************************************************************
 * Syntax	: raw_ir_q_init                                         *
 *                                                                     	*
 * Description	: This function reads data from the raw ir queue. 	*
 *                                                                     	*
 * Parameters	: None                                                	*
 *                                                                     	*
 * Return 	: None				                 	*
 *                                                                     	*
 * Comment	: This is an exported function.			   	*
 *                                                                     	*
 ************************************************************************/

unsigned char  raw_ir_q_read()
{  
	unsigned char byte;
	byte = ir_queue[ir_out_index++];
	if (ir_out_index == IR_QUEUE_LEN)
		ir_out_index = 0;
	return byte;

}



/************************************************************************
 * Syntax	: NscIrRawDataTwirp                                     *
 *                                                                     	*
 * Description	: This function reads the raw IR data and then call  	*
 *		  other functions to perform the TWIRP decoding.	*
 *                                                                     	*
 * Parameters	: None                                                	*
 *                                                                     	*
 * Return 	: None				                 	*
 *                                                                     	*
 * Comment	: This is an exported function.			   	*
 *                                                                     	*
 ************************************************************************/

int NscIrRawDataTwirp(int  partNo)
{
	static char combined_sent = 0;
	static int on_edge_len = 0;
	static int off_edge_len = 0;

	unsigned short edge_width;
	unsigned char is_on_edge;
	static int max_rle = 0;
	static int count = 0;
	unsigned char byte;

	const unsigned int MICRO_SEC_PER_RLE_TICK = 10;	
	if(partNo == 1)
	{
		// dequeue one byte of data
	        byte = raw_ir_q_read();

		// calculate the width in microseconds
		edge_width = (byte & 0x7f) + 1;
		edge_width *= MICRO_SEC_PER_RLE_TICK;

		// determine the edge state
		is_on_edge = (byte & 0x80) ? FALSE : TRUE;

		#if REPORT_RAW
		//printk ("nscir: width = %d ticks, %d us.  State = %d\n", 
	//			(byte & 0x7f) + 1, edge_width, is_on_edge);
		#endif

		if (on_edge_len)
			TwirpDecodeIR (edge_width * 0.84, is_on_edge);
		else
		{
			if (byte == 0xff)
			{
				max_rle++;
				off_edge_len += edge_width;
			}
			else
			{
				off_edge_len += edge_width;
				off_edge_len *= 0.87; // fudge factor
				TwirpDecodeIR (off_edge_len, is_on_edge);
				off_edge_len = 0;
			}
		}

		if( max_rle == MAX_RLE)  return 1;
	}

	if(partNo == 2)
	{
		// stop reception if things are running away
		if (max_rle == MAX_RLE)
		{
			#if REPORT_STATUS
			//printk ("nscir:  stopping reception\n");
			#endif
			//stop_and_flush ();

			TwirpDecodeIR (off_edge_len, is_on_edge);
			combined_sent = 0;
			off_edge_len = 0;
			max_rle = 0;

			count = 0;
		}
	}
	if(reset_counter >= 500 ) 
	{
		reset_counter =0; 
		return 0xfff;
	}
	else
		return 0;
}

/************************************************************************
 * Syntax	: TwirpDecodeIR				                *
 *                                                                     	*
 * Description	: This function does the actual decoding for the TWIRP 	*
 *		  data.							*
 *                                                                     	*
 * Parameters	: edge_width                                          	*
 *		    - The width of the edge in microseconds.		*
 *		  is_on_edge						*
 *		    - True if the edge is ON, False if OFF		*
 *                                                                     	*
 * Return 	: None				                 	*
 *                                                                     	*
 * Comment	: This function is based on code from SolutionNets.   	*
 *                                                                     	*
 ************************************************************************/

static void TwirpDecodeIR (unsigned short edge_width, unsigned char is_on_edge)
{
	static char last_was_on_edge = FALSE;
	static int pulse_count = 0;
	static char in_pulse = FALSE;
	static int delta_time = 0;
	char is_out_of_sync = FALSE;
	char is_transition = FALSE;

	static unsigned short width_on;
	static unsigned short width_off;
	static unsigned short edge_sep;
	static int packet_count = 0;

	// Is this an edge transition, or further increments for
	// the current edge

	#if REPORT_EDGES
	//printk ("nscir: edge width: %d, edge type: %s\n",
//		edge_width, is_on_edge ? "ON" : "OFF");
	#endif

	// see if an edge transition occured:  OFF -> ON, ON -> OFF
	if (is_on_edge == last_was_on_edge)
	{
		// add to the current edge being measured
		//probably won't happen much, due to demodulation and joining in
		// raw_ir_in routine

		delta_time += edge_width;

		if (! last_was_on_edge)
		{
			if (delta_time > MAX_INTER_PULSE)
			{
				// if we are in the middle of a waveform capture
				// then this would count as a fatal error
				if (pulse_count)
				{
				//printk ("nscir:error in middle of capture\n");
				}

				// can't be a valid packet
				// We must have got the last packet, or missed 
				// a pulse and don't want to combine pulses 
				// from two different packets

				is_out_of_sync = TRUE;
				//printk ("nscir:  out of sync, delta time too \
						great (1)\n");
			}
		}
	} // if - same edge
	else
	{
		is_transition = TRUE;

		if (last_was_on_edge)
		{
			// transitioning to an on-edge (OFF => ON)
			if (in_pulse)
			{
				// already in an on-edge, need an off-edge first
				//printk ("nscir: two on-edges, out of sync\n");
				is_out_of_sync = TRUE;
			}
		}
		else
		{
			// transitioning to an off-edge (ON => OFF)
			if (pulse_count == 0)
			{
				is_out_of_sync = TRUE;

				#if REPORT_STATUS
				// leading off edge before On-edge
				//printk ("nscir:  ignore off-edge when not in \
						packet capture\n");
				//printk("nscir:out of sync,pulse count = 0\n");
				#endif
			}
			else if (! in_pulse)
			{
				// can't start a new Off edge without first 
				// being in an On edge
				is_out_of_sync = TRUE;
		 	        //printk ("nscir:  off-edge without on-edge \
  					out of sync\n");
			}
			else if (delta_time > MAX_INTER_PULSE)
			{
				//printk ("nscir: max interpulse time exceeded,\
 					 out of sync\n");
				is_out_of_sync = TRUE;
			}
		} 	// off edge
	}		// Edge transition

	// start of new pulse
	if ((is_transition) && (! is_out_of_sync))
	{
		if (! in_pulse)
		{
			in_pulse = TRUE;
			width_on = delta_time;

			if ((width_on >= MIN_PULSE_WIDTH) &&
				(width_on <= MAX_PULSE_WIDTH))
			{
				if (pulse_count == 0)
				{
					aPulseOffsets[pulse_count++] = 0;
				}
				else
				{
					if (++pulse_count == 
						NUM_PULSES_PER_TWIRP_PACKET)
					{
						// Complete packet captured
						packet_count++;
						#if REPORT_TWIRP_DECODE
						//printk ("nscir: packet  %d \
						    in! \n", packet_count);
						#endif
						process_waveform(aPulseOffsets);

						// reset for next packet
						in_pulse = FALSE;
						pulse_count = 0;
						last_was_on_edge = FALSE;
						delta_time = 0;

						is_transition = FALSE;
					}

					#if REPORT_EDGES
				/*	//printk ("nscir: %d/%d pulses seen \
					  this packet\n", pulse_count,
						 NUM_PULSES_PER_TWIRP_PACKET);
				*/
					#endif
				} 		// else - PPM data pulse
			}			// if - valid pulse width
			else
			{
				//printk ("nscir: invalid pulse width\n");
				//printk ("nscir: width is %d us, min is %d,\
				 max is %d\n",width_on, MIN_PULSE_WIDTH, \
				 MAX_PULSE_WIDTH);
				is_out_of_sync = TRUE;
			}
		}		// if - start of new pulse
		else
		{
			in_pulse = FALSE;
			width_off = delta_time;
			edge_sep = width_on + width_off;

			if ((edge_sep >= MIN_EDGE_SEPARATION) &&
				(edge_sep <= MAX_EDGE_SEPARATION))
			{
				aPulseOffsets[pulse_count] = edge_sep;
			}
			else
			{
				//printk ("nscir: pulse edge sep length out \
				of range: " "%d, min=%d, max=%d\n",edge_sep,\
				MIN_EDGE_SEPARATION, MAX_EDGE_SEPARATION);
				is_out_of_sync = TRUE;
			}	// else - invalid edge to edge separation
		}		// if - end of the current pulse
	}			// if - edge transition

	if (is_out_of_sync)
	{
		in_pulse = FALSE;
		pulse_count = 0;
		last_was_on_edge = FALSE;
		delta_time = 0;
		reset_counter ++;
	}

	// did we just start a new edge?
	if (is_transition)
	{
		delta_time = edge_width;
		last_was_on_edge = is_on_edge;
	}
}		




/************************************************************************
 * Syntax	: process_waveform                                      *
 *                                                                     	*
 * Description	: After generation of edge positions, this function  	*
 *		  performs PPM decode and obtains nibble value for the 	*
 *		  packet.						*
 *                                                                     	*
 * Parameters	: waveform                                             	*
 *		    - Waveform data as a set of edge position timings.	*
 *                                                                     	*
 * Return 	: None				                 	*
 *                                                                     	*
 * Comment	: None.						   	*
 *                                                                     	*
 ************************************************************************/

 static void process_waveform (unsigned short *waveform)
{
	int i;
	unsigned short edge_sep;
	unsigned char nibble;
	unsigned char modulo;
	unsigned char checksum;
	unsigned char cur_byte;
	unsigned char is_high_byte;
	unsigned char is_valid_packet;

	checksum = 0;
	cur_byte = 0;
	is_high_byte = TRUE;
	is_valid_packet = TRUE;

	// PPM decode each pulse
	for (i = 1; i < NUM_PULSES_PER_TWIRP_PACKET; i++)
	{
		edge_sep = waveform[i];
		edge_sep -= (TWIRP1X_PULSE_WIDTH + TWIRP1X_DET_SETTLING);

		// adjust for 1/2 PPM cell width
		edge_sep += (TWIRP1X_PPM_WIDTH / 2);

		// fine tune incoming signal
		if (i == 1)
			edge_sep += IR_DETECTOR_TUNING_FIRST;
		else
			edge_sep += IR_DETECTOR_TUNING_REST;

		// calculate PPM nibble value
		nibble = (unsigned char) (edge_sep / TWIRP1X_PPM_WIDTH);

		// see where in cell we land
		modulo = (unsigned char) (edge_sep % TWIRP1X_PPM_WIDTH);

		checksum += nibble;

		// do PPM clipping
		if ((TWIRP1X_CLIPPING < modulo) &&
			(modulo < (TWIRP1X_PPM_WIDTH - TWIRP1X_CLIPPING)))
		{
			if (is_high_byte)
			{
				packet[cur_byte] = nibble << 4;
				is_high_byte = FALSE;
			}
			else
			{
				packet[cur_byte] |= nibble;
				cur_byte++;
				is_high_byte = TRUE;
			}
		}
		else
		{
			//printk ("nscir: bad edge placement in packet\n");
			is_valid_packet = FALSE;
			break;
		}
	}

	// see if checksum is correct
	if (i == NUM_PULSES_PER_TWIRP_PACKET)
	{
		if ((checksum & 0x0f) != 0)
		{
			//printk ("nscir: bad checksum 0x%x\n", checksum);
			is_valid_packet = FALSE;
		}
	}

	// pass validated packet to next stage of protocol decode
	if (is_valid_packet)
	{
		#if REPORT_TWIRP_DECODE
		//printk ("nscir: inbound TWIRP packet validated\n");
		#endif
		process_packet (packet);
	}
}



/************************************************************************
 * Syntax	: process_packet	                                *
 *                                                                     	*
 * Description	: This function process a validate TWIRP packet and then*
 *		  maps IR keyboard and remote keys to system keyboard 	*
 *		  driver.						*
 *									*
 * Parameters	: packet                                             	*
 *		    - The packet to process.			.	*
 *                                                                     	*
 * Return 	: None				                 	*
 *                                                                     	*
 * Comment	: In this function the playloads are extracted, and 	*
 *		  depending on the contents processing can take two	*
 *		  tracks:						*
 *		  1. Registration data is noted in a static variable 	*
 *		  (current_device).This identifies the sending device,  *
 *		  and precedes packets containi scancode data.  Since 	*
 *		  this data comes in separate packets, there a 		*
 *		  requirement for handling a keypress in two separate 	*
 *		  calls to process_packet.				*
 *		  2. Scancode data follows a registration packet, and 	*
 *		  identifies which key/button was pressed.  Depending on*
 *		  the registered device, this data is enqueued in a 	*
 *		  device specific queue.				*
 ************************************************************************/

static void process_packet (unsigned short *packet)
{
	unsigned char command = packet[1] & 0x0f;
	static char current_device = -1;
	static char payload_seen = 0;
	static char scancode_queued = 0;
	char lb_pressed, rb_pressed;

	int i;
	char sum = 0;
	unsigned char temp1;

	const char BASIC_REGISTRATION = 0x0f;

	reset_counter =0;

	#if REPORT_RAW_PACKET
		for (i = 0; i < TWIRP1X_NUM_PACKET_BYTES; i++)
		{
			//printk ("nscir: byte %d:  0x%02x\n", i, packet[i]);
		}
	#endif

	// calculate checksum
	for (i = 0; i < TWIRP1X_NUM_PACKET_BYTES; i++)
	{

		unsigned char high, low;

		high = packet[i] >> 4;
		low = packet[i] & 0x0f;
		sum += high;
		sum += low;
	}

	// validate checksum
	if (sum & 0x0f)
	{
	//printk ("nscir: bad packet checksum: %d, dropping packet\n", sum);
		return;
	}

	// packets are either registration or payload.
	// registration packets are handled specially: the device id is
	// extracted and associated with all payload packets until another
	// registration occurs
	if (command == BASIC_REGISTRATION)
	{
		unsigned short reg_num = packet[2] << 8 | packet[3];

		#if REPORT_TWIRP_DECODE
		//printk ("nscir: registration for device %d (0x%x)\n", reg_num, reg_num);
		#endif
		current_device = reg_num;
		payload_seen = 0;
		return;
	}

	// associate payload with previous registration
	switch (current_device)
	{
	case TWIRP_KEYBOARD:
		#if REPORT_TWIRP_DECODE
		//printk ("nscir: data in for keyboard\n");
		#endif
		if (! payload_seen)
		{
		// look at the repeat flag, process repeated keys separately
			if (packet[1] & 0x80)
			{
				// already queued, no need to requeue
				if (scancode_queued == packet[2])
					return;
 			}

		#if REPORT_TWIRP_DECODE
			//printk ("nscir: repeated key:  scan = %d (0x%02x)\n", packet[2], packet[2]);

			if (packet[3])
					print_modifier (packet[3]);
		#endif

		scancode_queued = packet[2];
		temp1 = std_map[scancode_queued];
		if((temp1 > 119) && (temp1 < 128))
		{
			handle_scancode(56,1);
			handle_scancode(temp1-118,1);
			handle_scancode(56,0);
		}

		else
		{
		if(packet[3] & 0x40 )  // shift
			handle_scancode(42 ,1);

		if(packet[3] & 0x80 )   // control
			handle_scancode(29 ,1); 

		if(packet[3] & 0x20 )  //alt
			handle_scancode( 56 ,1);


		handle_scancode(std_map[scancode_queued & 0x7f],1);

		if(scancode_queued ==0x1f)    // CAPS
			handle_scancode(std_map[scancode_queued],0);


		if(packet[3] & 0x40 )   // shift 0
			handle_scancode(42 ,0); 

		if(packet[3] & 0x80 )
			handle_scancode(29 ,0); 

		if(packet[3] & 0x20 )
			handle_scancode(  56 ,0); 

		}

		}

		break;

	case TWIRP_REMOTE:
		#if REPORT_TWIRP_DECODE
		//printk ("nscir: data in for remote, last was 0x%02x\n", scancode_queued);
		#endif
		if (! payload_seen)
		{
		// look at the repeat flag, process repeated keys separately
			if (packet[1] & 0x80)
			{
				// already queued, no need to requeue
				if (scancode_queued == packet[2])
					return;
                       	}
			#if REPORT_TWIRP_DECODE
			//printk ("nscir: key scan code = %d (0x%02x)\n", packet[2], packet[2]);
			if (packet[3])
				print_modifier (packet[3]);
			#endif

			// queue the data, and inhibit repeat data until new
			//  key press
			scancode_queued = packet[2];

			temp1 = std_map[rmt_kbd[scancode_queued & 0x7f]];
			if((temp1 > 119) && (temp1 < 128))
			{
				handle_scancode(56,1);
				handle_scancode(temp1-118,1);
				handle_scancode(56,0);
			}
			else
		handle_scancode(std_map[rmt_kbd[scancode_queued & 0x7f]],1); 
		}
		break;

	case TWIRP_MOUSE:
		#if REPORT_TWIRP_DECODE
		//printk ("nscir: data in for mouse\n");
		if (packet[1] & 0x10)
			//printk ("nscir: Right button\n");
		if (packet[1] & 0x20)
			//printk ("nscir: Left button\n");
		#endif

		// convert button press data to single bit flags
		lb_pressed = ((packet[1] & 0x20) != 0);
		rb_pressed = ((packet[1] & 0x10) != 0);

		#if REPORT_TWIRP_DECODE
		//printk ("nscir: X-Pos:%d, Y-Pos: %d\n", packet[2], packet[3]);
		#endif

		// and enqueue the data for the application layer
		enqueue_mouse (lb_pressed, rb_pressed, packet[2], packet[3]);
		break;

	default:
		//printk ("nscir: data in for unknown device %d\n", current_device);
		break;
	}
}




/************************************************************************
 * Syntax	: enqueue_mouse		                                *
 *                                                                     	*
 * Description	: This function takes the mouse pointer position and the*
 *		  button state and then maps that with system busmouse	*
 *		  driver.						*
 *                                                                     	*
 * Parameters	: l_button                                             	*
 *		    - 1 if the left button is pressed.			*
 *		  r_button						*
 *		    - 1 if the right button is pressed.			*
 *		  x_pos							*
 *		    - The raw x-position of the pointer.		*
 *		  y_pos							*
 *		    - The raw y-position of the pointer.		*
 *                                                                     	*
 * Return 	: None				                 	*
 *                                                                     	*
 * Comment	: None.			   				*
 *                                                                     	*
 ************************************************************************/

static void enqueue_mouse (unsigned char l_button, unsigned char r_button,
	unsigned char x_pos, unsigned char y_pos)
{
	unsigned long val;
	char delta_x = 0;
	char delta_y = 0;
	char x_sign = 0;
	char y_sign = 0;
	char mag;
	unsigned char temp;

	// convert absolute pointer positions into deltas.
	// Acceleration is implemented by measuring position from center point.
	// This asymetric value should be linearly scalable.

	if (x_pos > 128)
	{
		mag = (x_pos - 128) / 10;
		x_sign = 1;
		delta_x = -mag;
	}
	else if (x_pos > 0)
	{
		mag = (x_pos / 2) + 1;
		x_sign = 0;
		delta_x = mag;
	}

	if (y_pos > 128)
	{
		mag = (y_pos - 128) / 10;
		y_sign = 1;
		delta_y = -mag;
	}
	else if (y_pos > 0)
	{
		mag = (y_pos / 2) + 1;
		y_sign = 0;
		delta_y = mag;
	}

	// build up the masked value
	val = (l_button << 23);
	val |= (r_button << 22);
	val |= (x_sign << 21);
	val |= (y_sign << 20);
	val |= ((delta_x & 0xff) << 8);
	val |= (delta_y & 0xff);

	#if REPORT_COOKED_DATA
	//printk ("nscir: dx = %d, dy = %d, enqueueing raw mouse 0x%x\n", delta_x, delta_y, val);
	#endif

	/*****************************************/
	temp=0x07;
	if(l_button)  temp &= 0xfb;
	if(r_button)  temp &= 0xfe;
	ansc_mv_mouse(delta_x,delta_y,temp);
	
	/*****************************************/

}

#if REPORT_TWIRP_DECODE

/************************************************************************
 * Syntax	: print_modifier                                        *
 *                                                                     	*
 * Description	: This is an debug routine to see the state of the  	*
 *		  modifier keys.					*
 *                                                                     	*
 * Parameters	: mod                                             	*
 *		    - A bitmap of modifier key.				*
 *                                                                     	*
 * Return 	: None				                 	*
 *                                                                     	*
 * Comment	: This function prints the name of the modifier key 	*
 *		  pressed.						*
 *                                                                     	*
 ************************************************************************/

static void print_modifier (char mod)
{
	printk ("nscir: Modifiers: ");

	if (mod == 0x00)
	{
		printk ("None!\n");
		return;
	}

	if (mod & 0x80)
		printk ("CTRL ");

	if (mod & 0x40)
		printk ("SHIFT ");

	if (mod & 0x20)
		printk ("ALT ");

	if (mod & 0x08)
		printk ("FUNC ");

	printk ("\n");
}
#endif
